/*
 * $Id: mvdata.c,v 1.42.2.2 2008/09/21 17:42:45 hess Exp $
 * 
 *                This source code is part of
 * 
 *                 G   R   O   M   A   C   S
 * 
 *          GROningen MAchine for Chemical Simulations
 * 
 *                        VERSION 3.2.0
 * Written by David van der Spoel, Erik Lindahl, Berk Hess, and others.
 * Copyright (c) 1991-2000, University of Groningen, The Netherlands.
 * Copyright (c) 2001-2004, The GROMACS development team,
 * check out http://www.gromacs.org for more information.

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * If you want to redistribute modifications, please consider that
 * scientific software is very special. Version control is crucial -
 * bugs must be traceable. We will be happy to consider code for
 * inclusion in the official distribution, but derived work must not
 * be called official GROMACS. Details are found in the README & COPYING
 * files - if they are missing, get the official version at www.gromacs.org.
 * 
 * To help us fund GROMACS development, we humbly ask that you cite
 * the papers on the package - you can find them in the top README file.
 * 
 * For more info, check our website at http://www.gromacs.org
 * 
 * And Hey:
 * GROningen Mixture of Alchemy and Childrens' Stories
 */
/* This file is completely threadsafe - keep it that way! */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sysstuff.h>
#include <string.h>
#include "typedefs.h"
#include "main.h"
#include "mvdata.h"
#include "network.h"
#include "smalloc.h"
#include "gmx_fatal.h"
#include "symtab.h"
#include "vec.h"
#include "tgroup.h"

#define   block_bc(cr,   d) gmx_bcast(     sizeof(d),     &(d),(cr))
#define  nblock_bc(cr,nr,d) gmx_bcast((nr)*sizeof((d)[0]), (d),(cr))
#define    snew_bc(cr,d,nr) { if (!MASTER(cr)) snew((d),(nr)); }
/* Dirty macro with bAlloc not as an argument */
#define nblock_abc(cr,nr,d) { if (bAlloc) snew((d),(nr)); nblock_bc(cr,(nr),(d)); }

static void bc_string(const t_commrec *cr,t_symtab *symtab,char ***s)
{
  int handle;
  
  if (MASTER(cr)) {
    handle = lookup_symtab(symtab,*s);
  }
  block_bc(cr,handle);
  if (!MASTER(cr)) {
    *s = get_symtab_handle(symtab,handle);
  }
}

static void bc_strings(const t_commrec *cr,t_symtab *symtab,int nr,char ****nm)
{
  int  i;
  int  *handle;
  char ***NM;

  snew(handle,nr);
  if (MASTER(cr)) {
    NM = *nm;
    for(i=0; (i<nr); i++)
      handle[i] = lookup_symtab(symtab,NM[i]);
  }
  nblock_bc(cr,nr,handle);

  if (!MASTER(cr)) {
    snew_bc(cr,*nm,nr);
    NM = *nm;
    for (i=0; (i<nr); i++) 
      (*nm)[i] = get_symtab_handle(symtab,handle[i]);
  }
  sfree(handle);
}

static void bc_symtab(const t_commrec *cr,t_symtab *symtab)
{
  int i,nr,len;
  t_symbuf *symbuf;

  block_bc(cr,symtab->nr);
  nr = symtab->nr;
  snew_bc(cr,symtab->symbuf,1);
  symbuf = symtab->symbuf;
  symbuf->bufsize = nr;
  snew_bc(cr,symbuf->buf,nr);
  for (i=0; i<nr; i++) {
    if (MASTER(cr))
      len = strlen(symbuf->buf[i]) + 1;
    block_bc(cr,len);
    snew_bc(cr,symbuf->buf[i],len);
    nblock_bc(cr,len,symbuf->buf[i]);
  }
}

static void bc_block(const t_commrec *cr,t_block *block)
{
  block_bc(cr,block->nr);
  snew_bc(cr,block->index,block->nr+1);
  nblock_bc(cr,block->nr+1,block->index);
}

static void bc_blocka(const t_commrec *cr,t_blocka *block)
{
  block_bc(cr,block->nr);
  snew_bc(cr,block->index,block->nr+1);
  nblock_bc(cr,block->nr+1,block->index);
  block_bc(cr,block->nra);
  if (block->nra) {
    snew_bc(cr,block->a,block->nra);
    nblock_bc(cr,block->nra,block->a);
  }
}

static void bc_grps(const t_commrec *cr,t_grps grps[])
{
  int i;
  
  for(i=0; (i<egcNR); i++) {
    block_bc(cr,grps[i].nr);
    snew_bc(cr,grps[i].nm_ind,grps[i].nr);
    nblock_bc(cr,grps[i].nr,grps[i].nm_ind);
  }
}

static void bc_atoms(const t_commrec *cr,t_symtab *symtab,t_atoms *atoms)
{
  int dummy;

  block_bc(cr,atoms->nr);
  snew_bc(cr,atoms->atom,atoms->nr);
  nblock_bc(cr,atoms->nr,atoms->atom);
  bc_strings(cr,symtab,atoms->nr,&atoms->atomname);
  block_bc(cr,atoms->nres);
  bc_strings(cr,symtab,atoms->nres,&atoms->resname);
  /* QMMM requires atomtypes to be known on all nodes as well */
  bc_strings(cr,symtab,atoms->nr,&atoms->atomtype);
}

static void bc_groups(const t_commrec *cr,t_symtab *symtab,
		      int natoms,gmx_groups_t *groups)
{
  int dummy;
  int g,n;

  bc_grps(cr,groups->grps);
  block_bc(cr,groups->ngrpname);
  bc_strings(cr,symtab,groups->ngrpname,&groups->grpname);
  for(g=0; g<egcNR; g++) {
    if (MASTER(cr)) {
      if (groups->grpnr[g]) {
	n = natoms;
      } else {
	n = 0;
      }
    }
    block_bc(cr,n);
    if (n == 0) {
      groups->grpnr[g] = NULL;
    } else {
      snew_bc(cr,groups->grpnr[g],n);
      nblock_bc(cr,n,groups->grpnr[g]);
    }
  }
  if (debug) fprintf(debug,"after bc_groups\n");
}

void bcast_state_setup(const t_commrec *cr,t_state *state)
{
  block_bc(cr,state->natoms);
  block_bc(cr,state->ngtc);
  block_bc(cr,state->nrng);
  block_bc(cr,state->nrngi);
  block_bc(cr,state->flags);
}

void bcast_state(const t_commrec *cr,t_state *state,bool bAlloc)
{
  int i;

  bcast_state_setup(cr,state);

  if (MASTER(cr)) {
    bAlloc = FALSE;
  }
  if (bAlloc) {
    state->nalloc = state->natoms;
  }
  for(i=0; i<estNR; i++) {
    if (state->flags & (1<<i)) {
      switch (i) {
      case estLAMBDA:  block_bc(cr,state->lambda); break;
      case estBOX:     block_bc(cr,state->box); break;
      case estBOX_REL: block_bc(cr,state->box_rel); break;
      case estBOXV:    block_bc(cr,state->boxv); break;
      case estPRES_PREV: block_bc(cr,state->pres_prev); break;
      case estNH_XI:   nblock_abc(cr,state->ngtc,state->nosehoover_xi); break;
      case estTC_INT:  nblock_abc(cr,state->ngtc,state->therm_integral); break;
      case estX:       nblock_abc(cr,state->natoms,state->x); break;
      case estV:       nblock_abc(cr,state->natoms,state->v); break;
      case estSDX:     nblock_abc(cr,state->natoms,state->sd_X); break;
      case estCGP:     nblock_abc(cr,state->natoms,state->cg_p); break;
      case estLD_RNG:  if (state->nrngi == 1) nblock_abc(cr,state->natoms,state->ld_rng); break;
      case estLD_RNGI: if (state->nrngi == 1) nblock_abc(cr,state->natoms,state->ld_rngi); break;
      case estDISRE_INITF: block_bc(cr,state->hist.disre_initf); break;
      case estDISRE_RM3TAV:
	block_bc(cr,state->hist.ndisrepairs);
	nblock_abc(cr,state->hist.ndisrepairs,state->hist.disre_rm3tav);
	break;
      case estORIRE_INITF: block_bc(cr,state->hist.orire_initf); break;
      case estORIRE_DTAV:
	block_bc(cr,state->hist.norire_Dtav);
	nblock_abc(cr,state->hist.norire_Dtav,state->hist.orire_Dtav);
	break;
      default:
	gmx_fatal(FARGS,
		  "Communication is not implemented for %s in bcast_state",
		  est_names[i]);
      }
    }
  }
}

static void bc_ilists(const t_commrec *cr,t_ilist *ilist)
{
  int ftype;

  /* Here we only communicate the non-zero length ilists */
  if (MASTER(cr)) {
    for(ftype=0; ftype<F_NRE; ftype++) {
      if (ilist[ftype].nr > 0) {
	block_bc(cr,ftype);
	block_bc(cr,ilist[ftype].nr);
	nblock_bc(cr,ilist[ftype].nr,ilist[ftype].iatoms);
      }
    }
    ftype = -1;
    block_bc(cr,ftype);
  } else {
    for(ftype=0; ftype<F_NRE; ftype++) {
      ilist[ftype].nr = 0;
    }
    do {
      block_bc(cr,ftype);
      if (ftype >= 0) {
	block_bc(cr,ilist[ftype].nr);
	snew_bc(cr,ilist[ftype].iatoms,ilist[ftype].nr);
	nblock_bc(cr,ilist[ftype].nr,ilist[ftype].iatoms);
      }
    } while (ftype >= 0);
  }

  if (debug) fprintf(debug,"after bc_ilists\n");
}

static void bc_idef(const t_commrec *cr,t_idef *idef)
{
  block_bc(cr,idef->ntypes);
  block_bc(cr,idef->atnr);
  snew_bc(cr,idef->functype,idef->ntypes);
  snew_bc(cr,idef->iparams,idef->ntypes);
  nblock_bc(cr,idef->ntypes,idef->functype);
  nblock_bc(cr,idef->ntypes,idef->iparams);
  block_bc(cr,idef->fudgeQQ);
  bc_ilists(cr,idef->il);
  block_bc(cr,idef->ilsort);
}

static void bc_ffparams(const t_commrec *cr,gmx_ffparams_t *ffp)
{
  int i;
  
  block_bc(cr,ffp->ntypes);
  block_bc(cr,ffp->atnr);
  snew_bc(cr,ffp->functype,ffp->ntypes);
  snew_bc(cr,ffp->iparams,ffp->ntypes);
  nblock_bc(cr,ffp->ntypes,ffp->functype);
  nblock_bc(cr,ffp->ntypes,ffp->iparams);
  block_bc(cr,ffp->fudgeQQ);
}

static void bc_grpopts(const t_commrec *cr,t_grpopts *g)
{
  int i,n;
  
  block_bc(cr,g->ngtc);
  block_bc(cr,g->ngacc);
  block_bc(cr,g->ngfrz);
  block_bc(cr,g->ngener);
  snew_bc(cr,g->nrdf,g->ngtc);
  snew_bc(cr,g->tau_t,g->ngtc);
  snew_bc(cr,g->ref_t,g->ngtc);
  snew_bc(cr,g->acc,g->ngacc);
  snew_bc(cr,g->nFreeze,g->ngfrz);
  snew_bc(cr,g->egp_flags,g->ngener*g->ngener);

  nblock_bc(cr,g->ngtc,g->nrdf);
  nblock_bc(cr,g->ngtc,g->tau_t);
  nblock_bc(cr,g->ngtc,g->ref_t);
  nblock_bc(cr,g->ngacc,g->acc);
  nblock_bc(cr,g->ngfrz,g->nFreeze);
  nblock_bc(cr,g->ngener*g->ngener,g->egp_flags);
  snew_bc(cr,g->annealing,g->ngtc);
  snew_bc(cr,g->anneal_npoints,g->ngtc);
  snew_bc(cr,g->anneal_time,g->ngtc);
  snew_bc(cr,g->anneal_temp,g->ngtc);
  nblock_bc(cr,g->ngtc,g->annealing);
  nblock_bc(cr,g->ngtc,g->anneal_npoints);
  for(i=0;(i<g->ngtc); i++) {
    n = g->anneal_npoints[i];
    if (n > 0) {
      snew_bc(cr,g->anneal_time[i],n);
      snew_bc(cr,g->anneal_temp[i],n);
      nblock_bc(cr,n,g->anneal_time[i]);
      nblock_bc(cr,n,g->anneal_temp[i]);
    }
  }

  /* QMMM stuff, see inputrec */
  block_bc(cr,g->ngQM);
  snew_bc(cr,g->QMmethod,g->ngQM);
  nblock_bc(cr,g->ngQM,g->QMmethod);
  snew_bc(cr,g->QMbasis,g->ngQM);
  nblock_bc(cr,g->ngQM,g->QMbasis);
  snew_bc(cr,g->QMcharge,g->ngQM);
  nblock_bc(cr,g->ngQM,g->QMcharge);
  snew_bc(cr,g->QMmult,g->ngQM);
  nblock_bc(cr,g->ngQM,g->QMmult);
  snew_bc(cr,g->bSH,g->ngQM);
  nblock_bc(cr,g->ngQM,g->bSH);
  snew_bc(cr,g->CASorbitals,g->ngQM);
  nblock_bc(cr,g->ngQM,g->CASorbitals);
  snew_bc(cr,g->CASelectrons,g->ngQM);
  nblock_bc(cr,g->ngQM,g->CASelectrons);
  snew_bc(cr,g->SAon,g->ngQM);
  nblock_bc(cr,g->ngQM,g->SAon);
  snew_bc(cr,g->SAoff,g->ngQM);
  nblock_bc(cr,g->ngQM,g->SAoff);
  snew_bc(cr,g->SAsteps,g->ngQM);
  nblock_bc(cr,g->ngQM,g->SAsteps);
  /* end of QMMM stuff */
}

static void bc_cosines(const t_commrec *cr,t_cosines *cs)
{
  block_bc(cr,cs->n);
  snew_bc(cr,cs->a,cs->n);
  snew_bc(cr,cs->phi,cs->n);
  if (cs->n > 0) {
    nblock_bc(cr,cs->n,cs->a);
    nblock_bc(cr,cs->n,cs->phi);
  }
}

static void bc_pullgrp(const t_commrec *cr,t_pullgrp *pgrp)
{
  block_bc(cr,*pgrp);
  if (pgrp->nat > 0) {
    snew_bc(cr,pgrp->ind,pgrp->nat);
    nblock_bc(cr,pgrp->nat,pgrp->ind);
  }
  if (pgrp->nweight > 0) {
    snew_bc(cr,pgrp->ind,pgrp->nweight);
    nblock_bc(cr,pgrp->nweight,pgrp->weight);
  }
}

static void bc_pull(const t_commrec *cr,t_pull *pull)
{
  int g;

  block_bc(cr,*pull);
  snew_bc(cr,pull->grp,pull->ngrp+1);
  for(g=0; g<pull->ngrp+1; g++)
    bc_pullgrp(cr,&pull->grp[g]);
}

static void bc_inputrec(const t_commrec *cr,t_inputrec *inputrec)
{
  int i;
  
  block_bc(cr,*inputrec);
  bc_grpopts(cr,&(inputrec->opts));
  if (inputrec->ePull != epullNO) {
    snew_bc(cr,inputrec->pull,1);
    bc_pull(cr,inputrec->pull);
  }
  for(i=0; (i<DIM); i++) {
    bc_cosines(cr,&(inputrec->ex[i]));
    bc_cosines(cr,&(inputrec->et[i]));
  }
}

static void bc_moltype(const t_commrec *cr,t_symtab *symtab,
		       gmx_moltype_t *moltype)
{
  bc_string(cr,symtab,&moltype->name);
  bc_atoms(cr,symtab,&moltype->atoms);
  if (debug) fprintf(debug,"after bc_atoms\n");

  bc_ilists(cr,moltype->ilist);
  bc_block(cr,&moltype->cgs);
  bc_blocka(cr,&moltype->excls);
}

static void bc_molblock(const t_commrec *cr,gmx_molblock_t *molb)
{
  bool bAlloc=TRUE;
  
  block_bc(cr,molb->type);
  block_bc(cr,molb->nmol);
  block_bc(cr,molb->natoms_mol);
  block_bc(cr,molb->nposres_xA);
  if (molb->nposres_xA > 0) {
    snew_bc(cr,molb->posres_xA,molb->nposres_xA);
    nblock_bc(cr,molb->nposres_xA*DIM,molb->posres_xA[0]);
  }
  block_bc(cr,molb->nposres_xB);
  if (molb->nposres_xB > 0) {
    snew_bc(cr,molb->posres_xB,molb->nposres_xB);
    nblock_bc(cr,molb->nposres_xB*DIM,molb->posres_xB[0]);
  }
  if (debug) fprintf(debug,"after bc_molblock\n");
}

void bcast_ir_mtop(const t_commrec *cr,t_inputrec *inputrec,gmx_mtop_t *mtop)
{
  int i; 
  if (debug) fprintf(debug,"in bc_data\n");
  bc_inputrec(cr,inputrec);
  if (debug) fprintf(debug,"after bc_inputrec\n");
  bc_symtab(cr,&mtop->symtab);
  if (debug) fprintf(debug,"after bc_symtab\n");
  bc_string(cr,&mtop->symtab,&mtop->name);
  if (debug) fprintf(debug,"after bc_name\n");

  bc_ffparams(cr,&mtop->ffparams);

  block_bc(cr,mtop->nmoltype);
  snew_bc(cr,mtop->moltype,mtop->nmoltype);
  for(i=0; i<mtop->nmoltype; i++) {
    bc_moltype(cr,&mtop->symtab,&mtop->moltype[i]);
  }

  block_bc(cr,mtop->nmolblock);
  snew_bc(cr,mtop->molblock,mtop->nmolblock);
  for(i=0; i<mtop->nmolblock; i++) {
    bc_molblock(cr,&mtop->molblock[i]);
  }

  block_bc(cr,mtop->natoms);
  /* communication for mtop->atomtypes is missing here */

  bc_block(cr,&mtop->mols);
  bc_groups(cr,&mtop->symtab,mtop->natoms,&mtop->groups);
}

/*>>> MFFA >>>*/

void register_mffa_boundary_meta(MFFA_BOUNDARY_META *data, MPI_Datatype *type)
{
	int i, n=13;
	int block_lengths[n];
	MPI_Aint displacements[n];
	MPI_Aint addresses[n];
	MPI_Datatype typelist[n];

	typelist[0]  = MPI_INT;      /* TYPE */
	typelist[1]  = MPI_INT;      /* AXIS */
	typelist[2]  = MPI_INT;      /* index_nr_out */
	typelist[3]  = MPI_INT;      /* index_nr_in */
	typelist[4]  = GMX_MPI_REAL; /* pos_in */
	typelist[5]  = GMX_MPI_REAL; /* pos_out */
	typelist[6]  = GMX_MPI_REAL; /* GRID */
	typelist[7]  = GMX_MPI_REAL; /* a_out */
	typelist[8]  = GMX_MPI_REAL; /* b_out */
	typelist[9]  = GMX_MPI_REAL; /* a_in */
	typelist[10] = GMX_MPI_REAL; /* b_in */
	typelist[11] = GMX_MPI_REAL; /* CUTOFF */
	typelist[12] = GMX_MPI_REAL; /* center */

	for (i=0; i<12; i++) 
		block_lengths[i] = 1;
	block_lengths[12] = DIM;

	MPI_Address(data, &addresses[0]);
	MPI_Address(&(data->TYPE), &addresses[1]);
	MPI_Address(&(data->AXIS), &addresses[2]);
	MPI_Address(&(data->index_nr_out), &addresses[3]);
	MPI_Address(&(data->index_nr_in), &addresses[4]);
	MPI_Address(&(data->pos_in), &addresses[5]);
	MPI_Address(&(data->pos_out), &addresses[6]);
	MPI_Address(&(data->GRID), &addresses[7]);
	MPI_Address(&(data->a_out), &addresses[8]);
	MPI_Address(&(data->b_out), &addresses[9]);
	MPI_Address(&(data->a_in), &addresses[10]);
	MPI_Address(&(data->b_in), &addresses[11]);
	MPI_Address(&(data->CUTOFF), &addresses[12]);
	MPI_Address(&(data->center), &addresses[13]);

	for (i=0; i<n; i++)
		displacements[i]  = addresses[i+1]  - addresses[0];
	
	MPI_Type_struct(n, block_lengths, displacements, typelist, type);
	MPI_Type_commit(type);
}

void register_mffa_meta(MFFA_META *data, MPI_Datatype *type)
{
	int i, n=10;
	int block_lengths[n];
	MPI_Aint displacements[n];
	MPI_Aint addresses[n+1];
	MPI_Datatype typelist[n];

	typelist[0] = MPI_INT;      /* size_ndx */
	typelist[1] = MPI_INT;      /* bSDB */
	typelist[2] = MPI_INT;      /* bPC */
	typelist[3] = MPI_INT;      /* bNDX */
	typelist[4] = MPI_INT;      /* bQUENCH */
	typelist[5] = MPI_INT;      /* PNUMB */
	typelist[6] = GMX_MPI_REAL; /* Fb_in */
	typelist[7] = GMX_MPI_REAL; /* Fb_out */
	typelist[8] = GMX_MPI_REAL; /* P_int */
	typelist[9] = GMX_MPI_REAL; /* int_vir */
	
	for (i=0; i<9; i++) 
		block_lengths[i] = 1;
	block_lengths[9] = DIM * DIM;
	
	MPI_Address(data, &addresses[0]);
	MPI_Address(&(data->size_ndx), &addresses[1]);
	MPI_Address(&(data->bSDB), &addresses[2]);
	MPI_Address(&(data->bPC), &addresses[3]);
	MPI_Address(&(data->bNDX), &addresses[4]);
	MPI_Address(&(data->bQUENCH), &addresses[5]);
	MPI_Address(&(data->PNUMB), &addresses[6]);
	MPI_Address(&(data->Fb_in), &addresses[7]);
	MPI_Address(&(data->Fb_out), &addresses[8]);
	MPI_Address(&(data->P_int), &addresses[9]);
	MPI_Address(&(data->int_vir), &addresses[10]);

	for (i=0; i<n; i++)
		displacements[i]  = addresses[i+1]  - addresses[0];
	
	MPI_Type_struct(n, block_lengths, displacements, typelist, type);
	MPI_Type_commit(type);
}

void register_mffa_dynamic(MFFA_DYNAMIC *data, MPI_Datatype *type)
{
	int i, n=6;
	int block_lengths[n];
	MPI_Aint displacements[n];
	MPI_Aint addresses[n+1];
	MPI_Datatype typelist[n];

	for (i=0; i<n; i++) {
		typelist[i] = GMX_MPI_REAL;	
		block_lengths[i] = 1;
	}
	MPI_Address(data, &addresses[0]);
	MPI_Address(&(data->TFb_in), &addresses[1]);
	MPI_Address(&(data->TFb_out), &addresses[2]);
	MPI_Address(&(data->TPb_in), &addresses[3]);
	MPI_Address(&(data->TPb_out), &addresses[4]);
	MPI_Address(&(data->AREA_in), &addresses[5]);
	MPI_Address(&(data->AREA_out), &addresses[6]);

	for (i=0; i<n; i++)
		displacements[i]  = addresses[i+1]  - addresses[0];
	
	MPI_Type_struct(n, block_lengths, displacements, typelist, type);
	MPI_Type_commit(type);
}

void bcast_mffa(const t_commrec *cr, MFFA *data)
{
	int i, j, n=0, c=0, force_size=0;
	int *buffer;
	real *force_buffer;
	int **index;
	MFFA_META           proto_meta;
	MFFA_DYNAMIC        proto_dynamic;
	MFFA_BOUNDARY_META  proto_boundary_meta;
	MPI_Datatype mpi_meta, mpi_dynamic, mpi_boundary_meta;

	/* register custom MPI datatypes */
	register_mffa_meta(&proto_meta, &mpi_meta);
	register_mffa_dynamic(&proto_dynamic, &mpi_dynamic);
	register_mffa_boundary_meta(&proto_boundary_meta, &mpi_boundary_meta);

	if (MASTER(cr))
		fprintf(stderr, "Broadcasting MFFA.\n");
	/* first meta */
	MPI_Bcast(&(data->meta), 1, mpi_meta, MASTERRANK(cr), 
			cr->mpi_comm_mygroup);
	if (debug)
		fprintf(stderr, "  ..meta @ node %d\n", cr->nodeid);
	/* prepare dynamic arrays for receiving */
	if (!MASTER(cr)) {
		snew(data->boundary, data->meta.PNUMB);
		snew(data->dynamic, data->meta.PNUMB);
	}
	/* dynamic data for each boundary */
	for (i=0; i < data->meta.PNUMB; i++)
		MPI_Bcast(&(data->dynamic[i]), 1, mpi_dynamic, 
			MASTERRANK(cr), cr->mpi_comm_mygroup);
	if (debug)
		fprintf(stderr, "  ..dynamic @ node %d\n", cr->nodeid);

	/* prepare force_buffer */
	if (MASTER(cr)) {
		for (i=0; i < data->meta.PNUMB; i++) 
			force_size += data->boundary[i].meta.index_nr_out;
	}
	MPI_Bcast(&force_size, 1, MPI_INT, MASTERRANK(cr), cr->mpi_comm_mygroup);
	snew(force_buffer, force_size);
	/* meta data for each boundary & buffer forces in master node */
	for (i=0; i < data->meta.PNUMB; i++) {
		MPI_Bcast(&(data->boundary[i].meta), 1, mpi_boundary_meta, 
				MASTERRANK(cr), cr->mpi_comm_mygroup);
		if (debug) 
			fprintf(stderr, "  ..boundary[%d] @ node %d\n", i, cr->nodeid);
		if (MASTER(cr))
			for (j=0; j < data->boundary[i].meta.index_nr_out; j++) {
				force_buffer[c] = data->boundary[i].F_bounds[j];
				c++;
			}
	}
	MPI_Bcast(force_buffer, force_size, GMX_MPI_REAL, MASTERRANK(cr), 
			cr->mpi_comm_mygroup);
	/* store forces in non-master nodes */
	if (!MASTER(cr)) {
		for (i=0; i < data->meta.PNUMB; i++) {
			snew(data->boundary[i].F_bounds, 
					data->boundary[i].meta.index_nr_out);
			for (j=0; j < data->boundary[i].meta.index_nr_out; j++) {
				data->boundary[i].F_bounds[j] = force_buffer[c];
				c++;
			}
		}
	}
	if (debug) 
		fprintf(stderr, "  ..boundary[].F_bounds @ node %d\n",  cr->nodeid);
	/* boundary indeces */
	n = data->meta.size_ndx * (data->meta.PNUMB + 1);
	snew(buffer, n);
	if (MASTER(cr))
		for (i=0; i < data->meta.size_ndx; i++) 
			for (j=0; j <= data->meta.PNUMB; j++) 
				buffer[i * (data->meta.PNUMB + 1) + j] = data->index[i][j];
	MPI_Bcast(buffer, n, MPI_INT, MASTERRANK(cr), cr->mpi_comm_mygroup);
	if (!MASTER(cr)) {
		snew(data->index, data->meta.size_ndx);
		for (i=0; i < data->meta.size_ndx; i++) {
			snew(data->index[i], data->meta.PNUMB + 1);
		}
		for (i=0; i < data->meta.size_ndx; i++) {
			for (j=0; j <= data->meta.PNUMB; j++) 
				data->index[i][j] = buffer[i * (data->meta.PNUMB + 1) + j];
		}
	}
	if (debug) 
		fprintf(stderr, "  ..index @ node %d\n", cr->nodeid);
}
/*<<< MFFA <<<*/

