/*
 * $Id: md.c,v 1.287.2.26 2009/05/08 08:28:40 hess Exp $
 * 
 *                This source code is part of
 * 
 *                 G   R   O   M   A   C   S
 * 
 *          GROningen MAchine for Chemical Simulations
 * 
 *                        VERSION 3.2.0
 * Written by David van der Spoel, Erik Lindahl, Berk Hess, and others.
 * Copyright (c) 1991-2000, University of Groningen, The Netherlands.
 * Copyright (c) 2001-2004, The GROMACS development team,
 * check out http://www.gromacs.org for more information.

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * If you want to redistribute modifications, please consider that
 * scientific software is very special. Version control is crucial -
 * bugs must be traceable. We will be happy to consider code for
 * inclusion in the official distribution, but derived work must not
 * be called official GROMACS. Details are found in the README & COPYING
 * files - if they are missing, get the official version at www.gromacs.org.
 * 
 * To help us fund GROMACS development, we humbly ask that you cite
 * the papers on the package - you can find them in the top README file.
 * 
 * For more info, check our website at http://www.gromacs.org
 * 
 * And Hey:
 * Gallium Rubidium Oxygen Manganese Argon Carbon Silicon
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "gmx_fatal.h"
#include "typedefs.h"
#include "smalloc.h"
#include "sysstuff.h"
#include "vec.h"
#include "statutil.h"
#include "vcm.h"
#include "mdebin.h"
#include "nrnb.h"
#include "calcmu.h"
#include "index.h"
#include "vsite.h"
#include "update.h"
#include "ns.h"
#include "trnio.h"
#include "xtcio.h"
#include "mdrun.h"
#include "confio.h"
#include "network.h"
#include "pull.h"
#include "xvgr.h"
#include "physics.h"
#include "names.h"
#include "xmdrun.h"
#include "ionize.h"
#include "disre.h"
#include "orires.h"
#include "dihre.h"
#include "pppm.h"
#include "pme.h"
#include "mdatoms.h"
#include "repl_ex.h"
#include "qmmm.h"
#include "mpelogging.h"
#include "domdec.h"
#include "partdec.h"
#include "topsort.h"
#include "coulomb.h"
#include "constr.h"
#include "shellfc.h"
#include "compute_io.h"
#include "mvdata.h"
#include "checkpoint.h"
#include "mtop_util.h"

#ifdef GMX_MPI
#include <mpi.h>
#endif

/* The following two variables and the signal_handler function
 * is used from pme.c as well 
 */
extern bool bGotTermSignal, bGotUsr1Signal;

/*>>> MFFA >>> */
/*TODO: remove */
typedef struct {
  real distance;
  real fric;
} FRIC;
/*<<< MFFA <<< */

static RETSIGTYPE signal_handler(int n)
{
  switch (n) {
  case SIGTERM:
    bGotTermSignal = TRUE;
    break;
  case SIGUSR1:
    bGotUsr1Signal = TRUE;
    break;
  }
}

typedef struct { 
  gmx_integrator_t *func;
} gmx_intp_t;

/*>>> MFFA >>> */

void collect_mffa(MFFA_DYNAMIC *dynamic, real *pressure_sphere, 
		rvec *pressure_plane, int PNUMB, t_commrec *cr)
{
	int i, j, position=0;
#define SIZE 4
#define SIZE_BUFFER (2 + 2*DIM + SIZE*PNUMB)
	real buffer_send[SIZE_BUFFER], buffer_recv[SIZE_BUFFER];
	
	buffer_send[0] = pressure_sphere[0];
	buffer_send[1] = pressure_sphere[1];
	for (i=0; i < 2; i++) 
		for (j=0; j < DIM; j++)
			buffer_send[i*DIM + j + 2] = pressure_plane[i][j];
	for (i=0; i < PNUMB; i++) {
		position = 2*DIM + 2 + i*SIZE;
		buffer_send[position    ] = dynamic[i].TFb_in;
		buffer_send[position + 1] = dynamic[i].TFb_out;
		buffer_send[position + 2] = dynamic[i].TPb_in;
		buffer_send[position + 3] = dynamic[i].TPb_out;
	}
	MPI_Allreduce(buffer_send, buffer_recv, SIZE_BUFFER, 
			GMX_MPI_REAL, MPI_SUM, cr->mpi_comm_mygroup);
	pressure_sphere[0] = buffer_recv[0];
	pressure_sphere[1] = buffer_recv[1];
	for (i=0; i < 2; i++) 
		for (j=0; j < DIM; j++)
			pressure_plane[i][j] = buffer_recv[i*DIM + j + 2];
	for (i=0; i < PNUMB; i++) {
		position = 2*DIM + 2 + i*SIZE;
		dynamic[i].TFb_in   = buffer_recv[position    ];
		dynamic[i].TFb_out  = buffer_recv[position + 1];
		dynamic[i].TPb_in   = buffer_recv[position + 2];
		dynamic[i].TPb_out  = buffer_recv[position + 3];
	}
}

void correct_shell(MFFA_BOUNDARY_META *ref, MFFA_BOUNDARY_META *targ, 
		real *shift_in, real *shift_out) 
{
  /* hack for properly shifting thin shells used for restraining SHELL 
   * particles */
  real width;
  real middle;
  
  width = ref->pos_out - ref->pos_in;
  middle = ref->pos_in + 0.5*width;
  
  if(ref->pos_out < middle)
	  *shift_out = *shift_in;
  
  if(ref->pos_in > middle)
	  *shift_in = *shift_out;
}

void sdb_pc(int start, int nr_atoms, real *rout, real *rin, real refp, 
		real Pt, double dt, real tau, real *shift_out, real *shift_in, 
		real comp, rvec x[], rvec center)
{
  real dP,rout_old,rin_old;
  rvec mu;
  rvec dcent,old_cent,new_cent;
  int d,n,g;
  
  rout_old = *rout;
  rin_old = *rin;
  
  /* Pressure is now in bar, everywhere. */
  for (d=0; d<DIM; d++)
	  mu[d] = 1.0 - (comp*dt*(refp - Pt))/(DIM*tau);
  
  if (mu[XX]<0.99 || mu[XX]>1.01) 
	  fprintf(stderr,"Warning, scaling more than one pr.!!: mu %f\n", mu[XX]);
  
  /*calculate center shift*/
  new_cent[XX] = mu[XX]*center[XX];
  new_cent[YY] = mu[YY]*center[YY];
  new_cent[ZZ] = mu[ZZ]*center[ZZ];
  
  rvec_sub(center, new_cent, dcent);
  for (n=start; n<start+nr_atoms; n++) {
	  x[n][XX] = mu[XX]*x[n][XX] + dcent[XX];
	  x[n][YY] = mu[YY]*x[n][YY] + dcent[YY];
	  x[n][ZZ] = mu[ZZ]*x[n][ZZ] + dcent[ZZ];
  }
  
  *rout = mu[XX]*rout_old;
  *rin  = mu[XX]*rin_old;
  *shift_out = (*rout) - rout_old;
  *shift_in  = (*rin) - rin_old;
}

real calc_volume(MFFA_META *meta, MFFA_BOUNDARY_META *boundary, tensor box, 
		real *rad_in, real *rad_out)
{
  /* Note some set ups are not rock proof for every combination of potential 
   * type and/or index file !!! */
  real volume=0,r;
  int i;
  rvec new_box;
#define vsphere(r)((float)((4.0/3.0)*3.1415927*r*r*r))
#define vcilind(r,h)((float)(3.1415927*r*r*h))
  
  /* look at first sdb file only */
  if (meta->bSDB == TRUE) {
	  if (boundary->pos_in !=0)
		  *rad_in = boundary->pos_in - boundary->CUTOFF;
	  else
		  *rad_in = 0;
	  
	  if (boundary->pos_out !=0)
		  *rad_out = boundary->pos_out + boundary->CUTOFF;
	  else
		  *rad_out=0;
	  
	  if (boundary->TYPE == efSPHERE) {
		  if (*rad_out == 0)
			  volume = det(box) - vsphere(*rad_in);
		  /* difficult to say but I guess we have to treath it like a 
		   * big particle and put it's contributions in the virial*/
		  else
			  volume = vsphere(*rad_out)- vsphere(*rad_in);
	  } else if (boundary->TYPE == efLIN) {
		  /* now a special case for test runs, a uniform planar boundary 
		   * condition without wild cards*/
		  if(!meta->bNDX) {
			  for (i=0; i<DIM; i++)
				  new_box[i]=box[i][i];
			  new_box[boundary->AXIS] = *rad_out-*rad_in;
			  volume = new_box[0]*new_box[1]*new_box[2];
		  } else 
			  volume=det(box);
	  } else if (boundary->TYPE == efCILIND) {
		  /* cilindric cavities */
		  if (!meta->bNDX) {
			  if (boundary->pos_out == 0)
				  volume = det(box) - 
					  vcilind(*rad_in,box[boundary->AXIS][boundary->AXIS]);
			  if (boundary->pos_out != 0)
				  volume = vcilind(*rad_out,box[boundary->AXIS][boundary->AXIS]) 
					  - vcilind(*rad_in,box[boundary->AXIS][boundary->AXIS]);
		  } else
			  volume = det(box);
	  }
  } else
	  volume=det(box);
  
  return volume;
}

void split_string(char *string, char ***parts,int *numb)
{
  char *buf;
  int count, check=0, size=0, psize, i, next=0;
  
  /* number of potentials */
  snew(*parts,10);
  psize = 1;
  count = 0;
  size=(int)strlen(string);
  snew(buf, size+4); /* add 4 to size for backups _old */
  
  for (i=0; i<size; i++) {
	  if(string[i]=='#' && i!=0) {
		  next=0;
		  snew((*parts)[count],psize);
		  (*parts)[count] = strdup(buf);
		  /* clear buffer */
		  free(buf);
		  snew(buf,size);
		  psize = 0;
		  count++;
	  } else {
		  buf[next] = string[i];
		  next++;
		  psize++;
	  }
  }
  snew( (*parts)[count],psize);
  (*parts)[count] = strdup(buf);
  *numb = count+1;
}

/* interface for different potentials */
bool read_sdb_ndx(bool bNDB, char *filenm, int natoms, int ***SDB_IND, 
		int PNUMB)
{
  FILE *in;
  int i, j, loop=0, EXCLUDE=0;
  bool bWILD=FALSE, bEXCLUDE=FALSE;
  
  snew(*SDB_IND,natoms);

  if(bNDB) {
	  fprintf(stderr,"Reading index file.\n");
	  in = ffopen(filenm,"r");
	  for (i=0; i<natoms; i++) {
		  fscanf(in,"%d\n",&loop);
		  /* actually this has to be the real size of loop for economical 
		   * reasons.....however this assymetry will you give a hard time 
		   * during communication */
		  /* Note: +1 slaves should be told this in mvdata.c */
		  snew((*SDB_IND)[i],PNUMB+1);
		  (*SDB_IND)[i][0]=loop;
		  bEXCLUDE = FALSE;
		  for (j=1; j<=loop; j++) {
			  fscanf(in,"%d\n", &(*SDB_IND)[i][j]);
			  /* check for wildcards */
			  if ((*SDB_IND)[i][j] >= PNUMB)
				  bWILD = TRUE;
			  if ((*SDB_IND)[i][j] < PNUMB)
				  bEXCLUDE = TRUE;
		  }
		  if (bEXCLUDE)
			  EXCLUDE++;
	  }
	  ffclose(in);
  } else {
	  /* default each potential works on each atom */
	  for (i=0; i<natoms; i++) {  
		  snew((*SDB_IND)[i],PNUMB+1);
		  (*SDB_IND)[i][0] = PNUMB;
		  for (j=1; j<=PNUMB; j++)
			  (*SDB_IND)[i][1]= j-1;
	  }
  }
  /* the first boundary potential is dominating the PC coupling, even if 
   * exclusion groups are used............ when acting on each atom the 
   * first potential gives a hard geometrical contrain on the volume */
  if(EXCLUDE == natoms)
	  bWILD == FALSE;
  /* Now bWILD is always false when a boundary potential works one every 
   * atom in the system !!!! */
  return bWILD;
}

//TODO: CLEAN
void dump_mffa_boundary(MFFA_BOUNDARY *boundary, char *filenm, 
		bool bQUENCH, real time)
{
  FILE *out;
  int i;
  
  out = fopen(filenm, "w");
  if (boundary->meta.TYPE == efSPHERE) {
	  if (bQUENCH == FALSE)
		  fprintf(out, "SPHERE\n");
	  else 
		  fprintf(out, "QUENCH\n");
  } else if (boundary->meta.TYPE == efLIN) {
	  fprintf(out, "LINEAR\n");
	  fprintf(out, "%c\n", 88 + boundary->meta.AXIS);
  } else if (boundary->meta.TYPE == efCILIND) {
	  fprintf(out, "CILIND\n");
	  fprintf(out, "%c\n", 88 + boundary->meta.AXIS);
  }
  if (boundary->meta.pos_in != 0)
	  fprintf(out, "%f\n", (real)(boundary->meta.pos_in 
				  - boundary->meta.CUTOFF));
  else
	  fprintf(out, "%f\n", 0.0);
  
  if (boundary->meta.pos_out != 0)
	  fprintf(out, "%f\n", (real)(boundary->meta.pos_out 
				  + boundary->meta.CUTOFF));
  else
	  fprintf(out, "%f\n", 0.0);
  
  fprintf(out, "%f\n", boundary->meta.CUTOFF);
  fprintf(out, "%f\n", boundary->meta.GRID);
  fprintf(out, "%f\n", boundary->meta.a_in);
  fprintf(out, "%f\n", boundary->meta.a_out);
  fprintf(out, "%f\n", boundary->meta.center[XX]);
  fprintf(out, "%f\n", boundary->meta.center[YY]);
  fprintf(out, "%f\n", boundary->meta.center[ZZ]);
  
  if (boundary->meta.TYPE == efSPHERE)
	  fprintf(out, "%c type\n%c inner cutoff\n%c outer cutoff\n%c cutoff\n%c gridsize\n%c a in\n%c a out\n%c center X Y Z\n",37,37,37,37,37,37,37,37);
  else
	  fprintf(out, "%c type\n%c Axis\n%c inner cutoff\n%c outer cutoff\n%c cutoff\n%c gridsize\n%c a in (=> 0)\n%c a out (=>0)\n%c center X Y Z\n",37,37,37,37,37,37,37,37,37);
  
  fprintf(out, "%c version 4.0, 25 3 2008\n", 37);
  fprintf(out, "%c t = %8.3f (ns)\n", 37, time);
  fprintf(out, "#\n");
  for (i=0; i < boundary->meta.index_nr_in; i++)
	  fprintf(out, "%f\n", boundary->F_bounds[i]);
  fprintf(out, "#\n");
  for(; i < boundary->meta.index_nr_out; i++)
	  fprintf(out, "%f\n", boundary->F_bounds[i]);
  fprintf(out, "#\n");
  fclose(out);
}

void read_sdb(char *gfile, MFFA_BOUNDARY *SDB, int *TYPE, int *bQUENCH)
{
  int i=0, count, count2, check;
  real *buffer, number;
  FILE *in;
  fpos_t pos0, pos1, pos2;
  char c='a', type[10], buf[STRLEN], axtitle[STRLEN], VERS[STRLEN];
  bool loop=TRUE, loop2=TRUE, First1=TRUE, First2=TRUE, bCHECK=TRUE;
  real cutoff, radin, radout;
  
  if ((in=fopen(gfile,"r")) == NULL)
	  gmx_fatal(FARGS, "No .sdb file found!\n");
  
  fprintf(stderr, "Reading: %s\n", gfile);
  fscanf(in,"%s",type); 
  *bQUENCH = 0;
  if (strncmp(type,"SPHERE",6)==0)
	  *TYPE = efSPHERE;
  else if (strncmp(type,"LINEAR",6)==0)
	  *TYPE = efLIN;
  else if (strncmp(type,"QUENCH",6)==0) {
	  *TYPE = efSPHERE;
	  *bQUENCH = 1;
  } else if(strncmp(type,"CILIND",6)==0)
	  *TYPE = efCILIND;
  else
	  bCHECK = FALSE;
  
  if (!bCHECK)
	  gmx_fatal(FARGS, "Invalid potential type in '%s'.\n", gfile);
  else if (debug)
	  fprintf(stderr, "Selected potential type: %s\n", type);
  
  if (*TYPE == efLIN || *TYPE == efCILIND) {
	  fscanf(in,"%s",axtitle);
	  SDB->meta.AXIS = toupper(axtitle[0]) - 'X';
	  if(SDB->meta.AXIS < 0 || SDB->meta.AXIS > 2)
		  gmx_fatal(FARGS,"Axis specified in '%s' is incorrect.\n", gfile);
	  if (debug)
		  fprintf(stderr, "Potential along %c-axis\n", axtitle[0]);
  }
  fscanf(in,"%f",&radin);
  fscanf(in,"%f",&radout);
  fscanf(in,"%f",&cutoff);
  
  SDB->meta.CUTOFF = cutoff;
  if (radin!=0)
	  SDB->meta.pos_in=(real)(radin+cutoff);
  else
	  SDB->meta.pos_in=0;
  if (debug)
	  fprintf(stderr, "pos_in set to '%f' from the center\n", 
			  SDB->meta.pos_in);
  if(radout!=0)
	  SDB->meta.pos_out=(real)(radout-cutoff);
  else
	  SDB->meta.pos_out=0;
  if (debug)
	  fprintf(stderr, "pos_out set to '%f' from the center\n", 
			  SDB->meta.pos_out);
  
  fscanf(in, "%f", &SDB->meta.GRID);
  fscanf(in, "%f", &SDB->meta.a_in);
  fscanf(in, "%f", &SDB->meta.a_out);
  fscanf(in, "%f", &SDB->meta.center[XX]);
  fscanf(in, "%f", &SDB->meta.center[YY]);
  fscanf(in, "%f", &SDB->meta.center[ZZ]);
  if (debug) {
	  fprintf(stderr, "GRID-size set to: %f\n", SDB->meta.GRID);
	  fprintf(stderr, "a_in set to: %f\n", SDB->meta.a_in);
	  fprintf(stderr, "a_out set to: %f\n", SDB->meta.a_out);
	  fprintf(stderr, "center set to: %f %f %f\n", SDB->meta.center[XX], 
			  SDB->meta.center[YY], SDB->meta.center[ZZ]);
  }
  /* check version */
  /* newer version shorter length */
  fscanf(in,"%s",VERS);
  if (strncmp(VERS,"%",1)!=0 && strncmp(VERS,"#",1) != 0)
	  gmx_fatal(FARGS, "Format not correct! Seems you are using an old .sdb version\n");
  
  /* make sure a_in a_out are not negative */
  SDB->meta.a_in  = fabs(SDB->meta.a_in);
  SDB->meta.a_out = fabs(SDB->meta.a_out);  
  
  if (debug)
	  fprintf(stderr, "Reading potentials...\n");
  check=0;
  count=-1;
  count2=-1;
  rewind(in);
  while ((c = getc(in))!=EOF && check != 3) {
	  if (c=='#')
		  check++;
	  if (check==1) {
		  if (First1)
			  fgetpos(in,&pos1);
		  if (c=='\n')
			  count++;
		  First1=FALSE;
	  }
	  if (check==2) {
		  if (First2) 
			  fgetpos(in,&pos2);
		  if (c=='\n')
			  count2++;
		  First2=FALSE;
	  }
  }
  
  if (check!=3)
	  gmx_fatal(FARGS,"Less than 3 #-characters in the .sdb file");
  
  SDB->meta.index_nr_in = count;
  SDB->meta.index_nr_out = count2 + count;
  fprintf(stderr, "Read %d data points for the inner and %d for the outer potential.\n", count, count2);
  
  snew(SDB->F_bounds,(count+count2));
  fsetpos(in,&pos1);
  for (i=0;i<count;i++) {
	  fscanf(in,"%f\n",&SDB->F_bounds[i]);
	  if (debug)
		  fprintf(stderr, "inner force: %f  %f\n", 
				  (float)(SDB->meta.pos_in 
					  - (SDB->meta.GRID * (SDB->meta.index_nr_in - i))), 
				  SDB->F_bounds[i]);
  }
  fsetpos(in,&pos2);
  for (i=0;i<count2;i++) {
	  fscanf(in,"%f\n",&SDB->F_bounds[i+count]);
	  if (debug)
		  fprintf(stderr, "outer force: %f  %f\n", 
				  (float)((SDB->meta.GRID*i) + SDB->meta.pos_out), 
				  SDB->F_bounds[i+count]);
  }
  fclose(in);
}

void do_sdb_force(int start, int homenr, rvec f[], rvec x[], 
		MFFA *mffa, rvec *pressure_plane, real *pressure_sphere, 
		matrix box, matrix force_vir)
{
  int j, m, n, d, k, index, NUMB=0, i, axis;
  real distance=0, fit, hbox=0, shift=0, boxz=0;
  rvec distvec, f_old, box_vec;
  bool bWILDCARD=TRUE, bEMPTY;
  real RIN=0, ROUT=0, bforce=0, pot_dist=0, LAST_IN=0, LAST_OUT=0;
  matrix bound_vir;
  static bool bFIRST = FALSE;
  
  /* clear arrays for pressures and forces */
  clear_mat(bound_vir);
  for (k=0; k<mffa->meta.PNUMB; k++) { 
	  mffa->dynamic[k].TFb_in  = 0;
	  mffa->dynamic[k].TFb_out = 0;
	  mffa->dynamic[k].TPb_in  = 0;
	  mffa->dynamic[k].TPb_out = 0;
  }
  
  clear_rvec(pressure_plane[0]);
  clear_rvec(pressure_plane[1]);
  pressure_sphere[0] = 0;
  pressure_sphere[1] = 0;
  
  /* loop over n atoms */
  clear_rvec(distvec);
  bWILDCARD = FALSE;
  bforce = 0;
  for (n=start; n<start+homenr; n++) { 
	  /* store old force */
	  for (d=0; d<DIM; d++)
		  f_old[d] = f[n][d];
	  for (k=1; k <= mffa->index[n][0]; k++) { 
		  clear_rvec(distvec);
		  bWILDCARD = TRUE;
		  NUMB = mffa->index[n][k];
		  if (NUMB >= mffa->meta.PNUMB)
			  bWILDCARD = FALSE;
		  /* If FALSE; don't touch me............. */
		  /* reset bforce */
		  bforce = 0;
		  if (bWILDCARD) {
			  bEMPTY = FALSE;
			  /* hack to prevent empty arrays */
			  if (mffa->boundary[NUMB].meta.index_nr_out != 0)
				  LAST_OUT = mffa->boundary[NUMB].F_bounds[mffa->boundary[NUMB].meta.index_nr_out - 1];
			  else {
				  LAST_OUT = 0; 
				  bEMPTY = TRUE;
			  }
			  if (mffa->boundary[NUMB].meta.index_nr_in != 0)
				  LAST_IN = mffa->boundary[NUMB].F_bounds[0];
			  else
				  LAST_IN = 0;
			  if (mffa->boundary[NUMB].meta.TYPE == efSPHERE) {
				  /* spheres */
				  for (d=0; d<DIM; d++)
					  distvec[d] = x[n][d] 
						  - mffa->boundary[NUMB].meta.center[d];
				  distance = sqrt(iprod(distvec,distvec));
				  /* unit normal vec */
				  for (d=0; d<DIM; d++)
					  distvec[d] = distvec[d]/distance;
			  } else if (mffa->boundary[NUMB].meta.TYPE == efLIN) {
				  /* planes */
				  axis = mffa->boundary[NUMB].meta.AXIS;
				  distance = x[n][axis];
				  /* unit normal vec */
				  clear_rvec(distvec);
				  distvec[axis] = 1.0;
			  } else if (mffa->boundary[NUMB].meta.TYPE == efCILIND) {
				  /* Cylinder */
				  for (d=0; d<DIM; d++)
					  distvec[d] = x[n][d] 
						  - mffa->boundary[NUMB].meta.center[d];
				  axis = mffa->boundary[NUMB].meta.AXIS;
				  /* cilindric axis */
				  distvec[axis] = 0;
				  distance = sqrt(iprod(distvec,distvec));
				  /* unit normal vec */
				  for (d=0; d<DIM; d++)
					  distvec[d] = distvec[d]/distance;
			  } else
				  gmx_fatal(FARGS,"Type error in one of the .sdb files!\n");
			  if (mffa->boundary[NUMB].meta.pos_out != 0)
				  /* check outher boundary is switched on anyway */
				  if (distance > mffa->boundary[NUMB].meta.pos_out) {
					  index = ((int)((distance - mffa->boundary[NUMB].meta.pos_out) 
								  / mffa->boundary[NUMB].meta.GRID)) 
						  + mffa->boundary[NUMB].meta.index_nr_in;
					  if (index < mffa->boundary[NUMB].meta.index_nr_out 
							  && !bEMPTY) {
						  bforce = mffa->boundary[NUMB].F_bounds[index];
						  mffa->dynamic[NUMB].TFb_out += bforce; 
						  for (d=0; d<DIM; d++)
							  f[n][d] += distvec[d]*bforce;
					  } else {
						  /* a_in > 0 pot_dist > 0 */
						  pot_dist = (distance 
								- mffa->boundary[NUMB].meta.pos_out) 
							  	- ((mffa->boundary[NUMB].meta.index_nr_out 
								- mffa->boundary[NUMB].meta.index_nr_in) 
								* mffa->boundary[NUMB].meta.GRID);
						  fit = LAST_OUT - (mffa->boundary[NUMB].meta.a_out 
								  * pot_dist);
						  bforce = fit;
						  mffa->dynamic[NUMB].TFb_out += bforce;
						  for (d=0; d<DIM; d++)
							  f[n][d] += distvec[d]*bforce;
					  }
					  for (j=0; j < DIM; j++)
						  for (m=0; m < DIM; m++)
							  bound_vir[j][m] += bforce * distvec[j] 
								  	* distvec[m] 
									* (mffa->boundary[NUMB].meta.pos_out 
									+ mffa->boundary[NUMB].meta.CUTOFF 
									- distance);
				  }
			  if (distance < mffa->boundary[NUMB].meta.pos_in) {
				  index = (int)(((distance 
						- mffa->boundary[NUMB].meta.pos_in) 
						/ mffa->boundary[NUMB].meta.GRID) 
						+ mffa->boundary[NUMB].meta.index_nr_in);
				  if (index<0) {
					  /* a_in > 0 and pot_dist > 0 */ 
					  pot_dist = (mffa->boundary[NUMB].meta.pos_in 
							- distance) 
						  	- (mffa->boundary[NUMB].meta.index_nr_in 
							* mffa->boundary[NUMB].meta.GRID);
					  fit = LAST_IN 
						  - (mffa->boundary[NUMB].meta.a_in * pot_dist);
					  bforce = fit;
					  mffa->dynamic[NUMB].TFb_in += bforce;
					  for (d=0; d<DIM; d++)
						  f[n][d] += distvec[d] * bforce * -1.0;
				  } else if (index<mffa->boundary[NUMB].meta.index_nr_in) {
					  bforce = mffa->boundary[NUMB].F_bounds[index];
					  mffa->dynamic[NUMB].TFb_in += bforce;
					  for (d=0; d<DIM; d++)
						  f[n][d] += distvec[d]* bforce * -1.0;
				  }
				  for (j=0; j < DIM; j++)
					  for (m=0; m < DIM; m++)
						  bound_vir[j][m] += bforce * distvec[j] 
							  	* distvec[m] * (distance 
								- mffa->boundary[NUMB].meta.pos_in 
								+ mffa->boundary[NUMB].meta.CUTOFF);
			  }
		  }
	  }
	  /**********uniform***********/
  }
  /* multiply by half due to the gromacs virial definition */
  msmul(bound_vir, 0.5, bound_vir); 
  /* add boundary contributions to virial */
  m_add(bound_vir,force_vir,force_vir);
  /* Boundary pressure, AREA, etc */
  ROUT=0;
  RIN=0;
#define asphere(r)((float)(4.0*3.1415927*r*r))
#define laplace(r,s)((float)((s*2)/r))
#define acilind(r,h)((float)(3.1415927*2.0*r*h))
  
  for (k=0; k < mffa->meta.PNUMB; k++) {
	  if (mffa->boundary[k].meta.pos_in != 0)
		  RIN = mffa->boundary[k].meta.pos_in 
			  - mffa->boundary[k].meta.CUTOFF;
	  else
		  RIN  = 0;
	  if (mffa->boundary[k].meta.pos_out != 0)
		  ROUT = mffa->boundary[k].meta.pos_out 
			  + mffa->boundary[k].meta.CUTOFF;
	  else
		  ROUT = 0;
	  
	  if (mffa->boundary[k].meta.TYPE == efLIN) {
		  for (d=0; d<DIM; d++)
			  box_vec[d] = box[d][d];
		  box_vec[mffa->boundary[k].meta.AXIS] = 1.0;
		  /* rectangular only */
		  mffa->dynamic[k].AREA_in = mffa->dynamic[k].AREA_out 
			  = box_vec[0]*box_vec[1]*box_vec[2];
	  } else if (mffa->boundary[k].meta.TYPE == efSPHERE) {
		  mffa->dynamic[k].AREA_in = asphere(RIN);
		  mffa->dynamic[k].AREA_out = asphere(ROUT);
	  } else if (mffa->boundary[k].meta.TYPE == efCILIND) {
		  boxz = box_vec[mffa->boundary[k].meta.AXIS];
		  mffa->dynamic[k].AREA_in = acilind(RIN, boxz);
		  mffa->dynamic[k].AREA_out = acilind(ROUT,boxz);
	  }
	  if (RIN != 0)
		  mffa->dynamic[k].TPb_in = -1.0 * PRESFAC 
			  * (mffa->dynamic[k].TFb_in / mffa->dynamic[k].AREA_in);
	  else
		  mffa->dynamic[k].TPb_in = 0;
	  
	  if (mffa->boundary[k].meta.TYPE == efSPHERE)
		  pressure_sphere[0] += mffa->dynamic[k].TPb_in;
	  else if (mffa->boundary[k].meta.TYPE == efLIN)
		  pressure_plane[0][mffa->boundary[k].meta.AXIS] 
			  += mffa->dynamic[k].TPb_in;
	  
	  /* total uniform */
	  if (ROUT != 0)
		  mffa->dynamic[k].TPb_out = -1.0 * PRESFAC 
			  * (mffa->dynamic[k].TFb_out 
				/ mffa->dynamic[k].AREA_out);
	  else
		  mffa->dynamic[k].TPb_out = 0;
	  /* total uniform */
	  if (mffa->boundary[k].meta.TYPE == efSPHERE)
		  pressure_sphere[1] += mffa->dynamic[k].TPb_out;
	  else if (mffa->boundary[k].meta.TYPE == efLIN)
		  pressure_plane[1][mffa->boundary[k].meta.AXIS] 
			  += mffa->dynamic[k].TPb_out;
  }
  bFIRST = TRUE;
}
/*<<< MFFA <<<*/

/* The array should match the eI array in include/types/enums.h */
const gmx_intp_t integrator[eiNR] = { {do_md}, {do_steep}, {do_cg}, {do_md}, {do_md}, {do_nm}, {do_lbfgs}, {do_tpi}, {do_tpi}, {do_md} };

void mdrunner(FILE *fplog,t_commrec *cr,int nfile,t_filenm fnm[],
	      bool bVerbose,bool bCompact,
	      ivec ddxyz,int dd_node_order,real rdd,real rconstr,
	      char *dddlb_opt,real dlb_scale,
	      char *ddcsx,char *ddcsy,char *ddcsz,
	      int nstepout,gmx_edsam_t ed,int repl_ex_nst,int repl_ex_seed,
	      real pforce,real cpt_period,real max_hours,
	      unsigned long Flags)
{
  double     nodetime=0,realtime;
  t_inputrec *inputrec;
  t_state    *state=NULL;
  matrix     box;
  rvec       *buf=NULL,*f=NULL;
  real       tmpr1,tmpr2;
  t_nrnb     *nrnb;
  gmx_mtop_t *mtop=NULL;
  t_mdatoms  *mdatoms=NULL;
  t_forcerec *fr=NULL;
  t_fcdata   *fcd=NULL;
  real       ewaldcoeff=0;
  gmx_pme_t  *pmedata=NULL;
  time_t     start_t=0;
  gmx_vsite_t *vsite=NULL;
  gmx_constr_t constr;
  int        i,m,nChargePerturbed=-1,status,nalloc;
  char       *gro;
  gmx_wallcycle_t wcycle;
  bool       bReadRNG,bReadEkin;
  int        list;
  int        nsteps_done;
  /*>>> MFFA >>>*/
  MFFA               *mffa=NULL;
  /*<<< MFFA <<<*/

  snew(inputrec,1);
  snew(mtop,1);
  /*>>> MFFA >>>*/
  snew(mffa, 1);

  mffa->meta.bSDB = FALSE;
  mffa->meta.bPC = FALSE;
  /*<<< MFFA <<<*/

  if (bVerbose && SIMMASTER(cr)) 
    fprintf(stderr,"Getting Loaded...\n");
  
  if (Flags & MD_APPENDFILES) 
  {
	  fplog = NULL;
  }
	
  if (PAR(cr)) {
    /* The master thread on the master node reads from disk, 
     * then distributes everything to the other processors.
     */
	
    list = (SIMMASTER(cr) && !(Flags & MD_APPENDFILES)) ?  (LIST_SCALARS | LIST_INPUTREC) : 0;
		  
    snew(state,1);
    init_parallel(fplog,ftp2fn(efTPX,nfile,fnm),cr,
		  inputrec,mtop,state,list);
	  
  } else {
    /* Read a file for a single processor */
    snew(state,1);
    init_single(fplog,inputrec,ftp2fn(efTPX,nfile,fnm),mtop,state);
  }
  if (!EEL_PME(inputrec->coulombtype) || (Flags & MD_PARTDEC)) {
    cr->npmenodes = 0;
  }
  
  /* NMR restraints must be initialized before load_checkpoint,
   * since with time averaging the history is added to t_state.
   * For proper consistency check we therefore need to extend
   * t_state here.
   * So the PME-only nodes (if present) will also initialize
   * the distance restraints.
   */
  snew(fcd,1);

  /* This needs to be called before read_checkpoint to extend the state */
  init_disres(fplog,mtop,inputrec,cr,Flags & MD_PARTDEC,fcd,state);
					  
  if (gmx_mtop_ftype_count(mtop,F_ORIRES) > 0) {
    if (PAR(cr) && !(Flags & MD_PARTDEC)) {
      gmx_fatal(FARGS,"Orientation restraints do not work (yet) with domain decomposition, use particle decomposition (mdrun option -pd)");
    }
    /* Orientation restraints */
    if (MASTER(cr)) {
      init_orires(fplog,mtop,state->x,inputrec,cr->ms,&(fcd->orires),state);
    }
  }
  
  if (DEFORM(*inputrec)) {
    /* Store the deform reference box before reading the checkpoint */
    if (SIMMASTER(cr)) {
      copy_mat(state->box,box);
    }
    if (PAR(cr)) {
      gmx_bcast(sizeof(box),box,cr);
    }
    set_deform_reference_box(inputrec->init_step,box);
  }

  if (opt2bSet("-cpi",nfile,fnm)) 
  {
      /* Check if checkpoint file exists before doing continuation.
       * This way we can use identical input options for the first and subsequent runs...
       */
      if( fexist(opt2fn("-cpi",nfile,fnm)) )
      {
	  load_checkpoint(opt2fn("-cpi",nfile,fnm),fplog,
			  cr,Flags & MD_PARTDEC,ddxyz,
			  inputrec,state,&bReadRNG,&bReadEkin,
			  (Flags & MD_APPENDFILES));
	  
	  if (bReadRNG)
	  {
	      Flags |= MD_READ_RNG;
	  }
	  if (bReadEkin)
	  {
	      Flags |= MD_READ_EKIN;
	  }
      }
  }
  
  if (MASTER(cr) && (Flags & MD_APPENDFILES))
  {
      fplog = gmx_log_open(ftp2fn(efLOG,nfile,fnm),cr, !(Flags & MD_SEPPOT) ,Flags);
  }
  
  if (SIMMASTER(cr)) 
  {
      copy_mat(state->box,box);
  }
	
  if (PAR(cr)) 
  {
      gmx_bcast(sizeof(box),box,cr);
  }
	
  if (bVerbose && SIMMASTER(cr))
    fprintf(stderr,"Loaded with Money\n\n");

  if (PAR(cr) && !((Flags & MD_PARTDEC) || EI_TPI(inputrec->eI))) {
    cr->dd = init_domain_decomposition(fplog,cr,Flags,ddxyz,rdd,rconstr,
				       dddlb_opt,dlb_scale,
				       ddcsx,ddcsy,ddcsz,
				       mtop,inputrec,
				       box,state->x);
    
    make_dd_communicators(fplog,cr,dd_node_order);

    /* Set overallocation to avoid frequent reallocation of arrays */
    set_over_alloc_dd(TRUE);
  } else {
    cr->duty = (DUTY_PP | DUTY_PME);

    if (inputrec->ePBC == epbcSCREW)
      gmx_fatal(FARGS,"pbc=%s is only implemented with domain decomposition",
		epbc_names[inputrec->ePBC]);
  }

  if (PAR(cr)) {
    /* After possible communicator splitting in make_dd_communicators.
     * we can set up the intra/inter node communication.
     */
    gmx_setup_nodecomm(fplog,cr);
  }
	
  wcycle = wallcycle_init(fplog,cr);

  snew(nrnb,1);
  if (cr->duty & DUTY_PP) {
    /* For domain decomposition we allocate dynamically
     * in dd_partition_system.
     */
    if (DOMAINDECOMP(cr)) {
      bcast_state_setup(cr,state);
    } else {
      if (PAR(cr)) {
	if (!MASTER(cr)) {
	  snew(state,1);
	}
	bcast_state(cr,state,TRUE);
      }

      snew(buf,mtop->natoms);
      snew(f,mtop->natoms);
    }
    
    /* Dihedral Restraints */
    if (gmx_mtop_ftype_count(mtop,F_DIHRES) > 0) {
      init_dihres(fplog,mtop,inputrec,fcd);
    }
    
    /* Initiate forcerecord */
    fr = mk_forcerec();
    init_forcerec(fplog,fr,fcd,inputrec,mtop,cr,box,FALSE,
		  opt2fn("-table",nfile,fnm),opt2fn("-tablep",nfile,fnm),
		  opt2fn("-tableb",nfile,fnm),FALSE,pforce);
    fr->bSepDVDL = ((Flags & MD_SEPPOT) == MD_SEPPOT);
    
    /* Initialize QM-MM */
    if(fr->bQMMM)
      init_QMMMrec(cr,box,mtop,inputrec,fr);
    
    /* Initialize the mdatoms structure.
     * mdatoms is not filled with atom data,
     * as this can not be done now with domain decomposition.
     */
    mdatoms = init_mdatoms(fplog,mtop,inputrec->efep!=efepNO);
    
    /* Initialize the virtual site communication */
    vsite = init_vsite(mtop,cr);

    calc_shifts(box,fr->shift_vec);

    /* With periodic molecules the charge groups should be whole at start up
     * and the virtual sites should not be far from their proper positions.
     */
    if (!inputrec->bContinuation && MASTER(cr) &&
	!(inputrec->ePBC != epbcNONE && inputrec->bPeriodicMols)) {
      /* Make molecules whole at start of run */
      if (fr->ePBC != epbcNONE)  {
	do_pbc_first_mtop(fplog,inputrec->ePBC,box,mtop,state->x);
      }
      if (vsite) {
	/* Correct initial vsite positions are required
	 * for the initial distribution in the domain decomposition
	 * and for the initial shell prediction.
	 */
	construct_vsites_mtop(fplog,vsite,mtop,state->x);
      }
    }
	  
    /* Initiate PPPM if necessary */
    if (fr->eeltype == eelPPPM) {
      if (mdatoms->nChargePerturbed)
	gmx_fatal(FARGS,"Free energy with %s is not implemented",
		  eel_names[fr->eeltype]);
      status = gmx_pppm_init(fplog,cr,FALSE,TRUE,box,
			     getenv("GMXGHAT"),inputrec, (Flags & MD_REPRODUCIBLE));
      if (status != 0)
	gmx_fatal(FARGS,"Error %d initializing PPPM",status);
    }

    if (EEL_PME(fr->eeltype)) {
      ewaldcoeff = fr->ewaldcoeff;
      pmedata = &fr->pmedata;
    } else {
      pmedata = NULL;
    }
  } else {
    /* This is a PME only node */
    
    /* We don't need the state */
    done_state(state);

    ewaldcoeff = calc_ewaldcoeff(inputrec->rcoulomb, inputrec->ewald_rtol);
    snew(pmedata,1);
  }
  
  /*>>> MFFA >>>*/
  /* initialise SDB routines */
  if (MASTER(cr)) {
	  if (opt2bSet("-sb",nfile,fnm)) {
		  fprintf(stderr, "Reading boundary potentials.\n");
		  split_string(opt2fn("-sb",nfile,fnm), &mffa->filenames, 
				  &mffa->meta.PNUMB);
		  fprintf(stderr, "Found:");
		  for (i=0; i < mffa->meta.PNUMB; i++)
			  fprintf(stderr," %s", mffa->filenames[i]);
		  fprintf(stderr, "\n");
		  snew(mffa->boundary, mffa->meta.PNUMB);
		  snew(mffa->dynamic, mffa->meta.PNUMB);
		  int QUENCH=0, COOL=0;
		  if (mffa->meta.PNUMB >= 1) {
			  for (i=0; i < mffa->meta.PNUMB; i++) {
				  read_sdb(mffa->filenames[i], &(mffa->boundary[i]), 
						  &(mffa->boundary[i].meta.TYPE), &QUENCH);
				  COOL += QUENCH;
			  }
		  } else
			  gmx_fatal(FARGS,"No .sdb file found!\n");
		  if (COOL >= 1) {
			  mffa->meta.bQUENCH = TRUE;
			  fprintf(stderr, "Quench routine set to active (for all spherical boundaries)!\n");
		  } else if (COOL == 0)
			  mffa->meta.bQUENCH = FALSE;
		  
		  mffa->meta.bSDB = TRUE;
		  if (PAR(cr)) {
			  MPI_Bcast(&(mffa->meta.bSDB), 1, MPI_INT, MASTERRANK(cr), 
					  cr->mpi_comm_mygroup);
		  }
		  mffa->meta.bNDX = opt2bSet("-nb",nfile,fnm);

		  /* note that SDB->bNDX is used for wildcard check after returning */
		  mffa->meta.bNDX = read_sdb_ndx(mffa->meta.bNDX, 
				  opt2fn("-nb",nfile,fnm), state->natoms, 
				  &(mffa->index), mffa->meta.PNUMB);
		  if (mffa->meta.bNDX)
			  fprintf(stderr, "Found boundary exclusion group in index file for first potential!\n");
		  
		  mffa->meta.size_ndx = state->natoms;
	  }
  } else {
	  if (PAR(cr)) {
		  MPI_Bcast(&(mffa->meta.bSDB), 1, MPI_INT, MASTERRANK(cr), 
				  cr->mpi_comm_mygroup);
	  }
  }
  /* broadcast MFFA boundaries to non-master nodes */
  if (PAR(cr) && mffa->meta.bSDB) {
	  bcast_mffa(cr, mffa);
  }
  if (mffa->meta.bSDB) {
	  if (debug) {
		  echo_mffa_meta(cr, &mffa->meta);
		  for (i=0; i < mffa->meta.PNUMB; i++) {
			  echo_mffa_boundary(cr, &(mffa->boundary[i]));
		  }
	  }
	  /* only for union shell systems in this implementation */
	  if (mffa->boundary[0].meta.pos_out != 0 
			  && mffa->boundary[0].meta.TYPE == efSPHERE) {
		  if (fr->ePBC == epbcNONE)
			  if (inputrec->ref_p[XX][XX] != 0) {
				  if (MASTER(cr))
					  fprintf(stderr, "SDB-pressure coupling switched on.\n");
				  mffa->meta.bPC = TRUE;
			  }
		  if (inputrec->epc != epcNO) {
			  /* NVT with pressure : although its thermodynamicle bullshit */
			  if (inputrec->ref_p[XX][XX] == 0)
				  inputrec->compress[XX][XX] = 0;
			  if (MASTER(cr))
				  fprintf(stderr, "SDB-pressure coupling switched on.\n");
			  mffa->meta.bPC = TRUE;
		  }
	  }
	  /* put normal routine out for sphere */
	  if (mffa->meta.bPC)
		  inputrec->epc=epcNO;
  }
  /*<<< MFFA <<<*/
  /* Initiate PME if necessary,
   * either on all nodes or on dedicated PME nodes only. */
  if (EEL_PME(inputrec->coulombtype)) {
    if (mdatoms) {
      nChargePerturbed = mdatoms->nChargePerturbed;
    }
    if (cr->npmenodes > 0) {
      /* The PME only nodes need to know nChargePerturbed */
      gmx_bcast_sim(sizeof(nChargePerturbed),&nChargePerturbed,cr);
    }
    if (cr->duty & DUTY_PME) {
      status = gmx_pme_init(pmedata,cr,inputrec,
			    mtop ? mtop->natoms : 0,nChargePerturbed,
			    (Flags & MD_REPRODUCIBLE));
      if (status != 0)
	gmx_fatal(FARGS,"Error %d initializing PME",status);
    }
  }

  if (integrator[inputrec->eI].func == do_md) {
    /* Turn on signal handling on all nodes */
    /*
     * (A user signal from the PME nodes (if any)
     * is communicated to the PP nodes.
     */
    if (getenv("GMX_NO_TERM") == NULL) {
      if (debug)
	fprintf(debug,"Installing signal handler for SIGTERM\n");
      signal(SIGTERM,signal_handler);
    }
    if (getenv("GMX_NO_USR1") == NULL) {
      if (debug)
	fprintf(debug,"Installing signal handler for SIGUSR1\n");
      signal(SIGUSR1,signal_handler);
    }
  }

  if (cr->duty & DUTY_PP) {
    if (inputrec->ePull != epullNO) {
      /* Initialize pull code */
      init_pull(fplog,inputrec,nfile,fnm,mtop,cr,
		EI_DYNAMICS(inputrec->eI) && MASTER(cr),Flags);
    }

    constr = init_constraints(fplog,mtop,inputrec,ed,state,cr);

    if (DOMAINDECOMP(cr)) {
      dd_init_bondeds(fplog,cr->dd,mtop,vsite,constr,inputrec,
		      Flags & MD_DDBONDCHECK,fr->cginfo_global);

      set_dd_parameters(fplog,cr->dd,dlb_scale,inputrec,fr,box);
     
      setup_dd_grid(fplog,cr->dd);
    }

	/*>>> MFFA >>>*/
    /* Now do whatever the user wants us to do (how flexible...) */
    start_t = integrator[inputrec->eI].func(fplog,cr,nfile,fnm,
					    bVerbose,bCompact,
					    vsite,constr,
					    nstepout,inputrec,mtop,
					    fcd,state,f,buf,
					    mdatoms,nrnb,wcycle,ed,fr,
					    repl_ex_nst,repl_ex_seed,
					    cpt_period,max_hours,
					    Flags,
					    &nsteps_done, mffa);
	/*<<< MFFA <<<*/

    if (inputrec->ePull != epullNO)
      finish_pull(fplog,inputrec->pull);
  } else {
    /* do PME only */
    gmx_pmeonly(*pmedata,cr,nrnb,wcycle,ewaldcoeff,FALSE);
  }
 
  /* Some timing stats */  
  if (MASTER(cr)) {
    realtime=difftime(time(NULL),start_t);
    if ((nodetime=node_time()) == 0)
      nodetime=realtime;
  }
  else 
    realtime=0;

  wallcycle_stop(wcycle,ewcRUN);
    
  /* Finish up, write some stuff
   * if rerunMD, don't write last frame again 
   */
  finish_run(fplog,cr,ftp2fn(efSTO,nfile,fnm),
	     inputrec,nrnb,wcycle,nodetime,realtime,nsteps_done,
	     EI_DYNAMICS(inputrec->eI) && !MULTISIM(cr));
  
  /* Does what it says */  
  print_date_and_time(fplog,cr->nodeid,"Finished mdrun");

	/* Close logfile already here if we were appending to it */
	if (MASTER(cr) && (Flags & MD_APPENDFILES))
	{
		gmx_log_close(fplog);
	}	
}

time_t do_md(FILE *fplog,t_commrec *cr,int nfile,t_filenm fnm[],
	     bool bVerbose,bool bCompact,
	     gmx_vsite_t *vsite,gmx_constr_t constr,
	     int stepout,t_inputrec *ir,
	     gmx_mtop_t *top_global,
	     t_fcdata *fcd,
	     t_state *state_global,rvec f[],
	     rvec buf[],t_mdatoms *mdatoms,
	     t_nrnb *nrnb,gmx_wallcycle_t wcycle,
	     gmx_edsam_t ed,t_forcerec *fr,
	     int repl_ex_nst,int repl_ex_seed,
	     real cpt_period,real max_hours,
	     unsigned long Flags,
	     int *nsteps_done, MFFA *mffa) /*<<< MFFA >>>*/
{
  int        fp_ene=0,fp_trn=0,fp_xtc=0,step,step_rel,step_ene;
  char       *fn_cpt;
  FILE       *fp_dgdl=NULL,*fp_field=NULL;
  time_t     start_t;
  double     run_time;
  real       t,t0,lam0;
  bool       bGStatEveryStep,bGStat;
  bool       bNS,bSimAnn,bStopCM,bRerunMD,bNotLastFrame=FALSE,
             bFirstStep,bStateFromTPX,bLastStep;
  bool       bNEMD,do_ene,do_log,do_verbose,bRerunWarnNoV=TRUE,
	         bForceUpdate=FALSE,bX,bV,bF,bXTC,bCPT;
  bool       bMasterState;
  tensor     force_vir,shake_vir,total_vir,pres,ekin;
  int        i,m,status;
  rvec       mu_tot;
  t_vcm      *vcm;
  int        step_ns=0,step_nscheck=0,nns=0,nabnsb=0,ns_lt;
  double     ns_s1=0,ns_s2=0,ns_ab=0,ns_lt_runav=0,ns_lt_runav2=0;
  matrix     *scale_tot;
  t_trxframe rerun_fr;
  gmx_repl_ex_t repl_ex=NULL;
  int        nchkpt=1;
  /* Booleans (disguised as a reals) to checkpoint and terminate mdrun */  
  real       chkpt=0,terminate=0,terminate_now=0;

  gmx_localtop_t *top;	
  t_mdebin *mdebin=NULL;
  t_state    *state=NULL;
  rvec       *f_global=NULL;
  gmx_enerdata_t *enerd;
  gmx_stochd_t sd=NULL;
  t_graph    *graph=NULL;

  bool        bFFscan;
  gmx_groups_t *groups;
  gmx_ekindata_t *ekind;
  gmx_shellfc_t shellfc;
  int         count,nconverged=0;
  real        timestep=0;
  double      tcount=0;
  bool        bHaveConstr=FALSE,bIonize=FALSE,bGlas=FALSE;
  bool        bTCR=FALSE,bConverged=TRUE,bOK,bSumEkinhOld,bExchanged;
  bool        bAppend;
  real        temp0,mu_aver=0,dvdl;
  int         a0,a1,gnx=0,ii;
  atom_id     *grpindex=NULL;
  char        *grpname;
  t_coupl_rec *tcr=NULL;
  rvec        *xcopy=NULL,*vcopy=NULL, center;
  matrix      boxcopy,lastbox;
  double      cycles;
  /*>>> MFFA >>>*/
  /* partial pressures */
  rvec pressure_plane[2];
  real pressure_sphere[2];
  /* some running parameters for SDB hacks */
  real P_out_tot=0, P_out_interval=0;
  real P_in_tot=0, P_in_interval=0;
  real P_vir_interval=0, P_vir_tot =0;
  int count1=0, count2=0, n, j;
  real shift_out, shift_in, pos_out_ref=0, pos_in_ref=0;
  real volume=0, rin, rout, AREA=0;
  bool bUPDSDB;

  bUPDSDB = (ir->compress[XX][XX] == 0 && mffa->meta.bSDB);
  /*<<< MFFA <<<*/

  /* Check for special mdrun options */
  bRerunMD = (Flags & MD_RERUN);
  bIonize  = (Flags & MD_IONIZE);
  bGlas    = (Flags & MD_GLAS);
  bFFscan  = (Flags & MD_FFSCAN);
  bGStatEveryStep = !(Flags & MD_NOGSTAT);
  bAppend  = (Flags & MD_APPENDFILES);
	
  if (!bGStatEveryStep && !EI_DYNAMICS(ir->eI)) {
    char *warn="\nWARNING:\nNo energy summing can only be used with dynamics, ignoring this option\n";
    fprintf(stderr,"%s\n",warn);
    if (fplog)
      fprintf(fplog,"%s\n",warn);
    bGStatEveryStep = TRUE;
  }
  if (!bGStatEveryStep) {
    if (fplog) {
      fprintf(fplog,"\nWill not sum the energies at every step,\n"
	      "therefore the energy file does not contain exact averages and fluctuations.\n\n");
      if (ir->etc != etcNO || ir->epc != epcNO) {
	fprintf(fplog,"WARNING:\nThe temperature and/or pressure for scaling will only be updated every nstlist (%d) steps\n\n",ir->nstlist);
      }
      if (ir->nstlist == -1) {
	fprintf(fplog,
		"To reduce the energy communication with nstlist = -1\n"
		"the neighbor list validity should not be checked at every step,\n"
		"this means that exact integration is not guaranteed.\n"
		"The neighbor list validity is checked after:\n"
		"  <n.list life time> - 2*std.dev.(n.list life time)  steps.\n"
		"In most cases this will result in exact integration.\n"
		"This reduces the energy communication by a factor of 2 to 3.\n"
		"If you want less energy communication, set nstlist > 3.\n\n");
      }
    }
    if (ir->comm_mode != ecmNO && ir->nstcomm == 1) {
      if (fplog) {
	fprintf(fplog,"WARNING:\nWe should not remove the COM motion every step with option -nosum,\n");
      }
      if (ir->nstlist > 0) {
	ir->nstcomm = ir->nstlist;
	if (fplog) {
	  fprintf(fplog,"setting nstcomm to nstlist (%d)\n\n",ir->nstcomm);
	}
      } else {
	ir->nstcomm = 100;
	if (fplog) {
	  fprintf(fplog,"setting nstcomm to %d\n\n",ir->nstcomm);
	}
      }
    }
  }

  if (bRerunMD || bFFscan)
    ir->nstxtcout = 0;

  /*>>> MFFA >>>*/
  /* Make sure rotational momentum is removed !!!!!!!!!!!!*/
  if(mffa->meta.bSDB)
	  if(mffa->boundary[0].meta.TYPE == efSPHERE) {
		  ir->comm_mode = ecmANGULAR;
		  ir->nstcomm = 1;
		  if (MASTER(cr))
			  fprintf(stderr, "Simulating isolated cluster in vacuum. Angular momentum will be removed at every step!\n");
	  }
  /*<<< MFFA <<<*/

  groups = &top_global->groups;

  /* Initial values */
  init_md(fplog,cr,ir,&t,&t0,&state_global->lambda,&lam0,
	  nrnb,top_global,&sd,
	  nfile,fnm,&fp_trn,&fp_xtc,&fp_ene,&fn_cpt,
	  &fp_dgdl,&fp_field,&mdebin,
	  force_vir,shake_vir,mu_tot,&bNEMD,&bSimAnn,&vcm,Flags);

  /* Energy terms and groups */
  snew(enerd,1);
  init_enerdata(fplog,top_global->groups.grps[egcENER].nr,enerd);
  /* Kinetic energy data */
  snew(ekind,1);
  init_ekindata(fplog,top_global,&(ir->opts),ekind);
  /* Copy the cos acceleration to the groups struct */
  ekind->cosacc.cos_accel = ir->cos_accel;

  debug_gmx();

  /* Check for polarizable models and flexible constraints */
  shellfc = init_shell_flexcon(fplog,
			       top_global,n_flexible_constraints(constr),
			       (ir->bContinuation || 
				(DOMAINDECOMP(cr) && !MASTER(cr))) ?
			       NULL : state_global->x);

  {
    double io = compute_io(ir,top_global->natoms,groups,mdebin->ebin->nener,1);
    if ((io > 2000) && MASTER(cr))
      fprintf(stderr,
	      "\nWARNING: This run will generate roughly %.0f Mb of data\n\n",
	      io);
  }
 
  if (DOMAINDECOMP(cr)) {
    top = dd_init_local_top(top_global);

    snew(state,1);
    dd_init_local_state(cr->dd,state_global,state);

    if (DDMASTER(cr->dd) && ir->nstfout) {
      snew(f_global,state_global->natoms);
    }
  } else {
    if (PAR(cr)) {
      /* Initialize the particle decomposition and split the topology */
      top = split_system(fplog,top_global,ir,cr);

      pd_cg_range(cr,&fr->cg0,&fr->hcg);
      pd_at_range(cr,&a0,&a1);
    } else {
      top = gmx_mtop_generate_local_top(top_global,ir);

      a0 = 0;
      a1 = top_global->natoms;
    }

    state = partdec_init_local_state(cr,state_global);
    f_global = f;

    atoms2md(top_global,ir,0,NULL,a0,a1-a0,mdatoms);

    if (vsite) {
      set_vsite_top(vsite,top,mdatoms,cr);
    }

    if (ir->ePBC != epbcNONE && !ir->bPeriodicMols) {
      graph = mk_graph(fplog,&(top->idef),0,top_global->natoms,FALSE,FALSE);
    }

    if (shellfc) {
      make_local_shells(cr,mdatoms,shellfc);
    }

    if (ir->pull && PAR(cr)) {
      dd_make_local_pull_groups(NULL,ir->pull,mdatoms);
    }
  }

  if (DOMAINDECOMP(cr)) {
    /* Distribute the charge groups over the nodes from the master node */
    dd_partition_system(fplog,ir->init_step,cr,TRUE,
			state_global,top_global,ir,
			state,&f,&buf,mdatoms,top,fr,vsite,shellfc,constr,
			nrnb,wcycle,FALSE);
  }

  update_mdatoms(mdatoms,state->lambda);

  if (MASTER(cr))
  {
    /* Update mdebin with energy history if appending to output files */
    if ( Flags & MD_APPENDFILES )
    {
      restore_energyhistory_from_state(mdebin,&state_global->enerhist);
    }
    /* Set the initial energy history in state to zero by updating once */
    update_energyhistory(&state_global->enerhist,mdebin);
  }	
	
	
  if (sd && (Flags & MD_READ_RNG)) {
    /* Set the random state if we read a checkpoint file */
    set_stochd_state(sd,state);
  }

  /* Initialize constraints */
  if (constr) {
    if (!DOMAINDECOMP(cr))
      set_constraints(constr,top,ir,mdatoms,NULL);
    bHaveConstr = TRUE;
  }

  /* Check whether we have to GCT stuff */
  bTCR = ftp2bSet(efGCT,nfile,fnm);
  if (bTCR) {
    if (MASTER(cr)) {
      fprintf(stderr,"Will do General Coupling Theory!\n");
    }
    gnx = top_global->mols.nr;
    snew(grpindex,gnx);
    for(i=0; (i<gnx); i++) {
      grpindex[i] = i;
    }
  }
  

  if (repl_ex_nst > 0 && MASTER(cr)) {
    repl_ex = init_replica_exchange(fplog,cr->ms,state_global,ir,
				    repl_ex_nst,repl_ex_seed);

    if (max_hours > 0 && MASTERSIM(cr->ms)) {
      fprintf(stderr,"\nWARNING: Using -maxh with REMD can cause the simulation to hang after maxh hours\n\n");
    }
  }

  if (!ir->bContinuation && !bRerunMD) {
    if (mdatoms->cFREEZE && (state->flags & (1<<estV))) {
      /* Set the velocities of frozen particles to zero */
      for(i=mdatoms->start; i<mdatoms->homenr; i++) {
	for(m=0; m<DIM; m++) {
	  if (ir->opts.nFreeze[mdatoms->cFREEZE[i]][m]) {
	    state->v[i][m] = 0;
	  }
	}
      }
    }
    if (bHaveConstr) {
      /* Constrain the initial coordinates and velocities */
      do_shakefirst(fplog,constr,ir,mdatoms,state,buf,f,
		    graph,cr,nrnb,fr,&top->idef);
    }
    if (vsite) {
      /* Construct the virtual sites for the initial configuration */
      construct_vsites(fplog,vsite,state->x,nrnb,ir->delta_t,NULL,
		       top->idef.iparams,top->idef.il,
		       fr->ePBC,fr->bMolPBC,graph,cr,state->box);
    }
  }

  debug_gmx();

  if (Flags & MD_READ_EKIN)
  {
      restore_ekinstate_from_state(cr,ekind,&state_global->ekinstate);
  }
  else
  {
      /* Compute initial EKin for all.. */
      if (ekind->cosacc.cos_accel == 0) {
          calc_ke_part(state->v,&(ir->opts),mdatoms,ekind,nrnb,state->lambda);
      } else {
	  calc_ke_part_visc(state->box,state->x,state->v,&(ir->opts),
			    mdatoms,ekind,nrnb,state->lambda);
      }
      debug_gmx();

      if (PAR(cr))
      {
	  GMX_MPE_LOG(ev_global_stat_start);
	  
	  global_stat(fplog,cr,enerd,force_vir,shake_vir,mu_tot,
		      ir,ekind,FALSE,constr,vcm,NULL,NULL,&terminate);
	  
	  GMX_MPE_LOG(ev_global_stat_finish);
      }
       debug_gmx();
  }
  
  /* Calculate the initial half step temperature */
  temp0 = sum_ekin(TRUE,&(ir->opts),ekind,ekin,NULL);

  debug_gmx();
   
  /* Initiate data for the special cases */
  if (bFFscan) {
    snew(xcopy,state->natoms);
    snew(vcopy,state->natoms);
    for(ii=0; (ii<state->natoms); ii++) {
      copy_rvec(state->x[ii],xcopy[ii]);
      copy_rvec(state->v[ii],vcopy[ii]);
    }
    copy_mat(state->box,boxcopy);
  } 

  if (MASTER(cr)) {
    if (constr && !ir->bContinuation && ir->eConstrAlg == econtLINCS)
      fprintf(fplog,
	      "RMS relative constraint deviation after constraining: %.2e\n",
	      constr_rmsd(constr,FALSE));
    fprintf(fplog,"Initial temperature: %g K\n",temp0);
    if (bRerunMD) {
      fprintf(stderr,"starting md rerun '%s', reading coordinates from"
	      " input trajectory '%s'\n\n",
	      *(top_global->name),opt2fn("-rerun",nfile,fnm));
      if (bVerbose)
	fprintf(stderr,"Calculated time to finish depends on nsteps from "
		"run input file,\nwhich may not correspond to the time "
		"needed to process input trajectory.\n\n");
    } else {
		if(ir->init_step>0)
		{
			fprintf(stderr,"starting mdrun '%s'\n%d steps, %8.1f ps (continuing from step %d, %8.1f ps).\n",
					*(top_global->name),ir->nsteps+ir->init_step,(ir->nsteps+ir->init_step)*ir->delta_t,
					ir->init_step,ir->init_step*ir->delta_t);
		}
		else
		{
			fprintf(stderr,"starting mdrun '%s'\n%d steps, %8.1f ps.\n",
					*(top_global->name),ir->nsteps,ir->nsteps*ir->delta_t);
		}
    }
    fprintf(fplog,"\n");
  }

  if (ir->nstlist == -1) {
    snew(scale_tot,1);
  } else {
    scale_tot = NULL;
  }

  /* Write start time */
  start_t=print_date_and_time(fplog,cr->nodeid,"Started mdrun");
  wallcycle_start(wcycle,ewcRUN);
  if (fplog)
    fprintf(fplog,"\n");

  /* Set the node time counter to 0 after initialisation */
  start_time();
  debug_gmx();
  /***********************************************************
   *
   *             Loop over MD steps 
   *
   ************************************************************/
  
  /* if rerunMD then read coordinates and velocities from input trajectory */
  if (bRerunMD) {
    if (getenv("GMX_FORCE_UPDATE"))
      bForceUpdate = TRUE;

    bNotLastFrame = read_first_frame(&status,opt2fn("-rerun",nfile,fnm),
				     &rerun_fr,TRX_NEED_X | TRX_READ_V);
    if (rerun_fr.natoms != top_global->natoms)
      gmx_fatal(FARGS,"Number of atoms in trajectory (%d) does not match the "
		"run input file (%d)\n",rerun_fr.natoms,top_global->natoms);
    if (ir->ePBC != epbcNONE) {
      if (!rerun_fr.bBox)
	gmx_fatal(FARGS,"Rerun trajectory frame step %d time %f does not contain a box, while pbc is used",rerun_fr.step,rerun_fr.time);
      if (max_cutoff2(ir->ePBC,rerun_fr.box) < sqr(fr->rlistlong))
	gmx_fatal(FARGS,"Rerun trajectory frame step %d time %f has too small box dimensions",rerun_fr.step,rerun_fr.time);

      /* Set the shift vectors.
       * Necessary here when have a static box different from the tpr box.
       */
      calc_shifts(rerun_fr.box,fr->shift_vec);
    }
  }

  /* loop over MD steps or if rerunMD to end of input trajectory */
  bFirstStep = TRUE;
  /* Skip the first Nose-Hoover integration when we get the state from tpx */
  bStateFromTPX = !opt2bSet("-cpi",nfile,fnm);
  bLastStep = FALSE;
  bSumEkinhOld = FALSE,
  bExchanged = FALSE;

  step = ir->init_step;
  step_rel = 0;
  step_ene = 0;

  bLastStep = (bRerunMD || step_rel > ir->nsteps);
  while (!bLastStep || (bRerunMD && bNotLastFrame)) {

    wallcycle_start(wcycle,ewcSTEP);

    GMX_MPE_LOG(ev_timestep1);

    if (bRerunMD) {
      if (rerun_fr.bStep) {
	step = rerun_fr.step;
	step_rel = step - ir->init_step;
      }
      if (rerun_fr.bTime)
	t = rerun_fr.time;
      else
	t = step;
    } else {
      bLastStep = (step_rel == ir->nsteps);

      t = t0 + step*ir->delta_t;
    }
    if (Flags & MD_APPENDFILES) {
      step_ene = step;
    } else {
      step_ene = step_rel;
    }

    if (ir->efep != efepNO) {
      if (bRerunMD && rerun_fr.bLambda && (ir->delta_lambda!=0))
	state_global->lambda = rerun_fr.lambda;
      else
	state_global->lambda = lam0 + step*ir->delta_lambda;
      state->lambda = state_global->lambda;
    }
    
    if (bSimAnn) 
      update_annealing_target_temp(&(ir->opts),t);
    
    if (bRerunMD) {
      if (!(DOMAINDECOMP(cr) && !MASTER(cr))) {
	for(i=0; i<state_global->natoms; i++) {
	  copy_rvec(rerun_fr.x[i],state_global->x[i]);
	}
	if (rerun_fr.bV) {
	  for(i=0; i<state_global->natoms; i++) {
	    copy_rvec(rerun_fr.v[i],state_global->v[i]);
	  }
	} else {
	  for(i=0; i<state_global->natoms; i++) {
	    clear_rvec(state_global->v[i]);
	  }
	  if (bRerunWarnNoV) {
	    fprintf(stderr,"\nWARNING: Some frames do not contain velocities.\n"
		    "         Ekin, temperature and pressure are incorrect,\n"
		    "         the virial will be incorrect when constraints are present.\n"
		    "\n");
	    bRerunWarnNoV = FALSE;
	  }
	}
      }
      copy_mat(rerun_fr.box,state_global->box);
      copy_mat(state_global->box,state->box);

      if (vsite && (Flags & MD_RERUN_VSITE)) {
	if (DOMAINDECOMP(cr)) {
	  gmx_fatal(FARGS,"Vsite recalculation with -rerun is not implemented for domain decomposition, use particle decomposition");
	}
	if (graph) {
	  /* Following is necessary because the graph may get out of sync
	   * with the coordinates if we only have every N'th coordinate set
	   */
	  mk_mshift(fplog,graph,fr->ePBC,state->box,state->x);
	  shift_self(graph,state->box,state->x);
	}
	construct_vsites(fplog,vsite,state->x,nrnb,ir->delta_t,state->v,
			 top->idef.iparams,top->idef.il,
			 fr->ePBC,fr->bMolPBC,graph,cr,state->box);
	if (graph)
	  unshift_self(graph,state->box,state->x);
      }
    }

    /* Stop Center of Mass motion */
    bStopCM = (ir->comm_mode != ecmNO && do_per_step(step,ir->nstcomm));

    /* Copy back starting coordinates in case we're doing a forcefield scan */
    if (bFFscan) {
      for(ii=0; (ii<state->natoms); ii++) {
	copy_rvec(xcopy[ii],state->x[ii]);
	copy_rvec(vcopy[ii],state->v[ii]);
      }
      copy_mat(boxcopy,state->box);
    }
    
    bNS = bFirstStep;
    if (bRerunMD) {
      /* for rerun MD always do Neighbour Searching */
      if (ir->nstlist != 0) {
	bNS = TRUE;
      }
    } else {
      /* Determine whether or not to do Neighbour Searching */
      if (bExchanged || (ir->nstlist > 0 && (step % ir->nstlist == 0))) {
	bNS = TRUE;
      } else if (ir->nstlist == -1) {
	bNS = (bFirstStep || nabnsb > 0);
	if (bNS) {
	  if (bFirstStep) {
	    ns_lt_runav = 0;
	    step_nscheck = step;
	  } else {
	    /* Determine the neighbor list life time */
	    ns_lt = step - step_ns;
	    if (debug) {
	      fprintf(debug,"%d atoms beyond ns buffer, updating neighbor list after %d steps\n",nabnsb,ns_lt);
	    }
	    nns++;
	    ns_s1 += ns_lt;
	    ns_s2 += ns_lt*ns_lt;
	    ns_ab += nabnsb;
	    if (ns_lt_runav == 0) {
	      ns_lt_runav  = ns_lt;
	      /* Initialize the fluctuation average such that at startup
	       * we check after 0 steps.
	       */
	      ns_lt_runav2 = sqr(ns_lt/2.0);
	    }
	    /* Running average with 0.9 gives an exp. history of 9.5 */
	    ns_lt_runav2 = 0.9*ns_lt_runav2 + 0.1*sqr(ns_lt_runav - ns_lt);
	    ns_lt_runav  = 0.9*ns_lt_runav  + 0.1*ns_lt;
	    if (bGStatEveryStep) {
	      /* Always check the nlist validity */
	      step_nscheck = step;
	    } else {
	      /* We check after:  <life time> - 2*sigma
	       * The factor 2 is quite conservative,
	       * but we assume that with nstlist=-1 the user prefers
	       * exact integration over performance.
	       */
	      step_nscheck = step
		+ (int)(ns_lt_runav - 2.0*sqrt(ns_lt_runav2)) - 1;
	    }
	    if (debug) {
	      fprintf(debug,"nlist life time %d run av. %4.1f sig %3.1f check %d check with -nosum %d\n",
		      ns_lt,ns_lt_runav,sqrt(ns_lt_runav2),
		      step_nscheck-step+1,
		      (int)(ns_lt_runav - 2.0*sqrt(ns_lt_runav2)));
	    }
	  }
	  step_ns = step;
	  /* Initialize the cumulative coordinate scaling matrix */
	  clear_mat(*scale_tot);
	  for(ii=0; ii<DIM; ii++)
	    (*scale_tot)[ii][ii] = 1.0;
	}
      }
    } 

    if (terminate_now > 0 || (terminate_now < 0 && bNS)) {
      bLastStep = TRUE;
    }

    do_log = do_per_step(step,ir->nstlog) || bFirstStep || bLastStep;
    do_verbose = bVerbose && (step % stepout == 0 || bFirstStep || bLastStep);

    if (bNS && !(bFirstStep && ir->bContinuation && !bRerunMD)) {
      if (bRerunMD) {
	bMasterState = TRUE;
      } else {
	bMasterState = FALSE;
	/* Correct the new box if it is too skewed */
	if (DYNAMIC_BOX(*ir)) {
	  if (correct_box(fplog,step,state->box,graph))
	    bMasterState = TRUE;
	}
	if (DOMAINDECOMP(cr) && bMasterState)
	  dd_collect_state(cr->dd,state,state_global);
      }
      
      if (DOMAINDECOMP(cr)) {
	/* Repartition the domain decomposition */
	wallcycle_start(wcycle,ewcDOMDEC);
	dd_partition_system(fplog,step,cr,bMasterState,
			    state_global,top_global,ir,
			    state,&f,&buf,mdatoms,top,fr,vsite,shellfc,constr,
			    nrnb,wcycle,do_verbose);
	wallcycle_stop(wcycle,ewcDOMDEC);
      }
    }

    if (MASTER(cr) && do_log && !bFFscan)
      print_ebin_header(fplog,step,t,state->lambda);

    if (ir->efep != efepNO)
      update_mdatoms(mdatoms,state->lambda); 

    if (bRerunMD && rerun_fr.bV) {
      /* We need the kinetic energy at minus the half step for determining
       * the full step kinetic energy and possibly for T-coupling.
       */
      calc_ke_part(state->v,&(ir->opts),mdatoms,ekind,nrnb,state->lambda);
      if (PAR(cr)) {
	global_stat(fplog,cr,enerd,force_vir,shake_vir,mu_tot,
		    ir,ekind,FALSE,constr,vcm,NULL,NULL,&terminate);
      }
      sum_ekin(FALSE,&(ir->opts),ekind,ekin,NULL);
    }
    
    clear_mat(force_vir);
    
    /* Ionize the atoms if necessary */
    if (bIonize) {
      ionize(fplog,mdatoms,top_global,t,ir,state->x,state->v,
	     mdatoms->start,mdatoms->start+mdatoms->homenr,state->box,cr);
    }
      
    /* Update force field in ffscan program */
    if (bFFscan) {
      if (update_forcefield(fplog,
			    nfile,fnm,fr,mdatoms->nr,state->x,state->box)) {
	if (gmx_parallel_env)
	  gmx_finalize(cr);
	exit(0);
      }
    }

    GMX_MPE_LOG(ev_timestep2);

    if (shellfc) {
      /* Now is the time to relax the shells */
      count=relax_shell_flexcon(fplog,cr,bVerbose,bFFscan ? step+1 : step,
				ir,bNS,bStopCM,top,constr,enerd,fcd,
				state,f,buf,force_vir,mdatoms,
				nrnb,wcycle,graph,groups,
				shellfc,fr,t,mu_tot,
				state->natoms,&bConverged,vsite,
				fp_field);
      tcount+=count;
      
      if (bConverged)
	nconverged++;
    }
    else {
      /* The coordinates (x) are shifted (to get whole molecules) in do_force
       * This is parallellized as well, and does communication too. 
       * Check comments in sim_util.c
       */
      do_force(fplog,cr,ir,step,nrnb,wcycle,top,groups,
	       state->box,state->x,&state->hist,
	       f,buf,force_vir,mdatoms,enerd,fcd,
	       state->lambda,graph,
	       fr,vsite,mu_tot,t,fp_field,ed,
	       GMX_FORCE_STATECHANGED | (bNS ? GMX_FORCE_NS : 0) |
	       GMX_FORCE_ALLFORCES);
    }

    GMX_BARRIER(cr->mpi_comm_mygroup);

    if (bTCR)
      mu_aver = calc_mu_aver(cr,state->x,mdatoms->chargeA,
			     mu_tot,&top_global->mols,mdatoms,gnx,grpindex);
    if (bGlas)
      do_glas(fplog,mdatoms->start,mdatoms->homenr,state->x,f,
	      fr,mdatoms,top->idef.atnr,ir,enerd->term);
    
    if (bTCR && bFirstStep) {
      tcr=init_coupling(fplog,nfile,fnm,cr,fr,mdatoms,&(top->idef));
      fprintf(fplog,"Done init_coupling\n"); 
      fflush(fplog);
    }

    /* Now we have the energies and forces corresponding to the 
     * coordinates at time t. We must output all of this before
     * the update.
     * for RerunMD t is read from input trajectory
     */

    GMX_MPE_LOG(ev_output_start);

    bX   = do_per_step(step,ir->nstxout);
    bV   = do_per_step(step,ir->nstvout);
    bF   = do_per_step(step,ir->nstfout);
    bXTC = do_per_step(step,ir->nstxtcout);
    if ((bNS || bLastStep) && (step > ir->init_step) && !bRerunMD) {
      bCPT = ((chkpt < 0 && do_per_step(step,ir->nstenergy)) || chkpt > 0 ||
	       bLastStep);
      if (bCPT) {
	chkpt = 0;
      }
    } else {
      bCPT = FALSE;
    }

    if (bX || bV || bF || bXTC || bCPT) {
      wallcycle_start(wcycle,ewcTRAJ);
      if (bCPT) {
	if (sd) {
	  get_stochd_state(sd,state);
	}
	if (MASTER(cr)) {
	  if (bSumEkinhOld) {
	    state_global->ekinstate.bUpToDate = FALSE;
	  } else {
	    update_ekinstate(&state_global->ekinstate,ekind);
	    state_global->ekinstate.bUpToDate = TRUE;
	  }
	  update_energyhistory(&state_global->enerhist,mdebin);
	}
      }
      write_traj(fplog,cr,fp_trn,bX,bV,bF,fp_xtc,bXTC,ir->xtcprec,fn_cpt,bCPT,
				 top_global,ir->eI,ir->simulation_part,step,t,state,state_global,f,f_global);
      debug_gmx();

      if (bLastStep && step_rel == ir->nsteps &&
	  (Flags & MD_CONFOUT) && MASTER(cr) &&
	  !bRerunMD && !bFFscan) {
	/* x and v have been collected in write_traj */
	fprintf(stderr,"\nWriting final coordinates.\n");
	if (ir->ePBC != epbcNONE && !ir->bPeriodicMols && DOMAINDECOMP(cr)) {
	  /* Make molecules whole only for confout writing */
	  do_pbc_mtop(fplog,ir->ePBC,state->box,top_global,state_global->x);
	}
	write_sto_conf_mtop(ftp2fn(efSTO,nfile,fnm),
			    *top_global->name,top_global,
			    state_global->x,state_global->v,
			    ir->ePBC,state->box);
	debug_gmx();
      }
      wallcycle_stop(wcycle,ewcTRAJ);
    }
    GMX_MPE_LOG(ev_output_finish);

    clear_mat(shake_vir);
	/*>>> MFFA >>>*/
	/* position of the boundary force after virial calculation and before 
	 * update. One loop extra but more clear */
	if (mffa->meta.bSDB) {
		/* include boundary force */
		do_sdb_force(mdatoms->start, mdatoms->homenr, f, state->x, 
				mffa, pressure_plane, pressure_sphere, state->box, force_vir);
		/* sum all force and pressure contributions over the nodes */
		if (PAR(cr)) {
			/* sum up partial pressures & forces */
			collect_mffa(mffa->dynamic, pressure_sphere, pressure_plane, 
					mffa->meta.PNUMB, cr);
		}
		if (debug) {
			echo_pressures(cr, pressure_plane, pressure_sphere);
			echo_mffa_dynamic(cr, mffa->dynamic, mffa->meta.PNUMB);
		}
	}
	/*<<< MFFA <<<*/
    if (bFFscan)
      clear_rvecs(state->natoms,buf);

    /* Box is changed in update() when we do pressure coupling,
     * but we should still use the old box for energy corrections and when
     * writing it to the energy file, so it matches the trajectory files for
     * the same timestep above. Make a copy in a separate array.
     */
    copy_mat(state->box,lastbox);
 
    
    GMX_MPE_LOG(ev_update_start);
    /* This is also parallellized, but check code in update.c */
    /* bOK = update(nsb->natoms,START(nsb),HOMENR(nsb),step,state->lambda,&ener[F_DVDL], */
    bOK = TRUE;
    if (!bRerunMD || rerun_fr.bV || bForceUpdate) {
      wallcycle_start(wcycle,ewcUPDATE);
      dvdl = 0;
      /* We can only do Berendsen coupling after we have summed the kinetic
       * energy or virial. Since the happens in global_state after update,
       * we should only do it at step % nstlist = 1 with bGStatEveryStep=FALSE.
       */
      update(fplog,step,&dvdl,ir,mdatoms,state,graph,f,buf,fcd,
	     &top->idef,ekind,shake_vir,scale_tot,
	     cr,nrnb,wcycle,sd,constr,bHaveConstr,
	     bNEMD,bFirstStep && bStateFromTPX);
      if (fr->bSepDVDL && fplog && do_log) {
	fprintf(fplog,sepdvdlformat,"Constraint",0.0,dvdl);
      }
      enerd->term[F_DGDL_CON] += dvdl;
      wallcycle_stop(wcycle,ewcUPDATE);
    } else if (graph) {
      /* Need to unshift here */
      unshift_self(graph,state->box,state->x);
    }

    GMX_BARRIER(cr->mpi_comm_mygroup);
    GMX_MPE_LOG(ev_update_finish);

    if (!bOK && !bFFscan)
      gmx_fatal(FARGS,"Constraint error: Shake, Lincs or Settle could not solve the constrains");

    if (vsite) {
      wallcycle_start(wcycle,ewcVSITECONSTR);
      if (graph)
	shift_self(graph,state->box,state->x);
      
      construct_vsites(fplog,vsite,state->x,nrnb,ir->delta_t,state->v,
		       top->idef.iparams,top->idef.il,
		       fr->ePBC,fr->bMolPBC,graph,cr,state->box);
     
      if (graph)
	unshift_self(graph,state->box,state->x);
      wallcycle_stop(wcycle,ewcVSITECONSTR);
    }

    /* Non-equilibrium MD: this is parallellized, but only does communication
     * when there really is NEMD.
     */
    if (PAR(cr) && bNEMD) 
      accumulate_u(cr,&(ir->opts),ekind);

	/*>>> MFFA >>>*/
	/*******************************************/
	/*************BOUNDARY PRESSURE*************/
	/*******************************************/
	/* note that continuation isn't exact */
	/* only for enclosed spherical systems, not for planes */
	
	/* volume calculation */
	rin = 0;
	rout = 0;
	volume = calc_volume(&mffa->meta, &mffa->boundary[0].meta, lastbox, 
			&rin, &rout);
	if (debug)
		fprintf(debug, "volume=%f\n", volume);
	
	/* PC */
	if (mffa->meta.bPC && ! bFirstStep) {
		/* F/A */
		copy_rvec(mffa->boundary[0].meta.center, center);
		mffa->meta.P_int = trace(pres) / DIM;
		/* coupling to outer virial pressure */
		if (ir->compress[ZZ][ZZ]!=0) {
			/* rout,rin correspond to PNUMB = 0 */
			/* Note really efficient can better use the Mu in update.c. If 
			 * I have time.*/ 
			/* implement for cavity growing within the vesicle */
			if (mffa->meta.bQUENCH) {
				/* piston pressure coupling: no scaling */
				/* userint != 0, quench routine with both of the boundaries 
				 * scaling. This can be used for quenching big vesicles to 
				 * equ pressure */
				
				/* Now the cavity pressure is the virial pressure */
				if (ir->userint1 == 2)
					pressure_sphere[0] = mffa->meta.P_int;
				sphere_piston(state->x, mdatoms->start, mdatoms->homenr, 
						mffa->dynamic[0].AREA_out, pressure_sphere[1], 
						mffa->dynamic[0].AREA_in, pressure_sphere[0], 
						ir->ref_p[XX][XX], &shift_out, &shift_in, 
						ir->delta_t, ir->userreal2, ir->userreal3, 
						ir->userreal4, ir->opts.ref_t[0]);
				/* growth inner cavity with constant rate */
				if (ir->userint1 == 0) 
					shift_in = ir->userreal1 * ir->delta_t;
				
				/* set new rin en out */
				if (rin != 0)
					rin  += shift_in;
				if (rout != 0)
					rout += shift_out;
			} else
				/* mu scaling */
				sdb_pc(mdatoms->start, mdatoms->homenr, &rout, &rin, 
						ir->ref_p[XX][XX], mffa->meta.P_int, 
						ir->delta_t, ir->tau_p, &shift_out, 
						&shift_in, ir->compress[XX][XX], state->x, 
						center);
			
			/* carefull!!!! */
			/* n=0 gives the reference shell the rest is shifted along 
			 * with it*/
			for (n=0; n < mffa->meta.PNUMB; n++) {
				/* This approach is used to handle the shifting of 
				 * restraining shells within the system */
				/* scale the shell together with only the outher or inner 
				 * shell if the shell is to thin */
				if (mffa->boundary[n].meta.TYPE == efSPHERE) {
					correct_shell(&(mffa->boundary[0].meta), 
							(&mffa->boundary[n].meta), &shift_in, &shift_out);
					if (mffa->boundary[n].meta.pos_out != 0)
						mffa->boundary[n].meta.pos_out += shift_out;
					if (mffa->boundary[n].meta.pos_in !=0)
						mffa->boundary[n].meta.pos_in += shift_in;
				}
			}
		}
		P_out_interval += pressure_sphere[1];
		P_in_interval  += pressure_sphere[0];
		P_out_tot += pressure_sphere[1];
		P_in_tot  += pressure_sphere[0];
		P_vir_interval += mffa->meta.P_int;
		P_vir_tot += mffa->meta.P_int;
		count1++;
		count2++;
		
		if (do_per_step(step,ir->nstorireout)) {
			if (MASTER(cr)) {
				fprintf(stderr, "time: %f (ps)\n", t);
				fprintf(stderr, "bound pres in: %f pres out: %f \n", 
						pressure_sphere[0], pressure_sphere[1]);
				fprintf(stderr, "bound pres (interval) in : %f out: %f\n", 
						(float)(P_in_interval/count1), 
						(float)(P_out_interval/count1));
				fprintf(stderr, "pres gem (total) in: %f  out: %f\n", 
						(float)(P_in_tot/(count2)), 
						(float)(P_out_tot/(count2)));
				/* reset interval pressure */
				fprintf(stderr, "pres vir (interval) : %f", 
						(float)(P_vir_interval/count1));
				fprintf(stderr, "   pres vir (total) : %f\n", 
						(float)(P_vir_tot/count2));
				fprintf(stderr, "inner radius: %f outer: %f\n", rin, rout);
				fprintf(stderr, "Volume: %f (nm^3) \n", volume);
				fprintf(stderr, "Density: %f (kg/m^3)\n", 
						(float)(((1.66055*mdatoms->tmass)/volume)));
				if (debug)
					for (n=0; n < mffa->meta.PNUMB; n++) {
						fprintf(debug, "center %d: (%f,%f,%f)\n", n,
								mffa->boundary[n].meta.center[XX], 
								mffa->boundary[n].meta.center[YY],
								mffa->boundary[n].meta.center[ZZ]);
					}
			}
			count1 = 0;
			P_in_interval = 0;
			P_out_interval = 0;
			P_vir_interval = 0;
		}
	}
	/*<<< MFFA <<<*/
 
    debug_gmx();
    if (ekind->cosacc.cos_accel == 0) {
      calc_ke_part(state->v,&(ir->opts),mdatoms,ekind,nrnb,state->lambda);
    } else {
      calc_ke_part_visc(state->box,state->x,state->v,&(ir->opts),
			mdatoms,ekind,nrnb,state->lambda);
    }

    /* since we use the new coordinates in calc_ke_part_visc, we should use
     * the new box too. Still, won't this be offset by one timestep in the
     * energy file? / EL 20040121
     */ 

    debug_gmx();
    /* Calculate center of mass velocity if necessary, also parallellized */
    if (bStopCM && !bFFscan && !bRerunMD)
      calc_vcm_grp(fplog,mdatoms->start,mdatoms->homenr,mdatoms,
		   state->x,state->v,vcm);
    
    /* Determine the wallclock run time up till now */
    run_time = (double)time(NULL) - (double)start_t;

    /* Check whether everything is still allright */    
    if (bGotTermSignal || bGotUsr1Signal) {
      if (bGotTermSignal || ir->nstlist == 0)
	terminate = 1;
      else
	terminate = -1;
      if (!PAR(cr))
	terminate_now = terminate;
      if (fplog) {
	fprintf(fplog,
		"\n\nReceived the %s signal, stopping at the next %sstep\n\n",
		bGotTermSignal ? "TERM" : "USR1",terminate==-1 ? "NS " : "");
	fflush(fplog);
      }
      fprintf(stderr,
	      "\n\nReceived the %s signal, stopping at the next %sstep\n\n",
	      bGotTermSignal ? "TERM" : "USR1",terminate==-1 ? "NS " : "");
      fflush(stderr);
      bGotTermSignal = FALSE;
      bGotUsr1Signal = FALSE;
    } else if (MASTER(cr) && (bNS || ir->nstlist <= 0) &&
	       (max_hours > 0 && run_time > max_hours*60.0*60.0*0.99) &&
	       terminate == 0) {
      /* Signal to terminate the run */
      terminate = (ir->nstlist == 0 ? 1 : -1);
      if (!PAR(cr))
	terminate_now = terminate;
     if (fplog)
	fprintf(fplog,"\nStep %d: Run time exceeded %.3f hours, will terminate the run\n",step,max_hours*0.99);
      fprintf(stderr, "\nStep %d: Run time exceeded %.3f hours, will terminate the run\n",step,max_hours*0.99);
    }
    
    bGStat = (bGStatEveryStep || bStopCM);

    if (ir->nstlist == -1 && !bRerunMD) {
      /* When bGStatEveryStep=FALSE, global_stat is only called
       * when we check the atom displacements, not at NS steps.
       * This means that also the bonded interaction count check is not
       * performed immediately after NS. Therefore a few MD steps could
       * be performed with missing interactions. But wrong energies are never
       * written to file, since energies are only written after global_stat
       * has been called.
       */
      if (step >= step_nscheck) {
	nabnsb = natoms_beyond_ns_buffer(ir,fr,&top->cgs,*scale_tot,state->x);
	bGStat = TRUE;
      } else {
	/* This is not necessarily true,
	 * but step_nscheck is determined quite conservatively.
	 */
	nabnsb = 0;
      }
    } else {
      if (bNS) {
	bGStat = TRUE;
      }
    }

    /* In parallel we only have to check for checkpointing in steps
     * where we do global communication, otherwise the other nodes don't know.
     */
    if (MASTER(cr) && ((bGStat || !PAR(cr)) &&
		       cpt_period >= 0 &&
		       (cpt_period == 0 || 
			run_time >= nchkpt*cpt_period*60.0))) {
      if (chkpt == 0) {
	nchkpt++;
      }
      /* Write checkpoint at the next energy output step (if there is one),
       * or after 0.2*cpt_period at any step.
       */
      if (!bGStatEveryStep || ir->nstenergy == 0 || cpt_period == 0 ||
	  run_time >= (nchkpt + 0.2)*cpt_period*60.0) {
	chkpt = 1;
      } else {
	chkpt = -1;
      }
    }

    /* With exact energy averages (bGStatEveryStep=TRUE)
     * we should also write energy at first, last and continuation steps
     * such that we can get exact averages over a series of runs.
     * We therefore try to checkpoint at energy output frames.
     *
     * This is not necessary when we use the append-file-feature, so we avoid
     * the extra first frame in that case.
     */
    do_ene = (do_per_step(step,ir->nstenergy) ||
	      (bGStatEveryStep && ((bFirstStep && !bAppend) ||
				   bLastStep || bCPT)));

    if (do_ene || do_log) {
      bGStat = TRUE;
    }

    if (!bGStat) {
      /* We will not sum ekinh_old, so signal that we still have to do it */
      bSumEkinhOld = TRUE;
    } else {
      if (PAR(cr)) {
	wallcycle_start(wcycle,ewcMoveE);
	/* Globally (over all NODEs) sum energy, virial etc. 
	 * This includes communication 
	 */
	global_stat(fplog,cr,enerd,force_vir,shake_vir,mu_tot,
		    ir,ekind,bSumEkinhOld,constr,vcm,
		    ir->nstlist==-1 ? &nabnsb : NULL,&chkpt,&terminate);
	if (terminate != 0) {
	  terminate_now = terminate;
	  terminate = 0;
	}
	
	wallcycle_stop(wcycle,ewcMoveE);
	bSumEkinhOld = FALSE;
      }

      /* This is just for testing. Nothing is actually done to Ekin
       * since that would require extra communication.
       */
      if (!bNEMD && debug && (vcm->nr > 0)) {
	correct_ekin(debug,mdatoms->start,mdatoms->start+mdatoms->homenr,
		     state->v,vcm->group_p[0],
		     mdatoms->massT,mdatoms->tmass,ekin);
      }
    
      /* Do center of mass motion removal */
      if (bStopCM && !bFFscan && !bRerunMD) {
	check_cm_grp(fplog,vcm,1);
	do_stopcm_grp(fplog,mdatoms->start,mdatoms->homenr,mdatoms->cVCM,
		      state->x,state->v,vcm);
	inc_nrnb(nrnb,eNR_STOPCM,mdatoms->homenr);
	/*
	  calc_vcm_grp(fplog,START(nsb),HOMENR(nsb),mdatoms->massT,x,v,vcm);
	  check_cm_grp(fplog,vcm);
	  do_stopcm_grp(fplog,START(nsb),HOMENR(nsb),x,v,vcm);
	  check_cm_grp(fplog,vcm);
	*/
      }
      
      /* Add force and shake contribution to the virial */
      m_add(force_vir,shake_vir,total_vir);
      
      /* Calculate the amplitude of the cosine velocity profile */
      ekind->cosacc.vcos = ekind->cosacc.mvcos/mdatoms->tmass;
      
      /* Sum the kinetic energies of the groups & calc temp */
      enerd->term[F_TEMP] = sum_ekin((bRerunMD && !rerun_fr.bV),
				     &(ir->opts),ekind,ekin,
				     &(enerd->term[F_DKDL]));
      enerd->term[F_EKIN] = trace(ekin);
      
      /* Calculate pressure and apply LR correction if PPPM is used.
       * Use the box from last timestep since we already called update().
       */
	/*>>> MFFA >>>*/
      enerd->term[F_PRES] = 
		  calc_pres(fr->ePBC,ir->nwall,volume,ekin,total_vir,pres,
			     (fr->eeltype==eelPPPM)?enerd->term[F_COUL_RECIP]:0.0);
	/*<<< MFFA <<<*/
      
      /* Calculate long range corrections to pressure and energy */
      if (bTCR || bFFscan) {
	set_avcsixtwelve(fplog,fr,top_global);
      }
      
      /* Calculate long range corrections to pressure and energy */
      calc_dispcorr(fplog,ir,fr,step,top_global->natoms,
		    lastbox,state->lambda,
		    pres,total_vir,enerd->term);
      
      enerd->term[F_ETOT] = enerd->term[F_EPOT] + enerd->term[F_EKIN];
      
      switch (ir->etc) {
      case etcNO:
      case etcBERENDSEN:
	break;
      case etcNOSEHOOVER:
	enerd->term[F_ECONSERVED] =
	  enerd->term[F_ETOT] + nosehoover_energy(&(ir->opts),ekind,
						  state->nosehoover_xi,
						  state->therm_integral);
	break;
      case etcVRESCALE:
	enerd->term[F_ECONSERVED] =
	  enerd->term[F_ETOT] + vrescale_energy(&(ir->opts),
						state->therm_integral);
	break;
      }
      
      /* Complicated conditional when bGStatEveryStep=FALSE.
       * We can not just use bGStat, since then the simulation results
       * would depend on nstenergy and nstlog or step_nscheck.
       */
      if ((state->flags & (1<<estPRES_PREV)) &&
	  (bGStatEveryStep ||
	   (ir->nstlist > 0 && step % ir->nstlist == 0) ||
	   (ir->nstlist < 0 && nabnsb > 0) ||
	   (ir->nstlist == 0 && bGStat))) {
	/* Store the pressure in t_state for pressure coupling
	 * at the next MD step.
	 */
	copy_mat(pres,state->pres_prev);
      }

      /* Check for excessively large energies */
      if (bIonize) {
#ifdef GMX_DOUBLE
	real etot_max = 1e200;
#else
	real etot_max = 1e30;
#endif
	if (fabs(enerd->term[F_ETOT]) > etot_max) {
	  fprintf(stderr,"Energy too large (%g), giving up\n",
		  enerd->term[F_ETOT]);
	  break;
	}
      }
    }
    
    /* The coordinates (x) were unshifted in update */
    if (bFFscan && (shellfc==NULL || bConverged)) {
      if (print_forcefield(fplog,enerd->term,mdatoms->homenr,f,buf,xcopy,
			   &(top_global->mols),mdatoms->massT,pres)) {
	if (gmx_parallel_env)
	  gmx_finalize(cr);
	fprintf(stderr,"\n");
	exit(0);
      }
    }
    
    if (bTCR) {
      /* Only do GCT when the relaxation of shells (minimization) has converged,
       * otherwise we might be coupling to bogus energies. 
       * In parallel we must always do this, because the other sims might
       * update the FF.
       */
      
      /* Since this is called with the new coordinates state->x, I assume
       * we want the new box state->box too. / EL 20040121
       */
      do_coupling(fplog,nfile,fnm,tcr,t,step,enerd->term,fr,
		  ir,MASTER(cr),
		  mdatoms,&(top->idef),mu_aver,
		  top_global->mols.nr,cr,
		  state->box,total_vir,pres,
		  mu_tot,state->x,f,bConverged);
      debug_gmx();
    }

    /* Time for performance */
    if (((step % stepout) == 0) || bLastStep)
      update_time();

    /* Output stuff */
    if (MASTER(cr)) {
      bool do_dr,do_or;
      
      /* If we do reruns, the step numbers in the output energy frames
       * cannot be used for averages (since energies are only calculated
       * for trajectory frames).
       */
       upd_mdebin(mdebin,fp_dgdl,bGStatEveryStep && !bRerunMD,
		 mdatoms->tmass,step_ene,t,enerd,state,lastbox,
		 shake_vir,force_vir,total_vir,pres,
		 ekind,mu_tot,constr);
      
      do_dr  = do_per_step(step,ir->nstdisreout);
      do_or  = do_per_step(step,ir->nstorireout);
      
      print_ebin(fp_ene,do_ene,do_dr,do_or,do_log?fplog:NULL,step,step_ene,t,
		 eprNORMAL,bCompact,mdebin,fcd,groups,&(ir->opts));
      
      if (ir->ePull != epullNO)
	{
	  pull_print_output(ir->pull,step,t);
	}
      
      if (do_per_step(step,ir->nstlog))
	{
		if(fflush(fplog) != 0)
		{
			gmx_fatal(FARGS,"Cannot flush logfile - maybe you are out of quota?");
		}
	}
	/*>>> MFFA >>>*/
	if ((do_per_step(step,ir->nstvout) || bLastStep) 
			&& (step != 0 && !bUPDSDB ))
		if (MASTER(cr) && (mffa->meta.bSDB)) {
			for (i=0; i < mffa->meta.PNUMB; i++) {
				if (mffa->boundary[i].meta.TYPE == efSPHERE) {
					fprintf(stderr,"Dump: %s at time: %f\n ", 
							mffa->filenames[i], t);
					dump_mffa_boundary(&mffa->boundary[i], 
							mffa->filenames[i], 
							mffa->meta.bQUENCH, t);
				}
			}
		}
	}
	/*<<< MFFA <<<*/
    
    /* Remaining runtime */
    if (MULTIMASTER(cr) && do_verbose) {
      if (shellfc)
	fprintf(stderr,"\n");
      print_time(stderr,start_t,step,ir);
    }

    /* Replica exchange */
    bExchanged = FALSE;
    if ((repl_ex_nst > 0) && (step > 0) && !bLastStep &&
	do_per_step(step,repl_ex_nst))
      bExchanged = replica_exchange(fplog,cr,repl_ex,state_global,enerd->term,
				    state,step,t);
    if (bExchanged && PAR(cr)) {
      if (DOMAINDECOMP(cr)) {
	dd_partition_system(fplog,step,cr,TRUE,
			    state_global,top_global,ir,
			    state,&f,&buf,mdatoms,top,fr,vsite,shellfc,constr,
			    nrnb,wcycle,FALSE);
      } else {
	bcast_state(cr,state,FALSE);
      }
    }
    
    bFirstStep = FALSE;
    
    if (bRerunMD) 
      /* read next frame from input trajectory */
      bNotLastFrame = read_next_frame(status,&rerun_fr);

    if (!bRerunMD || !rerun_fr.bStep) {
      /* increase the MD step number */
      step++;
      step_rel++;
    }

    cycles = wallcycle_stop(wcycle,ewcSTEP);
    if (DOMAINDECOMP(cr) && wcycle)
      dd_cycles_add(cr->dd,cycles,ddCyclStep);
  }
  /* End of main MD loop */
  debug_gmx();

  if (bRerunMD)
    close_trj(status);

  if (!(cr->duty & DUTY_PME)) {
    /* Tell the PME only node to finish */
    gmx_pme_finish(cr);
  }
	  
  if (MASTER(cr)) {
    if (bGStatEveryStep && !bRerunMD) {
      print_ebin(fp_ene,FALSE,FALSE,FALSE,fplog,step,step_ene,t,
		 eprAVER,FALSE,mdebin,fcd,groups,&(ir->opts));
      print_ebin(fp_ene,FALSE,FALSE,FALSE,fplog,step,step_ene,t,
		 eprRMS,FALSE,mdebin,fcd,groups,&(ir->opts));
    }
    close_enx(fp_ene);
    if (ir->nstxtcout)
      close_xtc(fp_xtc);
    close_trn(fp_trn);
    if (fp_dgdl)
      gmx_fio_fclose(fp_dgdl);
    if (fp_field)
      gmx_fio_fclose(fp_field);
  }
  debug_gmx();
  
  if (ir->nstlist == -1 && nns>0 && fplog) {
    fprintf(fplog,"Average neighborlist lifetime: %.1f steps, std.dev.: %.1f steps\n",ns_s1/nns,sqrt(ns_s2/nns - sqr(ns_s1/nns)));
    fprintf(fplog,"Average number of atoms that crossed the half buffer length: %.1f\n\n",ns_ab/nns);
  }

  if (shellfc && fplog) {
    fprintf(fplog,"Fraction of iterations that converged:           %.2f %%\n",
	    (nconverged*100.0)/step_rel);
    fprintf(fplog,"Average number of force evaluations per MD step: %.2f\n\n",
	    tcount/step_rel);
  }

  if (repl_ex_nst > 0 && MASTER(cr)) {
    print_replica_exchange_statistics(fplog,repl_ex);
  }

  *nsteps_done = step_rel;
    
  return start_t;
}

/*>>> MFFA >>>*/
void echo_mffa_meta(const t_commrec *cr, MFFA_META *meta)
{
  int id, i;
  real r;

  id = cr->nodeid;

  fprintf(stderr, "node %d: size_ndx=%d\n", id, meta->size_ndx);
  if (meta->bSDB)
	  fprintf(stderr, "node %d: bSDB=True\n", id);
  else
	  fprintf(stderr, "node %d: bSDB=False\n", id);
  if (meta->bPC)
	  fprintf(stderr, "node %d: bPC=True\n", id);
  else
	  fprintf(stderr, "node %d: bPC=False\n", id);
  if (meta->bNDX)
	  fprintf(stderr, "node %d: bNDX=True\n", id);
  else
	  fprintf(stderr, "node %d: bNDX=False\n", id);
  if (meta->bQUENCH)
	  fprintf(stderr, "node %d: bQUENCH=True\n", id);
  else
	  fprintf(stderr, "node %d: bQUENCH=False\n", id);
  fprintf(stderr, "node %d: PNUMB=%d\n", id, meta->PNUMB);
  fprintf(stderr, "node %d: Fb_in=%f\n", id, meta->Fb_in);
  fprintf(stderr, "node %d: Fb_out=%f\n", id, meta->Fb_out);
  fprintf(stderr, "node %d: P_int=%f\n", id, meta->P_int);
  fprintf(stderr, "node %d: int_vir=|(%f,%f,%f),(%f,%f,%f),(%f,%f,%f)|\n", id, 
		  meta->int_vir[XX][XX], meta->int_vir[XX][YY], meta->int_vir[XX][ZZ], 
		  meta->int_vir[YY][XX], meta->int_vir[YY][YY], meta->int_vir[YY][ZZ], 
		  meta->int_vir[ZZ][XX], meta->int_vir[ZZ][YY], meta->int_vir[ZZ][ZZ]);
}

void echo_mffa_boundary_meta(const t_commrec *cr, MFFA_BOUNDARY_META *meta)
{
	int id;
	id = cr->nodeid;
	
	fprintf(stderr, "node %d: TYPE=%d\n", id, meta->TYPE);
	fprintf(stderr, "node %d: AXIS=%d\n", id, meta->AXIS);
	fprintf(stderr, "node %d: index_nr_out=%d\n", id, meta->index_nr_out);
	fprintf(stderr, "node %d: index_nr_in=%d\n", id, meta->index_nr_in);
	fprintf(stderr, "node %d: pos_in=%f\n", id, meta->pos_in);
	fprintf(stderr, "node %d: pos_out=%f\n", id, meta->pos_out);
	fprintf(stderr, "node %d: GRID=%f\n", id, meta->GRID);
	fprintf(stderr, "node %d: a_out=%f\n", id, meta->a_out);
	fprintf(stderr, "node %d: b_out=%f\n", id, meta->b_out);
	fprintf(stderr, "node %d: a_in=%f\n", id, meta->a_in);
	fprintf(stderr, "node %d: b_in=%f\n", id, meta->b_in);
	fprintf(stderr, "node %d: CUTOFF=%f\n", id, meta->CUTOFF);
	fprintf(stderr, "node %d: center=(%f,%f,%f)\n", id, 
			meta->center[XX], meta->center[YY], meta->center[ZZ]);
}

void echo_mffa_boundary(const t_commrec *cr, MFFA_BOUNDARY *boundary)
{
	int id, i;
	id = cr->nodeid;

	echo_mffa_boundary_meta(cr, &boundary->meta);
	fprintf(stderr, "node %d: Inner potential\n", id);
	for (i=0; i < boundary->meta.index_nr_in; i++)
		fprintf(stderr, "node %d: %f\n", cr->nodeid, boundary->F_bounds[i]);
	fprintf(stderr, "node %d: Outer potential\n", id);
	for (; i < boundary->meta.index_nr_out; i++)
		fprintf(stderr, "node %d: %f\n", cr->nodeid, boundary->F_bounds[i]);
}

void echo_mffa_dynamic(const t_commrec *cr, MFFA_DYNAMIC *mffa_dynamic, int n)
{
  int id, i;
  id = cr->nodeid;

  for (i=0; i < n; i++) { 
	  fprintf(stderr, "node %d: dynamic[%d]->TFb_in=%f\n", id, i, 
			  mffa_dynamic[i].TFb_in);
	  fprintf(stderr, "node %d: dynamic[%d]->TFb_out=%f\n", id, i, 
			  mffa_dynamic[i].TFb_out);
	  fprintf(stderr, "node %d: dynamic[%d]->TPb_in=%f\n", id, i, 
			  mffa_dynamic[i].TPb_in);
	  fprintf(stderr, "node %d: dynamic[%d]->TPb_out=%f\n", id, i, 
			  mffa_dynamic[i].TPb_out);
	  fprintf(stderr, "node %d: dynamic[%d]->AREA_in=%f\n", id, i, 
			  mffa_dynamic[i].AREA_in);
	  fprintf(stderr, "node %d: dynamic[%d]->AREA_out=%f\n", id, i, 
			  mffa_dynamic[i].AREA_out);
  }
}

void echo_pressures(const t_commrec *cr, rvec *pressure_plane, 
		real *pressure_sphere)
{
  int id;
  id = cr->nodeid;

  fprintf(stderr, "node %d: pressure_sphere=(%f,%f)\n", id, 
		  pressure_sphere[0], pressure_sphere[1]);
  fprintf(stderr, "node %d: pressure_plane=((%f,%f,%f),(%f,%f,%f))\n", id, 
		  pressure_plane[0][XX], pressure_plane[0][YY], 
		  pressure_plane[0][ZZ], 
		  pressure_plane[1][XX], pressure_plane[1][YY], 
		  pressure_plane[1][ZZ]);
}

/*<<< MFFA <<<*/

