/*
 * NAME
 *   g_entropy
 *     Calculate conformational entropy build-up.
 * SYNOPSIS
 *   g_entropy -s topol.tpr -f traj.xtc -o entropy.xvg
 * AUTHORS
 *   Martti Louhivuori (m.j.louhivuori@rug.nl)
 *   Alex de Vries (a.h.de.vries@rug.nl)
 * 
 * ---
 * Copyright (C) 2008  Martti Louhivuori, Alex de Vries
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * (version 2) as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
 * 02110-1301, USA.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <math.h>
#include <string.h>
#include <time.h>

#include "statutil.h"
#include "sysstuff.h"
#include "typedefs.h"
#include "smalloc.h"
#include "macros.h"
#include "vec.h"
#include "pbc.h"
#include "copyrite.h"
#include "futil.h"
#include "statutil.h"
#include "index.h"
#include "confio.h"
#include "trnio.h"
#include "mshift.h"
#include "xvgr.h"
#include "do_fit.h"
#include "rmpbc.h"
#include "txtdump.h"
#include "matio.h"
#include "eigio.h"
#include "eigensolver.h"

static const char *authors[] = {
	"Written by Martti Louhivuori (m.j.louhivuori@rug.nl)",
	"Copyright (c) 2008  Martti Louhivuori, Alex de Vries"
};
static const char *gpl[] = {
	"This program is free software; you can redistribute it and/or",
	"modify it under the terms of the GNU General Public License",
	"(version 2) as published by the Free Software Foundation."
};
void sp_print(FILE *out, const char *txt)
{
	int i, s;
	s = (int) (80 - strlen(txt)) / 2.0;
	for (i=0; i < s; i++) 
		fprintf(out, " ");
	fprintf(out, "%s\n", txt);
}

int gmx_entropy(int argc,char *argv[])
{
  static char *desc[] = {
    "[TT]g_covar[tt] calculates and diagonalizes the (mass-weighted)",
    "covariance matrix.",
    "All structures are fitted to the structure in the structure file.",
    "When this is not a run input file periodicity will not be taken into",
    "account. When the fit and analysis groups are identical and the analysis",
    "is non mass-weighted, the fit will also be non mass-weighted.",
    "[PAR]",
    "The eigenvectors are written to a trajectory file ([TT]-v[tt]).",
    "When the same atoms are used for the fit and the covariance analysis,",
    "the reference structure for the fit is written first with t=-1.",
    "The average (or reference when [TT]-ref[tt] is used) structure is",
    "written with t=0, the eigenvectors",
    "are written as frames with the eigenvector number as timestamp.",
    "[PAR]",
    "The eigenvectors can be analyzed with [TT]g_anaeig[tt].",
    "[PAR]",
    "Option [TT]-ascii[tt] writes the whole covariance matrix to",
    "an ASCII file. The order of the elements is: x1x1, x1y1, x1z1, x1x2, ...",
    "[PAR]",
    "Option [TT]-xpm[tt] writes the whole covariance matrix to an xpm file.",
    "[PAR]",
    "Option [TT]-xpma[tt] writes the atomic covariance matrix to an xpm file,",
    "i.e. for each atom pair the sum of the xx, yy and zz covariances is",
    "written.[PAR]",
    "With the [TT]-dist[tt] option a list of atom pairs is read from the",
    "index file and the distance between these pairs, minus the average",
    "distance is used for the covariance analysis. In this manner one",
    "can analyses correlations between distance fluctuations. The distances",
    "could be derived from e.g. NMR experiments (distance restraints) or",
    "from hydrogen bonds or your choosing."
  };
  static bool bFit=TRUE,bRef=FALSE,bM=TRUE,bPBC=TRUE,bRemTr=TRUE;
  static bool bMf=TRUE;
  static int step=1, minanal=2;
  static int debug = 0;
  static real temperature = 298.0;
  static real Drel = 1.0;
  static real facD = 0.0;
  static real converter = 15.2325;        /* k_B * T * e**2 / hbar**2 */
  static real converter2 = 0.5*8.3145112; /* 0.5 * k_b */
  FILE       *entropyfile=NULL, *tracefile=NULL;
  int        status;
  int        ePBC;
  t_topology top;
  t_atoms    *atoms;  
  rvec       *x, *xread, *xref, *xav;
  real       *cache_mat;
  matrix     box;
  real       *sqrtm, *mat, *eigval, trace, inv_nframes, t, tstart;
  real       xj, *w_rls=NULL, *w_rlsf=NULL;
  int        natoms, nat, ndim, nframes;
  char       *fn_reference, *fn_trajectory, *fn_index, *fn_index_cov;
  char       *fn_entropy, *fn_trace, *fn_average;
  char       str[STRLEN], *fitname, *ananame;
  int        i, j, k, l, d, dj, nfit;
  atom_id    *index,*ifit;
  real       *tmp;
  int        maxframes, maxread;
  real       antime;
  real       detC;
  real       diffcor;
  
  t_pargs pa[] = {
    { "-fit",  FALSE, etBOOL, {&bFit},
      "Fit to a reference structure"},
    { "-remtr",  FALSE, etBOOL, {&bRemTr},
      "Remove overall translation"},
    { "-ref",  FALSE, etBOOL, {&bRef},
      "Use the deviation from the conformation in the structure file instead of from the average" },
    { "-mwa",  TRUE, etBOOL, {&bM},
      "Mass-weighted covariance analysis"},
    { "-pbc",  FALSE,  etBOOL, {&bPBC},
      "Apply corrections for periodic boundary conditions" },
    { "-mwf",  TRUE, etBOOL, {&bMf},
      "Mass-weighted fit"},
    { "-step",  FALSE, etINT, {&step}, 
      "Output frequency (in frames)"},
    { "-minanal",  FALSE, etINT, {&minanal}, 
      "Calculate covariance matrix for at least minanal frames"},
    { "-temp",  FALSE, etREAL, {&temperature}, 
      "Simulation temperature (in K)"},
    { "-Drel",  FALSE, etREAL, {&Drel}, 
      "Diffusion coefficient for correcting rel motion"},
    { "-facD",  FALSE, etREAL, {&facD}, 
      "Factor with which to correct for rel motion"},
  };
  t_filenm fnm[] = { 
	  { efTRX, "-f",  NULL, ffREAD }, 
	  { efTPS, NULL,  NULL, ffREAD },
	  { efNDX, NULL,  NULL, ffOPTRD },
	  { efNDX, "-ncov",  NULL, ffOPTRD },
	  { efSTO, "-av", "average", ffOPTWR },
	  { efXVG, "-o", "entropy", ffWRITE },
	  { efXVG, "-t", "trace", ffOPTWR } 
  };
#define NFILE asize(fnm) 

  CopyRight(stderr,argv[0]);
	for (i=0; i < (int)asize(authors); i++)
		sp_print(stderr, authors[i]);
	fprintf(stderr, "\n");
	for (i=0; i < (int)asize(gpl); i++)
		sp_print(stderr, gpl[i]);
	fprintf(stderr, "\n");
  parse_common_args(&argc,argv,PCA_CAN_TIME | PCA_TIME_UNIT | PCA_BE_NICE,
		    NFILE,fnm,asize(pa),pa,asize(desc),desc,0,NULL);

  fn_reference  = ftp2fn(efTPS,NFILE,fnm);
  fn_trajectory = ftp2fn(efTRX,NFILE,fnm);
  fn_entropy    = ftp2fn(efXVG, NFILE, fnm);
  fn_index      = ftp2fn_null(efNDX,NFILE,fnm);
  fn_index_cov  = opt2fn_null("-ncov",NFILE,fnm);
  fn_average    = ftp2fn_null(efSTO,NFILE,fnm);
  fn_trace      = opt2fn_null("-t", NFILE, fnm);

  /* read topology */
  read_tps_conf(fn_reference, str, &top, &ePBC, &xref, NULL, box, TRUE);
  atoms=&top.atoms;

  /* read indeces */
  if (bFit || bRemTr) {
	  printf("\nChoose a group for the least squares fit\n"); 
	  get_index(atoms, fn_index, 1, &nfit, &ifit, &fitname);
	  if (nfit < 3) 
		  gmx_fatal(FARGS,"Need >= 3 points to fit!\n");
  } else {
	  nfit=0;
  }
  printf("\nChoose a group for the covariance analysis\n"); 
  if (fn_index_cov)
	  get_index(atoms, fn_index_cov, 1, &natoms, &index, &ananame);
  else
	  get_index(atoms, fn_index, 1, &natoms, &index, &ananame);

  /* prepare mass arrays */
  snew(w_rls, atoms->nr);
  snew(w_rlsf, atoms->nr);
  if (bFit || bRemTr) {
	  for (i=0; (i<nfit); i++) {
		  w_rls[ifit[i]] = atoms->atom[ifit[i]].m;
		  if (bMf) 
			  w_rlsf[ifit[i]] = w_rls[ifit[i]];
		  else
			  w_rlsf[ifit[i]] = 1.0;
	  }
  }
  snew(sqrtm, natoms);
  for(i=0; (i<natoms); i++) {
	  if (bM) 
		  sqrtm[i] = sqrt(atoms->atom[index[i]].m);
	  else 
		  sqrtm[i] = 1.0;
  }

  /* prepare reference frame */
  if (bPBC)
    rm_pbc(&(top.idef), ePBC, atoms->nr, box, xref, xref);
  if (bFit || bRemTr) 
    reset_x(nfit, ifit, atoms->nr, NULL, xref, w_rlsf);

  /* calculate global average structure */
  nat=read_first_x(&status, fn_trajectory, &t, &xread, box);
  if (nat != atoms->nr)
    fprintf(stderr,"\nWARNING: number of atoms in tpx (%d) and trajectory (%d) do not match\n",natoms,nat);
  maxframes = 0;
  snew(xav,natoms);
  if (bRef)
	  xav = xref;
  else {
	  do {
		  /* remove periodicity & fit */
		  if (bPBC)
			  rm_pbc(&(top.idef), ePBC, nat, box, xread, xread);
		  if (bFit || bRemTr) 
			  reset_x(nfit, ifit, nat, NULL, xread, w_rls);
		  if (bFit) 
			  do_fit(nat, w_rls, xref, xread);
		  /* update cache */
		  for (i=0; i<natoms; i++) 
			  rvec_inc(xav[i], xread[index[i]]);
		  maxframes++;
	  } while (read_next_x(status, &t, nat, xread, box));
	  close_trj(status); 
	  /* normalise w/ number of frames */
	  inv_nframes = 1.0/maxframes;
	  for (i=0; i<natoms; i++)
		  for (d=0; d<DIM; d++) {
			  xav[i][d] *= inv_nframes;
		  }
  }
  /* write average structure? */
  if (fn_average)
	  write_sto_conf_indexed(fn_average, "Average structure", 
			  atoms, xav, NULL, epbcNONE, box, natoms, index);

  /* open output files */
  entropyfile = xvgropen(fn_entropy, 
		  "Conformational entropy", "t [ps]", "S' [J / mol K]");  

  if (fn_trace) 
	  tracefile = xvgropen(fn_trace, 
			  "Trace of the Schlitter matrix", "t [ps]", "");

  /* main loop */
  nat=read_first_x(&status, fn_trajectory, &t, &xread, box);
  nframes=0;
  tstart = t;
  maxread = minanal;
  ndim=natoms*DIM;
  snew(mat,ndim*ndim);
  snew(cache_mat,ndim*ndim);
  do { 
	  nframes++;
	  snew(x,natoms);
	  /* remove periodicity & fit */
	  if (bPBC)
		  rm_pbc(&(top.idef), ePBC, nat, box, xread, xread);
	  if (bFit || bRemTr) 
		  reset_x(nfit, ifit, nat, NULL, xread, w_rls);
	  if (bFit) 
		  do_fit(nat, w_rls, xref, xread);
	  /* deviation from the average */
	  for (i=0; i<natoms; i++) {
		  rvec_sub(xread[index[i]], xav[i], x[i]);
	  }
	  /* add cross-products to the covariance matrix */
	  for (j=0; j<natoms; j++) {
		  for (dj=0; dj<DIM; dj++) {
			  k = ndim * (DIM * j + dj);
			  xj = x[j][dj];
			  for (i=j; i<natoms; i++) {
				  l = k + DIM * i;
				  for(d=0; d<DIM; d++)
					  cache_mat[l+d] += x[i][d] * xj;
			  }
		  }
	  }
	  /* output? */
	  if (nframes == maxread) {
		  /* currect time */
		  antime = convert_time(t) - convert_time(tstart);
		  /* correct the covariance matrix for mass, diffusion and frames */
		  diffcor = Drel * facD * (t - tstart);
		  inv_nframes = 1.0/nframes;
		  for (j=0; j<natoms; j++)
			  for (dj=0; dj<DIM; dj++)
				  for (i=j; i<natoms; i++) {
					  k = ndim*(DIM*j+dj)+DIM*i;
					  for (d=0; d<DIM; d++)
						  mat[k+d] = cache_mat[k+d] * inv_nframes * sqrtm[i] 
							  * sqrtm[j] - diffcor;
			  }
		  /* symmetrize the matrix */
		  for (j=0; j<ndim; j++)
			  for (i=j; i<ndim; i++)
				  mat[ndim*i+j]=mat[ndim*j+i];
		  /* calculate the trace */
		  trace=0;
		  for(i=0; i<ndim; i++)
			  trace+=mat[i*ndim+i];
		  /* output trace? */
		  if (fn_trace) fprintf(tracefile, "%g %g\n", antime, trace);
		  
		  /* build schlitter matrix */
		  for (i=0;i<ndim;i++) {
			  for (j=0;j<ndim;j++) {
				  if (i==j) {
					  mat[i*ndim+j] = 1.0 + converter*temperature*mat[i*ndim+j]; 
				  } else {
					  mat[i*ndim+j] = converter*temperature*mat[i*ndim+j]; 
				  }
			  }
		  }
		  /* call diagonalization routine */
		  snew(eigval, ndim);
		  snew(tmp, ndim*ndim);
		  memcpy(tmp, mat, ndim*ndim*sizeof(real));
		  eigensolver(tmp, ndim, 0, ndim, eigval, mat);
		  sfree(tmp);
		  
		  /* calculate entropy */
		  detC = 0.0;
		  for (i=0; (i<ndim); i++) {
			  if (eigval[ndim-1-i] > 1e-05) { 
				  detC += converter2 * log(eigval[ndim-1-i]); 
			  } else if (eigval[ndim-1-i] < 1e-05) { 
				  detC += converter2 * log(-1.0 * eigval[ndim-1-i]); 
			  }
		  }
		  /* output entropy */
		  fprintf(entropyfile, "%g %g\n", antime, detC);
		  /* udpate counter */
		  maxread += step;
	  }
  } while(read_next_x(status, &t, nat, xread, box));
  close_trj(status);

  fclose(entropyfile);
  if (fn_trace) fclose(tracefile);

  thanx(stderr);
  
  sfree(w_rlsf);
  sfree(w_rls);

  return 0;
}

