/*
 * NAME
 *   g_mffa
 *     Calculate a mean-field force approximation (MFFA) potential.
 * SYNOPSIS
 *   g_mffa -f rdf.xvg -o mffa.sdb
 * AUTHORS
 *   Jelger Risselada
 *   Martti Louhivuori (m.j.louhivuori@rug.nl)
 *
 * TODO: - add comments
 *       - switch to non-inverted (outer) slope
 * ---
 * Copyright (C) 2009  Jelger Risselada, Martti Louhivuori
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * (version 2) as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
 * 02110-1301, USA.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>
#include <ctype.h>
#include "macros.h"
#include "smalloc.h"
#include "typedefs.h"
#include "copyrite.h"
#include "statutil.h"
#include "tpxio.h"
#include "string2.h"
#include "vec.h"
#include "pbc.h"
#include "xvgr.h"

#define power6(x) (x*x*x*x*x*x)
#define power12(x) (power6(x) * power6(x))

static const char *authors[] = {
	"Written by Martti Louhivuori (m.j.louhivuori@rug.nl)",
	"Copyright (c) 2009  Jelger Risselada, Martti Louhivuori"
};
static const char *gpl[] = {
	"This program is free software; you can redistribute it and/or",
	"modify it under the terms of the GNU General Public License",
	"(version 2) as published by the Free Software Foundation."
};
void sp_print(FILE *out, const char *txt)
{
	int i, s;
	s = (int) (80 - strlen(txt)) / 2.0;
	for (i=0; i < s; i++) 
		fprintf(out, " ");
	fprintf(out, "%s\n", txt);
}

typedef struct {
	real position;
	real force;
} FORCE;

typedef struct {
	real inner;
	real outer;
	real cutoff;
	int angular;
	int radial;
	int steps;
	float resolution;
	real sigma;
	long double sigma6;
	long double sigma12;
	real epsilon;
	real rho;
	real grid;
	int size;
	real vdw;
	real tau;
	rvec center;
	real slope_inner;
	real slope_outer;
} OPTIONS;

void echo_options(OPTIONS options)
{
	printf("Options:\n");
	printf("  options->inner: %f\n", options.inner);
	printf("  options->outer: %f\n", options.outer);
	printf("  options->cutoff: %f\n", options.cutoff);
	printf("  options->angular: %d\n", options.angular);
	printf("  options->radial: %d\n", options.radial);
	printf("  options->steps: %d\n", options.steps);
	printf("  options->resolution: %f\n", options.resolution);
	printf("  options->sigma: %f\n", options.sigma);
	printf("  options->sigma6: %Lf\n", options.sigma6);
	printf("  options->sigma12: %Lf\n", options.sigma12);
	printf("  options->epsilon: %f\n", options.epsilon);
	printf("  options->rho: %f\n", options.rho);
	printf("  options->grid: %f\n", options.grid);
	printf("  options->size: %d\n", options.size);
	printf("  options->vdw: %f\n", options.vdw);
	printf("  options->tau: %f\n", options.tau);
}

void write_sdb(char *filename, FORCE *inner, FORCE *outer, OPTIONS options, 
		char *type)
{
	FILE *fp;
	int i;

	fp = fopen(filename, "w");
	fprintf(fp, "%s\n", type);
	fprintf(fp, "%f\n", options.inner);
	fprintf(fp, "%f\n", options.outer);
	fprintf(fp, "%f\n", options.cutoff);
	fprintf(fp, "%f\n", options.grid);
	fprintf(fp, "%f\n", options.slope_inner);
	fprintf(fp, "%f\n", options.slope_outer);
	fprintf(fp, "%f\n", options.center[0]);
	fprintf(fp, "%f\n", options.center[1]);
	fprintf(fp, "%f\n", options.center[2]);
	fprintf(fp, "%% type\n%% inner cutoff\n%% outer cutoff\n%% cutoff\n");
	fprintf(fp, "%% gridsize\n%% a in\n%% a out\n%% center X Y Z\n#\n");
	if (inner != NULL)
		for (i=0; i < options.steps; i++)
			fprintf(fp, "%f\n", inner[i].force);
	fprintf(fp, "#\n");
	if (outer != NULL)
		for (i=0; i < options.steps; i++)
			fprintf(fp, "%f\n", outer[i].force);
	fprintf(fp, "#\n");
	fclose(fp);
}

float theta_max(float R, float position, float cutoff)
{
	return ( (R*R + position*position - cutoff*cutoff) / (2 * R * position) );
}

void mara_inner(FORCE *buffer, float *slope, double **rdf, OPTIONS options)
{
	int i, j, index, z=0;
	double p, d, r, dr, x, c, w, offset, tau;
	long double lennard_jones, potential;

	fprintf(stdout, "Calculating the inner potential...");
	fflush(stdout);
	if (debug)
		fprintf(stdout, "\n");
	dr = options.cutoff / options.radial;
	if (debug)
		fprintf(stderr, "dr=%g\n", dr);
	for (z=0; z < options.steps; z++) {
		p = z * options.grid;
		d = options.inner + p;
		if (debug)
			fprintf(stderr, "p=%g  d=%g\n", p, d);
		potential = 0.0;
		for (r=options.cutoff; r > p; r -= dr) {
			x = (d*d - r*r + options.inner*options.inner) / (2*d);
			c = (d - x) / r;
			w = M_PI * r * (1 - c*c) * dr;
			index = (int) (r / options.resolution + 0.5);
			lennard_jones = 4.0 * options.epsilon * 
				(options.sigma12 / power12(r) - options.sigma6 / power6(r));
			potential += w * rdf[1][index] * options.rho * lennard_jones;
			if (debug)
				fprintf(stderr, "r=%g  x=%g  c=%g  w=%g  index=%d  LJ=%Lg  potential=%Lg\n", 
						r, x, c, w, index, lennard_jones, potential);
		}
		buffer[z].force = potential;
		buffer[z].position = d;
	}
	/* slope of potential at the interface */
	*slope = (buffer[1].force - buffer[0].force) / options.resolution;
	/* slope & offset of potential at the interface */
	z = (int) (options.tau * options.vdw / options.grid);
	if (debug)
		fprintf(stderr, "z=%d  force=%g %g\n", z, buffer[z].force, buffer[z+1].force);
	*slope = (buffer[z+1].force - buffer[z].force) 
		/ (buffer[z+1].position - buffer[z].position);
	offset = buffer[z].force - (*slope) * buffer[z].position;
	/* linearise forces */
	for (z=z; z >= 0; z--) {
		buffer[z].force = (*slope) * buffer[z].position + offset;
	}
	fprintf(stdout, "done.\n");
	fprintf(stdout, "g(x) = %f * x + %g\n", *slope, offset);
}

void mara_outer(FORCE *buffer, float *slope, double **rdf, OPTIONS options)
{
	int i, j, index, z=0;
	double p, d, d_o, r, dr, c, w, offset, tau;
	long double lennard_jones, potential;

	fprintf(stdout, "Calculating the outer potential...");
	fflush(stdout);
	if (debug)
		fprintf(stdout, "\n");
	dr = options.cutoff / options.radial;
	if (debug)
		fprintf(stderr, "dr=%g\n", dr);
	for (z=0; z < options.steps; z++)
	{
		p = z * options.grid; 
		d = options.outer - p;
		if (debug)
			fprintf(stderr, "p=%g  d=%g\n", p, d);
		potential = 0.0;
		for (r=options.cutoff; r > p; r -= dr) {
			d_o = sqrt(options.outer*options.outer - r*r);
			if (d > d_o) {
				c = (d - d_o) / r;
			} else {
				c = (d_o - d) / r;
			}
			w = M_PI * r * (1 - c*c) * dr;
			index = (int) (r / options.resolution + 0.5);
			lennard_jones = 4.0 * options.epsilon * 
				(options.sigma12 / power12(r) - options.sigma6 / power6(r));
			potential += w * rdf[1][index] * options.rho * lennard_jones;
			if (debug)
				fprintf(stderr, "r=%g  d_o=%g  c=%g  w=%g  index=%d  LJ=%Lg  potential=%Lg\n", 
						r, d_o, c, w, index, lennard_jones, potential);
		}
		buffer[options.steps - 1 - z].force = potential;
		buffer[options.steps - 1 - z].position = d;
	}
	/* slope & offset of potential at the interface */
	z = options.steps - 1 - (int) (options.tau * options.vdw / options.grid);
	if (debug)
		fprintf(stderr, "z=%d  force=%g %g\n", z, buffer[z - 1].force, buffer[z].force);
	*slope = (buffer[z].force - buffer[z-1].force) 
		/ (buffer[z].position - buffer[z-1].position);
	offset = buffer[z].force - (*slope) * buffer[z].position;
	/* linearise forces */
	for (z=z; z < options.steps; z++) {
		buffer[z].force = (*slope) * buffer[z].position + offset;
	}
	fprintf(stdout, "done.\n");
	fprintf(stdout, "f(x) = %f * x + %g\n", *slope, offset);
	*slope = -1 * (*slope); //FIXME: switch to non-inverted slope
}

void inner_void(FORCE *buffer, float *slope, double **rdf, OPTIONS options)
{
	int i, j, index, z=0;
	float position, rstep, astep, theta, r, lennard_jones;
	long double weight, potential;

	fprintf(stdout, "Calculating the inner potential...");
	fflush(stdout);
	if (debug)
		fprintf(stdout, "\n");
	for (z=0; z < options.steps; z++)
	{
		position = options.inner + z*options.grid;
		potential = 0.0;
		theta = theta_max(position, options.cutoff, options.inner);
		rstep = (1.0 - theta) / options.radial;
		if (debug)
			fprintf(stderr, "position=%f, theta=%f, rstep=%f\n", 
					position, theta, rstep);
		for (j=0; j < options.radial; j++) {
			weight = 0.0;
			// initial radius
			r = position * theta;
			r -= sqrt(r*r + options.inner*options.inner - position*position);
			// angular step size
			astep = (options.cutoff - r) / options.angular;
			for (i=0; i < options.angular; i++) {
				index = (int) (r / options.resolution + 0.5);
				if(index <  options.size) {
					lennard_jones = 4.0 * options.epsilon * 
						(options.sigma12 / power12(r) 
						 - options.sigma6 / power6(r));
					weight += r * astep * lennard_jones * theta
						* options.rho * rdf[1][index];
				}
				r += astep;
			} 
			potential += weight * rstep;
			theta += rstep;
		}
		buffer[z].force = 2.0*M_PI * potential;
		buffer[z].position = position;
	}
	/* slope of potential at the interface */
	*slope = (buffer[2].force - buffer[1].force) / options.resolution;
	/* potential at the interface */
	buffer[0].position = options.inner;
	buffer[0].force = buffer[1].force + (buffer[1].force - buffer[2].force);
	fprintf(stdout, "done.\n");
}

void outer_void(FORCE *buffer, float *slope, double **rdf, OPTIONS options)
{
	int i, j, index, z=0;
	float position, rstep, astep, theta, r, lennard_jones;
	double potential, weight, limit, c;

	fprintf(stdout, "Calculating the outer potential...");
	fflush(stdout);
	astep = (float) (options.cutoff / options.angular);
	if (debug)
		fprintf(stderr, "\nastep=%f\n", astep);
	for (z=0; z < options.steps-1; z++)
	{
		position = options.outer - options.cutoff + z * options.grid;
		potential = 0.0;
		theta = theta_max(position, options.cutoff, options.outer);
		rstep = (1.0 + theta) / options.radial;
		limit = options.outer * options.outer - position * position;
		if (debug)
			fprintf(stderr, "position=%f, rstep=%f, limit=%f, theta=%f\n", 
					position, rstep, limit, theta);
		for (j=0; j < options.radial; j++) {
			weight = 0.0;
			r = 0.0;
			c = 2.0 * theta * position;
			for (i=0; i < options.angular; i++) {
				if ( limit <= r * (r  - c) ) {
					index = (int) (r / options.resolution + 0.5);
					if (index <  options.size) {   
						lennard_jones = 4.0 * options.epsilon * 
							(options.sigma12 / power12(r) 
							 - options.sigma6 / power6(r));
						weight -= r * astep * lennard_jones * theta
							* options.rho * rdf[1][index];
					}
				}
				r += astep;
			}
			potential += weight * rstep;
			theta -= rstep;
		}
		buffer[z].force = 2.0*M_PI * potential;
		buffer[z].position = position;
	}
	/* slope of potential at interface */
	*slope = (buffer[options.steps - 2].force 
			- buffer[options.steps - 3].force) 
		/ options.resolution * -1;
	/* potential at the interface */
	buffer[options.steps - 1].position = options.outer;
	buffer[options.steps - 1].force = buffer[options.steps - 2].force + 
		(buffer[options.steps - 2].force - buffer[options.steps - 3].force);
	fprintf(stdout, "done.\n");
}

void outer_void_karplus(FORCE *buffer, double **rdf, OPTIONS options)
{
	int i, j, index, z=0;
	float position, rstep, astep, theta, r, limit, lennard_jones;
	float dist, trapez, gradient;
	double potential, weight;

	fprintf(stdout, "Calculating the outer potential...");
	fflush(stdout);
	if (debug)
		fprintf(stdout, "\n");
	for (z=0; z < options.steps; z++)
	{
		position = options.outer - options.cutoff + z * options.grid;
		if (debug)
			fprintf(stderr, "position: %f\n", position);
		potential = 0.0;
		theta = theta_max(options.outer, position, options.cutoff);
		rstep = (1.0 - theta) / options.radial;
		for (j=0; j < options.radial; j++) {
			weight = 0.0;
			// inner limit
			limit = position * theta;
			limit += sqrt(limit*limit + options.cutoff*options.cutoff 
					- position*position);
			// angular step size
			astep = (limit - options.outer) / options.angular;
			// initial radius
			r = options.outer;
			for (i=0; i < options.angular; i++) {
				dist = position*position + r*r - 2*position*r*theta;
				if (dist < 0.0)
					dist = 0.0;
				else
					dist = sqrt(dist);
				index = (int) (dist / options.resolution + 0.5);
				if (index <  options.size) {
					gradient = (r*theta - position) / dist;
					if (r == options.outer || r == limit)
						trapez = 0.5;
					else
						trapez = 1.0;
					lennard_jones = 4.0 * options.epsilon * 
						(options.sigma12 / power12(dist) 
						 - options.sigma6 / power6(dist));
					weight += r*r * trapez * gradient * options.rho 
						* (lennard_jones / dist) * rdf[1][index];
				}
				r += astep;
			}
			potential += weight * astep;
			theta += rstep;
		}
		buffer[z].force = 2.0*M_PI * rstep * potential;
		buffer[z].position = position;
	}
	fprintf(stdout, "done.\n");
}

void plane_void(FORCE *inner, FORCE *outer, double **rdf, OPTIONS options)
{
	int i, j, index, z=0;
	float position, astep, rstep, theta, r, lennard_jones;
	double potential, weight;
	FORCE *buf;
	float *positions;

	buf = malloc(options.steps * sizeof(FORCE));

	fprintf(stdout, "Calculating planar potentials...");
	fflush(stdout);
	if (debug)
		fprintf(stdout, "\n");
	for (z=0; z < options.steps; z++)
	{
		position = z * options.grid;
		if (debug)
			fprintf(stderr, "position: %f\n", position);
		potential = 0.0;
		theta = position / options.cutoff;
		rstep = (1.0 - theta) / options.steps;
		for (i=0; i < options.steps; i++) {
			weight = 0.0;
			theta += rstep;
			r = position / theta;
			astep = (options.cutoff - r) / options.angular;
			for (j=0; j < options.angular; j++) {
				r += astep;
				index = (int) (r / options.resolution + 0.5);
				if (index <  options.size) {
					lennard_jones = 4.0 * options.epsilon * 
						(options.sigma12 / power12(r) 
						 - options.sigma6 / power6(r));
					weight += r * astep * theta * options.rho 
						* lennard_jones * rdf[1][index];
				}
			}
			potential += weight * rstep;
		}
		inner[z].force = outer[options.steps - z - 1].force
			= (float) (2*M_PI * potential);
		inner[z].position = position + options.inner;
		outer[options.steps - z - 1].position = options.outer - position;
	}
	fprintf(stdout, "done.\n");
}

int gmx_mffa(int argc, char **argv)
{
	int i; 
	static bool bPLANE=FALSE, bQUENCH, bKARPLUS=FALSE;
	double **rdf;
	FORCE *inner, *outer;
	OPTIONS options;
	FILE *plot;
	real a, slope_inner, slope_outer;
	char *type, *fn_plot;

	// default values
	options.inner = 2.0;
	options.outer = 6.0;
	options.cutoff = 1.2;
	options.rho = 8.272676;
	options.epsilon = 5.0;
	options.sigma = 0.47;
	options.grid  = 0.01;
	options.vdw = 0.235;
	options.tau = 0.666666;
	options.angular = 1000;
	options.radial = 1000;
	options.center[0] = options.center[1] = options.center[2] = 20.0;
	options.slope_inner = 0.0;
	options.slope_outer = 0.0;

	static char *desc[] = {
		"Calculate a mean-field force approximation (MFFA) potential."
	};
	t_filenm fnm[] = {
		{ efXVG, "-f", "rdf", ffREAD },
		{ efSDB, "-o", "mffa", ffWRITE },
		{ efDAT, "-p", "plot", ffOPTWR }
	};

#define NFILE asize(fnm)

	t_pargs pa[] = {
		{ "-inner", FALSE, etREAL, {&(options.inner)},
			"inner radius"},
		{ "-outer", FALSE, etREAL, {&(options.outer)},
			"outer radius"},
		{ "-cutoff", FALSE, etREAL, {&(options.cutoff)},
			"cutoff, i.e. boundary width"},
		{ "-plane", FALSE, etBOOL, {&bPLANE},
			"planar system"},
		{ "-quench", FALSE, etBOOL, {&bQUENCH},
			"apply pressure coupling w/ shifting outer potential"},
		{ "-s", FALSE, etREAL, {&(options.sigma)},
			"sigma"},
		{ "-e", FALSE, etREAL, {&(options.epsilon)},
			"epsilon"},
		{ "-r", FALSE, etREAL, {&(options.rho)},
			"rho"},
		{ "-vdw", FALSE, etREAL, {&(options.vdw)}, 
			"van der Waals radius of solvent"}, 
		{ "-tau", FALSE, etREAL, {&(options.tau)}, 
			"hardness of the MFFA for particles close to (or beyond) the boundary"},
		{ "-grid", FALSE, etREAL, {&(options.grid)},
			"grid resolution"},
		{ "-angular", FALSE, etINT, {&(options.angular)},
			"# of angular steps to sample"},
		{ "-radial", FALSE, etINT, {&(options.radial)},
			"# of radial steps to sample"},
		{"-karplus", FALSE, etBOOL, {&bKARPLUS}, 
			"use Karplus method for outer radius calculation"},
		{"-center", FALSE, etRVEC, {&(options.center)}, 
			"center of the potentials"}
	};

	CopyRight(stderr, argv[0]);
	for (i=0; i < (int)asize(authors); i++)
		sp_print(stderr, authors[i]);
	fprintf(stderr, "\n");
	for (i=0; i < (int)asize(gpl); i++)
		sp_print(stderr, gpl[i]);
	fprintf(stderr, "\n");

	parse_common_args(&argc,argv,PCA_CAN_TIME | PCA_CAN_VIEW,
			NFILE,fnm,asize(pa),pa,asize(desc),desc,0,NULL);
	
	/* read rdf */
	options.size = read_xvg(opt2fn("-f", NFILE, fnm), &rdf, &i);
	if (i != 2)
		fprintf(stderr, 
				"WARNING: two columns expected, %d column(s) read.", i);
	// pre-calculate parameters
	options.sigma6 = -6.0 * power6(options.sigma);
	options.sigma12 = -12.0 * power12(options.sigma);
	options.resolution = (rdf[0][options.size-1] - rdf[0][0]) / options.size;
	printf("Resolution of RDF: %f\n", options.resolution);
	options.steps = (int) (options.cutoff / options.grid) + 1;
	printf("# of data points: %d\n", options.steps);
	if (debug)
		echo_options(options);
	// allocate memory
	snew(inner, options.steps);
	snew(outer, options.steps);
	for (i=0; i<options.steps; i++) {
		inner[i].force = 0.0;
		outer[i].force = 0.0;
	}
	if (bPLANE) {
		type = "LINEAR";
		plane_void(inner, outer, rdf, options);
	} else {
		if (bQUENCH)
			type = "QUENCH";
		else
			type = "SPHERE";
		if (options.outer != 0.0) {
			if (bKARPLUS)
				outer_void_karplus(outer, rdf, options);
			else
				mara_outer(outer, &slope_outer, rdf, options);
		}
		if (options.inner != 0.0)
			mara_inner(inner, &slope_inner, rdf, options);
	}
	options.slope_inner = slope_inner;
	options.slope_outer = slope_outer;
	write_sdb(opt2fn("-o", NFILE, fnm), inner, outer, options, type);
	// plot potentials?
	fn_plot = opt2fn_null("-p", NFILE, fnm);
	if (fn_plot != NULL) {
		plot = fopen(fn_plot, "w");
		for (i=0; i < options.steps; i++)
			fprintf(plot, "%f %f\n", inner[i].position, inner[i].force);
		for (i=0; i < options.steps; i++)
			fprintf(plot, "%f %f\n", outer[i].position, outer[i].force);
		fclose(plot);
	}
	return(0);
}

