/*
 * NAME
 *   g_vesicle_density
 *     Calculate spherical density profiles.
 * SYNOPSIS
 *   g_vesicle_density -s topol.tpr -f traj.xtc -o density.xvg
 * AUTHOR
 *   Martti Louhivuori (m.j.louhivuori@rug.nl)
 *
 * ADAPTED from g_density by Peter Tieleman
 * ---
 * Copyright (C) 2009  Martti Louhivuori
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * (version 2) as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
 * 02110-1301, USA.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <math.h>
#include <ctype.h>

#include "sysstuff.h"
#include "string.h"
#include "string2.h"
#include "typedefs.h"
#include "smalloc.h"
#include "macros.h"
#include "gstat.h"
#include "vec.h"
#include "xvgr.h"
#include "pbc.h"
#include "copyrite.h"
#include "futil.h"
#include "statutil.h"
#include "index.h"
#include "tpxio.h"
#include "physics.h"

static const char *authors[] = {
	"Written by Martti Louhivuori (m.j.louhivuori@rug.nl)",
	"Copyright (c) 2009  Martti Louhivuori"
};
static const char *gpl[] = {
	"This program is free software; you can redistribute it and/or",
	"modify it under the terms of the GNU General Public License",
	"(version 2) as published by the Free Software Foundation."
};
void sp_print(FILE *out, const char *txt)
{
	int i, s;
	s = (int) (80 - strlen(txt)) / 2.0;
	for (i=0; i < s; i++) 
		fprintf(out, " ");
	fprintf(out, "%s\n", txt);
}

typedef struct {
  char *atomname;
  int nr_el;
} t_electron;

/* used for sorting the list */
int compare(void *a, void *b)
{
  t_electron *tmp1,*tmp2;
  tmp1 = (t_electron *)a; tmp2 = (t_electron *)b;

  return strcmp(tmp1->atomname,tmp2->atomname);
}

int get_electrons(t_electron **eltab, char *fn)
{
  char buffer[256];  /* to read in a line   */
  char tempname[80]; /* buffer to hold name */
  int tempnr; 

  FILE *in;
  int nr;            /* number of atomstypes to read */
  int i;

  if ( !(in = fopen(fn,"r")))
    gmx_fatal(FARGS,"Couldn't open %s. Exiting.\n",fn);

  if(NULL==fgets(buffer, 255, in))
  {
      gmx_fatal(FARGS,"Error reading from file %s",fn);
  }
 
  if (sscanf(buffer, "%d", &nr) != 1)
    gmx_fatal(FARGS,"Invalid number of atomtypes in datafile\n");

  snew(*eltab,nr);

  for (i=0;i<nr;i++) {
    if (fgets(buffer, 255, in) == NULL)
      gmx_fatal(FARGS,"reading datafile. Check your datafile.\n");
    if (sscanf(buffer, "%s = %d", tempname, &tempnr) != 2)
      gmx_fatal(FARGS,"Invalid line in datafile at line %d\n",i+1);
    (*eltab)[i].nr_el = tempnr;
    (*eltab)[i].atomname = strdup(tempname);
  }
  
  /* sort the list */
  fprintf(stderr,"Sorting list..\n");
  qsort ((void*)*eltab, nr, sizeof(t_electron), 
	 (int(*)(const void*, const void*))compare);

  return nr;
}

void calculate_com (rvec *com, int n, atom_id index[], rvec x[], real m[])
{
	int i;
	real mass=0.0;
	rvec r;
	
	/* calculate the center of mass */
	(*com)[XX] = 0.0;
	(*com)[YY] = 0.0;
	(*com)[ZZ] = 0.0;
	for (i=0; i < n; i++) {
		svmul(m[index[i]], x[index[i]], r);
		rvec_inc(*com, r);
		mass += m[index[i]];
	}
	svmul(1.0 / mass, *com, *com);
}

void center_coords(t_atoms *atoms,matrix box,rvec x0[],int axis)
{
  int  i,m;
  real tmass,mm;
  rvec com,shift,box_center;
  
  tmass = 0;
  clear_rvec(com);
  for(i=0; (i<atoms->nr); i++) {
    mm     = atoms->atom[i].m;
    tmass += mm;
    for(m=0; (m<DIM); m++) 
      com[m] += mm*x0[i][m];
  }
  for(m=0; (m<DIM); m++) 
    com[m] /= tmass;
  calc_box_center(ecenterDEF,box,box_center);
  rvec_sub(box_center,com,shift);
  shift[axis] -= box_center[axis];
  
  for(i=0; (i<atoms->nr); i++) 
    rvec_dec(x0[i],shift);
}

void calc_electron_density(char *fn, atom_id **index, int gnx[], 
			   real ***slDensity, int *nslices, t_topology *top,
			   int ePBC,
			   int axis, int nr_grps, real *slWidth, 
			   t_electron eltab[], int nr,bool bCenter)
{
  rvec *x0;              /* coordinates without pbc */
  matrix box;            /* box (3x3) */
  int natoms,            /* nr. atoms in trj */
      status,  
      i,n,               /* loop indices */
      ax1=0, ax2=0,
      nr_frames = 0,     /* number of frames */
      slice;             /* current slice */
  t_electron *found;     /* found by bsearch */
  t_electron sought;     /* thingie thought by bsearch */
 
  real t, 
        z;

  switch(axis) {
  case 0:
    ax1 = 1; ax2 = 2;
    break;
  case 1:
    ax1 = 0; ax2 = 2;
    break;
  case 2:
    ax1 = 0; ax2 = 1;
    break;
  default:
    gmx_fatal(FARGS,"Invalid axes. Terminating\n");
  }

  if ((natoms = read_first_x(&status,fn,&t,&x0,box)) == 0)
    gmx_fatal(FARGS,"Could not read coordinates from statusfile\n");
  
  if (! *nslices)
    *nslices = (int)(box[axis][axis] * 10); /* default value */
  fprintf(stderr,"\nDividing the box in %d slices\n",*nslices);

  snew(*slDensity, nr_grps);
  for (i = 0; i < nr_grps; i++)
    snew((*slDensity)[i], *nslices);
  
  /*********** Start processing trajectory ***********/
  do {
    rm_pbc(&(top->idef),ePBC,top->atoms.nr,box,x0,x0);

    if (bCenter)
      center_coords(&top->atoms,box,x0,axis);
    
    *slWidth = box[axis][axis]/(*nslices);
    for (n = 0; n < nr_grps; n++) {      
      for (i = 0; i < gnx[n]; i++) {   /* loop over all atoms in index file */
	  z = x0[index[n][i]][axis];
	  while (z < 0) 
	    z += box[axis][axis];
	  while (z > box[axis][axis])
	    z -= box[axis][axis];
      
	  /* determine which slice atom is in */
	  slice = (z / (*slWidth)); 
	  sought.nr_el = 0;
	  sought.atomname = strdup(*(top->atoms.atomname[index[n][i]]));

	  /* now find the number of electrons. This is not efficient. */
	  found = (t_electron *)
	    bsearch((const void *)&sought,
		    (const void *)eltab, nr, sizeof(t_electron), 
		    (int(*)(const void*, const void*))compare);

	  if (found == NULL)
	    fprintf(stderr,"Couldn't find %s. Add it to the .dat file\n",
		    *(top->atoms.atomname[index[n][i]]));
	  else  
	    (*slDensity)[n][slice] += found->nr_el - 
	                              top->atoms.atom[index[n][i]].q;
	  free(sought.atomname);
	}
    }
      nr_frames++;
  } while (read_next_x(status,&t,natoms,x0,box));
  
  /*********** done with status file **********/
  close_trj(status);
  
/* slDensity now contains the total number of electrons per slice, summed 
   over all frames. Now divide by nr_frames and volume of slice 
*/

  fprintf(stderr,"\nRead %d frames from trajectory. Counting electrons\n",
	  nr_frames);

  for (n =0; n < nr_grps; n++) {
    for (i = 0; i < *nslices; i++)
      (*slDensity)[n][i] = (*slDensity)[n][i] * (*nslices) /
	( nr_frames * box[axis][axis] * box[ax1][ax1] * box[ax2][ax2]);
  }

  sfree(x0);  /* free memory used by coordinate array */
}

void calc_density(char *fn, atom_id **index, int sizes[], 
		real ***density, int *nshells, t_topology *top, int ePBC,
		int ng, real *width, bool bCenter, int size_com, 
		atom_id *index_com, bool bFraction)
{
	rvec *x;
	matrix box;
	int natoms, status;
	int i, n, nr_frames=0, s;
	real t, z, vol=1.0, *m;
	rvec com, shift, r;
	real *group_mass;

	/* read first frame */
	if ((natoms = read_first_x(&status, fn, &t, &x, box)) == 0)
		gmx_fatal(FARGS,"Could not read coordinates from statusfile\n");
	/* store masses */
	snew(m, top->atoms.nr);
	for (i=0; i < top->atoms.nr; i++)
		m[i] = top->atoms.atom[i].m;
	/* calculate number of shells? */
	if (! *nshells) 
		*nshells = (int)(box[XX][XX] * 5);
	*width = box[XX][XX] / (*nshells * 2.0);
	fprintf(stderr,"\nDividing the box in %d shells of %f width\n", 
			*nshells, *width);
	/* allocate memory for densities & calc. total mass of each group */
	snew(*density, ng);
	snew(group_mass, ng);
	for (n=0; n < ng; n++) {
		snew((*density)[n], *nshells);
		for (i=0; i < sizes[n]; i++)
			group_mass[n] += top->atoms.atom[index[n][i]].m;
	}
	/* loop through trajectory */
	do {
		rm_pbc(&(top->idef), ePBC, top->atoms.nr, box, x, x);
		/* shift CoM? */
		if (bCenter) 
			calculate_com(&com, size_com, index_com, x, m);
		else 
			calc_box_center(ecenterDEF, box, com);
		/* shift origo to zero */
		for (i=0; i < top->atoms.nr; i++) 
			rvec_dec(x[i], com);
		/* add-up each atom */
		for (n=0; n < ng; n++) {
			for (i=0; i < sizes[n]; i++) {
//				r = x[index[n][i]];
				s = (int)(norm(x[index[n][i]]) / *width);
				if (s >= *nshells)
					s = *nshells - 1;
				(*density)[n][s] += top->atoms.atom[index[n][i]].m;
			}
		}
		nr_frames++;
	} while ( read_next_x(status, &t, natoms, x, box) );
	close_trj(status);
	fprintf(stderr,"\nRead %d frames from trajectory. Calculating density\n",
			nr_frames);

	/* normalise densities */	
	for (i=0; i < *nshells; i++) {
		if (!bFraction)
			vol = (4 * M_PI / 3.0) * (3*i*i + 3*i + 1) * pow(*width, 3);
		for (n=0; n < ng; n++) {
			if (bFraction)
				vol = group_mass[n];
			(*density)[n][i] = (*density)[n][i] / (nr_frames * vol);
		}
	}
	sfree(x);
}

void plot_density(real *density[], char *afile, int nshells,
		  int ng, char *names[], real width, 
		  char **dens_opt, bool bFraction)
{
	FILE  *fp;
	char  *ylabel=NULL;
	
	int   s, n;
	real  d;

	switch (dens_opt[0][0]) {
		case 'm': 
			if (bFraction)
				ylabel = "Density fraction"; 
			else
				ylabel = "Density (kg m\\S-3\\N)";
			break;
		case 'n': ylabel = "Number density (nm\\S-3\\N)"; break;
		case 'c': ylabel = "Charge density (e nm\\S-3\\N)"; break;
		case 'e': ylabel = "Electron density (e nm\\S-3\\N)"; break;
	}
	
	fp = xvgropen(afile, "Partial densities", "Box (nm)", ylabel);
	xvgr_legend(fp, ng, names);
	
	for (s=0; s < nshells; s++) { 
		fprintf(fp, "%12g  ", s * width);
		for (n=0; n < ng; n++) {
			d = density[n][s];
			if (dens_opt[0][0] == 'm')
				fprintf(fp, "   %12g", d * AMU / (NANO*NANO*NANO));
			else
				fprintf(fp, "   %12g", d);
		}
		fprintf(fp, "\n");
	}
	fclose(fp);
}
 
int gmx_vesicle_density(int argc,char *argv[])
{
	static char *desc[] = {
		"Compute partial densities across the box, using an index file. Densities",
		"in kg/m^3, number densities or electron densities can be",
		"calculated. For electron densities, a file describing the number of",
		"electrons for each type of atom should be provided using [TT]-ei[tt].",
		"It should look like:[BR]",
		"   2[BR]",
		"   atomname = nrelectrons[BR]",
		"   atomname = nrelectrons[BR]",
		"The first line contains the number of lines to read from the file.",
		"There should be one line for each unique atom name in your system.",
		"The number of electrons for each atom is modified by its atomic",
		"partial charge.2"
	};
	
	static char *dens_opt[] = 
		{ NULL, "mass", "number", "charge", "electron", NULL };
	static int  nshells = 200;
	static int  ng   = 1;
	static bool bCenter=FALSE;
	static bool bFraction=FALSE;
	t_pargs pa[] = {
		{ "-nshells",  FALSE, etINT, {&nshells}, 
			"Divide the system in n shells." },
		{ "-density", FALSE, etENUM, {dens_opt},
			"Density"},
		{ "-ng",       FALSE, etINT, {&ng},
			"Number of groups to compute densities of" },
		{ "-center",  FALSE, etBOOL, {&bCenter},
			"Shift the center of mass along the axis to zero. This means if your axis is Z and your box is bX, bY, bZ, the center of mass will be at bX/2, bY/2, 0."}, 
		{ "-fraction", FALSE, etBOOL, {&bFraction}, 
			"Instead of normalising shell densities by the shell volume, normalise by the total mass of each group."}
	};
	
	static char *bugs[] = {
		"When calculating electron densities, atomnames are used instead of types. This is bad.",
	};
	
	t_topology *top; 
	int  ePBC;
	atom_id   **index;
	real **density;        /* density per slice           */
	real width;            /* width of one shell          */
	char **names;        /* group names                 */
	int  nr_electrons;     /* nr. electrons               */
	int  *sizes;             /* sizes of groups             */
	t_electron *el_tab;    /* table with nr. of electrons */
	int  i;
	atom_id *index_com;
	int size_com;
	char *name_com;
	
	t_filenm  fnm[] = {
		{ efTRX, "-f", NULL, ffREAD },
		{ efNDX, NULL, NULL, ffOPTRD }, 
		{ efTPX, NULL, NULL, ffREAD },
		{ efDAT, "-ei", "electrons", ffOPTRD },
		{ efXVG, "-o", "density", ffWRITE },
	};

#define NFILE asize(fnm)
	
	CopyRight(stderr, argv[0]);
	for (i=0; i < (int)asize(authors); i++)
		sp_print(stderr, authors[i]);
	fprintf(stderr, "\n");
	for (i=0; i < (int)asize(gpl); i++)
		sp_print(stderr, gpl[i]);
	fprintf(stderr, "\n");
	parse_common_args(&argc, argv, PCA_CAN_VIEW | PCA_CAN_TIME | PCA_BE_NICE,
			NFILE, fnm, asize(pa), pa, asize(desc), desc, asize(bugs), bugs);
	/* read topology */
	top = read_top(ftp2fn(efTPX, NFILE, fnm), &ePBC);
	/* number / charge density? modify masses of atoms */
	if (dens_opt[0][0] == 'n') {
		for (i=0; (i<top->atoms.nr); i++)
			top->atoms.atom[i].m = 1;  
	} else if (dens_opt[0][0] == 'c') {
		for (i=0; (i<top->atoms.nr); i++)
			top->atoms.atom[i].m = top->atoms.atom[i].q;  
	}
	/* read atom groups */
	if (bCenter) {
		printf("Select group for center-of-mass calculation\n");
		get_index(&top->atoms, ftp2fn_null(efNDX, NFILE, fnm), 1, &size_com, 
				&index_com, &name_com);
	}
	snew(names, ng);
	snew(index, ng);
	snew(sizes, ng);
	get_index(&top->atoms, ftp2fn_null(efNDX, NFILE, fnm), ng, sizes, 
			index, names);
	/* calculate density profile */
	if (dens_opt[0][0] == 'e') {
		/* electron density */
		/* FIXME: re-write!!! */
		fprintf(stderr, "WARNING: electron density calculation is NOT tested!");
		/*nr_electrons =  get_electrons(&el_tab, ftp2fn(efDAT, NFILE, fnm));
		fprintf(stderr, "Read %d atomtypes from datafile\n", nr_electrons);
		calc_electron_density(ftp2fn(efTRX, NFILE, fnm), index, ngx, 
				&density, &nshells, top, ePBC, ngrps, &width, el_tab, 
				nr_electrons, bCenter);*/
	} else
		calc_density(ftp2fn(efTRX, NFILE, fnm), index, sizes, &density, 
				&nshells, top, ePBC, ng, &width, bCenter, size_com, 
				index_com, bFraction); 
	/* write density profile */
	plot_density(density, opt2fn("-o", NFILE, fnm), nshells, ng, names, 
			width, dens_opt, bFraction);
	/* plot density profile */
	do_view(opt2fn("-o", NFILE, fnm), "-nxy");
	/* the end. */
	thanx(stderr);
	return 0;
}
