/*
 * NAME
 *   g_vesicle_flux
 *     Calculate ion fluxes through a lipid vesicle.
 * SYNOPSIS
 *   g_vesicle_flux -s topol.tpr -f traj.xtc -o flux.xvg
 * AUTHOR
 *   Martti Louhivuori (m.j.louhivuori@rug.nl)
 * 
 * ---
 * Copyright (C) 2008  Martti Louhivuori
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * (version 2) as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
 * 02110-1301, USA.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>
#include <ctype.h>

#include "macros.h"
#include "smalloc.h"
#include "typedefs.h"
#include "names.h"
#include "copyrite.h"
#include "statutil.h"
#include "tpxio.h"
#include "string2.h"
#include "strdb.h"
#include "vec.h"
#include "macros.h"
#include "index.h"
#include "pbc.h"
#include "xvgr.h"
#include "futil.h"
#include "matio.h"

#define TIME_EXPLICIT 0
#define TIME_CONTINUE 1
#define TIME_LAST     2
#ifndef FLT_MAX
#define FLT_MAX 1e36
#endif
#define FLAGS (TRX_READ_X | TRX_READ_V | TRX_READ_F)

static const char *authors[] = {
	"Written by Martti Louhivuori (m.j.louhivuori@rug.nl)",
	"Copyright (c) 2008  Martti Louhivuori"
};
static const char *gpl[] = {
	"This program is free software; you can redistribute it and/or",
	"modify it under the terms of the GNU General Public License",
	"(version 2) as published by the Free Software Foundation."
};
void sp_print(FILE *out, const char *txt)
{
	int i, s;
	s = (int) (80 - strlen(txt)) / 2.0;
	for (i=0; i < s; i++) 
		fprintf(out, " ");
	fprintf(out, "%s\n", txt);
}


void calculate_com (rvec *com, int n, atom_id index[], rvec x[])
{
	int i;
	
	/* calculate the center of mass */
	/*   TODO: actually use masses! */
	(*com)[XX] = 0.0;
	(*com)[YY] = 0.0;
	(*com)[ZZ] = 0.0;
	for (i=0; i < n; i++) {
		rvec_add(*com, x[index[i]], *com);
	}
	svmul(1.0 / n, *com, *com);
}

real calculate_avg_radius (int n, atom_id index[], rvec x[], rvec com)
{
	rvec v;
	real r=0.0;
	int i;

	/* calculate the average distance to CoM */
	for (i=0; i < n; i++) {
		rvec_sub(x[index[i]], com, v);
		r = r + norm(v);
	}
	r = r / n;
	return r;
}

real calculate_variance (int n, atom_id index[], rvec x[], rvec com, real r)
{
	rvec v;
	real e=0.0;
	int i;

	/* calculate the average distance to CoM */
	for (i=0; i < n; i++) {
		rvec_sub(x[index[i]], com, v);
		e += pow(norm(v) - r, 2);
	}
	return sqrt(e / n);
}

int calculate_flux (int *in, int *out, int n, atom_id index[], rvec com, 
		real r, rvec x[], bool bInverse)
{
	int i, prev, flux;
	real s;
	rvec v;

	/* previous number of ions inside */
	prev = *in;
	
	/* count ions */
	*in = 0;
	*out = 0;
	for (i=0; i < n; i++) {
		rvec_sub(x[index[i]], com, v);
		s = norm(v);
		if (s > r) 
			(*out)++;
		else 
			(*in)++;
	}

	/* invert flux direction? */
	if (bInverse)
		flux = (*in) - prev;
	else
		flux = prev - (*in);

	return flux;
}

void update_position (int *position, int *in, int *channel, int *out, 
		int size, atom_id index[], rvec x[], rvec com, real radius, 
		real variance, bool verbose) {
	int j, i=0, c=0, o=0;
	real lower, upper, r;
	rvec v;

	lower = radius - variance;
	upper = radius + variance;
	for (j=0; j < size; j++) {
		rvec_sub(x[index[j]], com, v);
		r = norm(v);
		if (r > lower && r < upper) {
			position[j] = 0;
			c++;
		} else if (r > upper) {
			position[j] = 1;
			o++;
		} else {
			position[j] = -1;
			i++;
		}
	}
	if (verbose) {
		fprintf(stdout, "\n[inside ]: %d\n", i);
		fprintf(stdout, "[channel]: %d\n", c);
		fprintf(stdout, "[outside]: %d\n", o);
	}
	*in = i;
	*channel = c;
	*out = o;
}

void update_side (int *side, int *position, int size) {
	int i;

	for (i=0; i < size; i++) {
		if (position[i] == 1) 
			side[i] = 1;
		else if (position[i] == -1) 
			side[i] = -1;
		else if (position[i] != 0)
			side[i] = 0;
	}
}

void check_transitions (int *df, int *transitions, int *ups, int *downs, 
		int *side, int *position, int size_ions, atom_id index_ions[], 
		bool bProtein, int size_protein, atom_id index_protein[], rvec x[], 
		real cutoff) {
	int i, j, flux=0, count=0;
	bool accept=TRUE;
	rvec r;

	for (i=0; i < size_ions; i++) {
		if (bProtein && position[i] != side[i] && position[i] != 0) {
			accept = FALSE;
			for (j=0; j < size_protein; j++) {
				rvec_sub(x[index_protein[j]], x[index_ions[i]], r);
				if (norm(r) < cutoff) {
					accept = TRUE;
					break;
				}
			}
		} else {
			accept = TRUE;
		}
		if (accept) {
			if (position[i] == 1 && side[i] == -1) {
				ups[i] += 1;
				flux += 1;
				count++;
			} else if (position[i] == -1 && side[i] == 1) {
				downs[i] += 1;
				flux -= 1;
				count++;
			}
		}
	}
	*df = flux;
	*transitions = count;
}

/*** MULTIPLE TRAJECTORIES ***/
static void scan_trj_files(char **fnms,int nfiles,
		real *readtime,real *timestep,atom_id imax)
{
	/* Check start time of all files */
	int i,status,natoms=0;
	real t;
	t_trxframe fr;
	bool ok;

	for(i=0;i<nfiles;i++) {
		ok=read_first_frame(&status,fnms[i],&fr,FLAGS);

		if(!ok) 
			gmx_fatal(FARGS,"\nCouldn't read frame from file.");
		if(fr.bTime)
			readtime[i]=fr.time;
		else {
			readtime[i]=0;
			fprintf(stderr,"\nWARNING: Couldn't find a time in the frame.\n");
		}

		if(i==0) {
			natoms=fr.natoms;
		}
		else {
			if (imax==NO_ATID) {
				if(natoms!=fr.natoms) 
					gmx_fatal(FARGS,"\nDifferent numbers of atoms (%d/%d) in files",
							natoms,fr.natoms);
			} else {
				if(fr.natoms <= imax)
					gmx_fatal(FARGS,"\nNot enough atoms (%d) for index group (%d)",
							fr.natoms,imax);
			}
		}
		ok=read_next_frame(status,&fr);
		if(ok && fr.bTime) {
			timestep[i] = fr.time - readtime[i];
		} else {
			timestep[i] = 0;
		}

		close_trj(status);
	}
	fprintf(stderr,"\n");

	sfree(fr.x);
}


static void sort_files(char **fnms,real *settime,int nfile)
{
	int i,j,minidx;
	real timeswap;
	char *chptr;

	for(i=0;i<nfile;i++) {
		minidx=i;
		for(j=i+1;j<nfile;j++) {
			if(settime[j]<settime[minidx])
				minidx=j;
		}
		if(minidx!=i) {
			timeswap=settime[i];
			settime[i]=settime[minidx];
			settime[minidx]=timeswap;
			chptr=fnms[i];
			fnms[i]=fnms[minidx];
			fnms[minidx]=chptr;
		}
	}
}


static void edit_files(char **fnms,int nfiles,real *readtime, real *timestep,
		real *settime, int *cont_type, bool bSetTime,bool bSort)
{
	int i;
	bool ok;
	char inputstring[STRLEN],*chptr;

	if(bSetTime) {
		fprintf(stderr,"\n\nEnter the new start time (%s) for each file.\n"
				"There are two special options, both disable sorting:\n\n"
				"c (continue) - The start time is taken from the end\n"
				"of the previous file. Use it when your continuation run\n"
				"restarts with t=0.\n\n"
				"l (last) - The time in this file will be changed the\n"
				"same amount as in the previous. Use it when the time in the\n"
				"new run continues from the end of the previous one,\n"
				"since this takes possible overlap into account.\n\n",
				time_unit() );

		fprintf(stderr,
				"          File             Current start (%s)  New start (%s)\n"
				"---------------------------------------------------------\n",
				time_unit(), time_unit() );

		for(i=0;i<nfiles;i++) {
			fprintf(stderr,"%25s   %10.3f %s          ",
					fnms[i],convert_time(readtime[i]), time_unit());
			ok=FALSE;
			do {
				if(NULL==fgets(inputstring,STRLEN-1,stdin))
				{
					gmx_fatal(FARGS,"Error reading user input");
				}

				inputstring[strlen(inputstring)-1]=0;

				if(inputstring[0]=='c' || inputstring[0]=='C') {
					cont_type[i]=TIME_CONTINUE;
					bSort=FALSE;
					ok=TRUE;
					settime[i]=FLT_MAX;
				}
				else if(inputstring[0]=='l' ||
						inputstring[0]=='L') {
					cont_type[i]=TIME_LAST;
					bSort=FALSE;
					ok=TRUE;
					settime[i]=FLT_MAX;			  
				}
				else {
					settime[i]=strtod(inputstring,&chptr)*time_invfactor();
					if(chptr==inputstring) {
						fprintf(stderr,"'%s' not recognized as a floating point number, 'c' or 'l'. "
								"Try again: ",inputstring);
					}
					else {
						cont_type[i]=TIME_EXPLICIT;
						ok=TRUE;
					}
				}
			} while (!ok);
		}
		if(cont_type[0]!=TIME_EXPLICIT) {
			cont_type[0]=TIME_EXPLICIT;
			settime[0]=0;
		}
	}
	else 
		for(i=0;i<nfiles;i++)
			settime[i]=readtime[i];

	if(!bSort) 
		fprintf(stderr,"Sorting disabled.\n");
	else 
		sort_files(fnms,settime,nfiles);


	/* Write out the new order and start times */
	fprintf(stderr,"\nSummary of files and start times used:\n\n"
			"          File                Start time       Time step\n"
			"---------------------------------------------------------\n");
	for(i=0;i<nfiles;i++)
		switch(cont_type[i]) {
			case TIME_EXPLICIT:
				fprintf(stderr,"%25s   %10.3f %s   %10.3f %s",
						fnms[i],
						convert_time(settime[i]),time_unit(),
						convert_time(timestep[i]),time_unit());
				if ( i>0 && 
						cont_type[i-1]==TIME_EXPLICIT && settime[i]==settime[i-1] )
					fprintf(stderr," WARNING: same Start time as previous");
				fprintf(stderr,"\n");
				break;
			case TIME_CONTINUE:
				fprintf(stderr,"%25s        Continue from last file\n",fnms[i]);
				break;	      
			case TIME_LAST:
				fprintf(stderr,"%25s        Change by same amount as last file\n",
						fnms[i]);
				break;
		}
	fprintf(stderr,"\n");

	settime[nfiles]=FLT_MAX;
	cont_type[nfiles]=TIME_EXPLICIT;
	readtime[nfiles]=FLT_MAX;
}
/*** MULTIPLE TRAJECTORIES ***/


int gmx_vesicle_flux (int argc, char *argv[])
{
	static char *desc[] = {
		"Calculate ion fluxes through lipid vesicles with or without ",
		"channel proteins.[PAR]",
		"For each frame calculates the average radius of the vesicle ",
		"and checks the number of ions inside and outside of this sphere. ",
		"The flux is then determined as the increase/decrease of inside ", 
		"ions from the previous time step. Positive flux direction is by ",
		"default from inside to outside, but this can be changed by ", 
		"[TT]-inverse[tt]. The average radius of the vesicle is ", 
		"calculated as a mass weighted average over all atoms selected ", 
		"interactively by the user. The ion group is also selected by the ",
		"user."
	};
	static bool bInverse=FALSE;
	bool        bFirst=TRUE, bEvents=FALSE, bCount=FALSE;
	bool        bProtein=FALSE, bBuffer=TRUE, verbose=FALSE, bIndex=FALSE;
	t_topology  top;
	t_atoms    *atoms;
	matrix      box;
	char        buf[255];
	rvec       *x, com;
	int         i, ePBC, status, natom, in=0, out=0, channel=0, 
				first=0;
	int         size_vesicle, size_ions, size_protein;
	int         df=0, flux=0;
	real        sigma=1.0, cutoff=2.4;
	atom_id    *index_vesicle, *index_ions, *index_protein;
	char       *grpname_vesicle, *grpname_ions, *grpname_protein;
	real        t, t0=0.0, r, e;
	FILE       *fp;
	int        *side, *position, *location, *ups, *downs, count=0;
	int        *histo, lo=0, hi=0;
	ivec       *counts;
	/* for index file */
	char     **gnames;
	t_blocka  *block;
	atom_id   *index_trans;
	/* multi-trajectory */
	char       **filenames, tmp[1];
	real         last_ok_t=0.0, begin=-1.0, end=-1.0, shift=0.0;
	static bool  bSort=TRUE;
	static bool  bKeepLast=FALSE;
	static bool  bSetTime=FALSE;
	int          n, *cont_type, frame=0, frame_out=0;
	real        *readtime, *timest, *settime, dt=0.0, lasttime=NOTSET, timestep;
	bool         bNewFile=FALSE, bAnalyse=TRUE;

	char *keys[] = {"Particle flux", "Total flux", "Transitions",  
		"Inside", "Channel", "Outside", "Radius", "Variance"};
	char *keys_events[] = {"# of particles"};
	char *keys_count[] = {"in->out", "out->in"};
	
	t_pargs pa[] = {
		{ "-b",       FALSE, etTIME, {&begin},
			"First time to use (%t)"},
		{ "-e",       FALSE, etTIME, {&end},
			"Last time to use (%t)"},
		{ "-dt",      FALSE, etTIME, {&dt},
			"Only write frame when t MOD dt = first time (%t)" },
		{ "-inverse", TRUE, etBOOL, {&bInverse}, 
			"inverse flux direction to be from outside to inside" },
		{ "-buffer", TRUE, etBOOL, {&bBuffer}, 
			"use a buffer zone to remove rapid flickering across the decision boundary" },
		{ "-sigma", FALSE, etREAL, {&sigma}, 
			"relative size of the buffer zone"}, 
		{ "-protein", TRUE, etBOOL, {&bProtein}, 
			"consider only flux close to a memrane protein" },
		{ "-cutoff", FALSE, etREAL, {&cutoff}, 
			"cutoff for being in the neighbourhood of the protein"}, 
		{ "-settime", FALSE, etBOOL, {&bSetTime}, 
			"Change starting time interactively" },
		{ "-sort",    FALSE, etBOOL, {&bSort},
			"Sort trajectory files (not frames)" },
		{ "-keeplast",FALSE, etBOOL, {&bKeepLast},
			"keep overlapping frames at end of trajectory" },
		{ "-v", FALSE, etBOOL, {&verbose}, 
			"be loud and noisy"}
	};
	t_filenm fnm[] = {
		{ efTRX, "-f", NULL, ffRDMULT },
		{ efTPS, "-s", NULL, ffREAD },
		{ efNDX, NULL, NULL, ffOPTRD },
		{ efXVG, NULL, "flux", ffWRITE },
		{ efNDX, "-no", "flux", ffOPTWR },
		{ efXVG, "-oe", "events", ffOPTWR },
		{ efXVG, "-oc", "count", ffOPTWR },
	};
#define NFILE asize(fnm)
	CopyRight(stderr, argv[0]);
	for (i=0; i < (int)asize(authors); i++)
		sp_print(stderr, authors[i]);
	fprintf(stderr, "\n");
	for (i=0; i < (int)asize(gpl); i++)
		sp_print(stderr, gpl[i]);
	fprintf(stderr, "\n");
	parse_common_args(&argc, argv, PCA_CAN_VIEW | PCA_CAN_TIME | PCA_BE_NICE, 
			NFILE, fnm, asize(pa), pa, asize(desc), desc, 0, NULL);

	/* write index? */
	bIndex = opt2bSet("-no", NFILE, fnm);
	/* write events? */
	bEvents = opt2bSet("-oe", NFILE, fnm);
	/* write flux counts? */
	bCount = opt2bSet("-oc", NFILE, fnm);
	
	/* read topology */
	read_tps_conf(ftp2fn(efTPS, NFILE, fnm), buf, &top, &ePBC, &x, NULL, 
			box, FALSE);
	atoms = &(top.atoms);
	/* read vesicle and ion groups */
	printf("Select group for the vesicle\n");
	get_index(atoms, ftp2fn_null(efNDX, NFILE, fnm), 1, &size_vesicle, 
			&index_vesicle, &grpname_vesicle);
	printf("Select group for the ions to monitor\n");
	get_index(atoms, ftp2fn_null(efNDX, NFILE, fnm), 1, &size_ions, 
			&index_ions, &grpname_ions);
	/* read protein group? */
	if (bProtein) {
		fprintf(stderr, "Select group for the protein\n");
		get_index(atoms, opt2fn_null("-n", NFILE, fnm), 1, &size_protein, 
				&index_protein, &grpname_protein);
	}
	fgets(tmp, 1, stdin);
	
	/* write header */
	fp = xvgropen(ftp2fn(efXVG, NFILE, fnm), 
			"Particle flux", "Time (ps)", "#");
	xvgr_legend(fp, 8, keys);

	/* allocate memory */
	side = malloc(size_ions * sizeof(int) + 1);
	position = malloc(size_ions * sizeof(int) + 1);
	location = malloc(size_ions * sizeof(int) + 1);
	ups = malloc(size_ions * sizeof(int) + 1);
	downs = malloc(size_ions * sizeof(int) + 1);
	/* initialise, just to be safe */
	for (i=0; i < size_ions; i++) {
		side[i] = 0;
		position[i] = 0;
		location[i] = 0;
		ups[i] = 0;
		downs[i] = 0;
	}

	/* multi-trajectory */
	n = opt2fns(&filenames, "-f", NFILE, fnm);
	snew(readtime, n+1);
	snew(timest, n+1);
	scan_trj_files(filenames, n, readtime, timest, atoms->nr - 1);
	snew(settime, n+1);
	snew(cont_type, n+1);
	edit_files(filenames, n, readtime, timest, settime, cont_type, bSetTime, 
			bSort);
	if (debug) {
		for (i=0; i<n; i++) {
			fprintf(stderr, "readtime[%d]=%g  timest[%d]=%g\n", i, 
					readtime[i], i, timest[i]);
			fprintf(stderr, "settime[%d]=%g  cont_type[%d]=%d\n", i, 
					settime[i], i, cont_type[i]);
		}
	}
    
	timestep = timest[0];
	/* read all trajectories */
    for (i=0; i<n; i++) {
		/* set the next time from the last frame in previous file */
		if (i > 0) {
			bNewFile=TRUE;
			if (frame_out >= 0) {
				if (cont_type[i] == TIME_CONTINUE) {
					begin = t; //frout.time;
					begin += 0.5*timestep;
					settime[i] = t; //frout.time;
					cont_type[i] = TIME_EXPLICIT;
				}
				else if (cont_type[i] == TIME_LAST) {
					begin = t; //frout.time;
					begin += 0.5*timestep;
				} 
			}
			if (cont_type[i] == TIME_EXPLICIT) 
				if ( (i < n) && (t < settime[i]-1.5*timestep) ) 
					fprintf(stderr, "WARNING: Frames around t=%f %s have a "
							"different spacing than the rest,\n"
							"might be a gap or overlap that couldn't be "
							"corrected automatically.\n", convert_time(t), 
							time_unit());
		}
		/* if we don't have a timestep in the current file, use the old one */
		if (timest[i] != 0)
			timestep = timest[i];

		/* echo final time of last frame? */ 
		if (i != 0 && lasttime != NOTSET)
			printf("\nStopped at frame %d, time %g.\n", frame, lasttime);
		/* read first frame */
		natom = read_first_x(&status, filenames[i], &t, &x, box);
		/* store initial time */
		if (i == 0) 
			t0 = t;
		/* shift times? */
		if (cont_type[i] == TIME_EXPLICIT) 
			shift = settime[i] - t;
		else
			shift = 0.0;
		/* main loop */
		do {
			t += shift;
			/* stop if we have reached the end */
			if ( (end > 0) && (t > end+GMX_REAL_EPS) ) {
				i=n;
				break;
			}
			/* analyse this frame? */
			if ( bKeepLast ) 
				bAnalyse = ( t > lasttime + 0.5*timestep );
			else 
				bAnalyse = ( t < settime[i+1] - 0.5*timestep );
			/* analyse only selected frames */
			if ( bAnalyse && (t >= begin) ) {
				frame++;
				lasttime = t;
				if (dt==0 || bRmod(t, t0, dt)) {
					frame_out++;
					last_ok_t = t;
					if (bNewFile) {
						fprintf(stderr, 
								"\nContinue analysing frames from %s t=%g %s, "
								"frame=%d      \n", filenames[i], 
								convert_time(t), time_unit(), frame);
						bNewFile=FALSE;
					}
					if ( ((frame % 10) == 0) || (frame < 10) )
						fprintf(stderr," ->  frame %6d time %8.3f %s     \r",
								frame_out, convert_time(t), time_unit());
				} else 
					bAnalyse = FALSE;
			}
			if (bAnalyse) {
				/* calculate position and size of vesicle */
				put_atoms_in_box(box, natom, x);
				calculate_com(&com, size_vesicle, index_vesicle, x);
				r = calculate_avg_radius(size_vesicle, index_vesicle, x, com);
				if (bBuffer)
					e = calculate_variance(size_vesicle, index_vesicle, x, 
							com, r);
				else
					e = 0.0;
				e *= sigma;
				if (debug)
					fprintf(stderr, "r=%g  e=%g\n", r, e);
				update_position(position, &in, &channel, &out, size_ions, 
						index_ions, x, com, r, e, verbose);
				if (bFirst) {
					df = 0;
					bFirst = FALSE;
				} else {
					check_transitions(&df, &count, ups, downs, side, position, 
							size_ions, index_ions, bProtein, size_protein, 
							index_protein, x, cutoff);
				}
				flux += df;
				fprintf(fp, "%g  %d  %d  %d  %d  %d  %d  %g  %g\n", t, df, 
						flux, count, in, channel, out, r, e);
				update_side(side, position, size_ions);
			}
		} while (read_next_x(status, &t, natom, x, box));
		close_trj(status);
	}  
	fclose(fp);
	/* echo final time of last frame? */ 
	if (lasttime != NOTSET)
		printf("\nStopped at frame %d, time %g.\n", frame, lasttime);
	/* echo final frame used */
    fprintf(stderr, "\nLast frame analysed was %d, time %f %s\n",
	    frame, convert_time(last_ok_t), time_unit());

	/* write an index file of fluxed ions? */
	if (bIndex) {
		snew(gnames,1);
		block = new_blocka();
		analyse(atoms, block, &gnames, FALSE, FALSE);
		
		snew(index_trans, size_ions);
		count=0;
		for (i=0; i < size_ions; i++) 
			if (ups[i] > 0)
				index_trans[count++] = index_ions[i];
		add_grp(block, &gnames, count, index_trans, "Fluxed+");
		for (i=0; i < size_ions; i++)
			index_trans[i] = 0;
		count=0;
		for (i=0; i < size_ions; i++) 
			if (downs[i] > 0)
				index_trans[count++] = index_ions[i];
		add_grp(block, &gnames, count, index_trans, "Fluxed-");
		
		write_index(opt2fn("-no", NFILE, fnm), block, gnames);
	}
	/* write flux event histogram? */
	if (bEvents) {
		for (i=0; i < size_ions; i++) {
			if (ups[i] > hi) 
				hi = ups[i];
			if (downs[i] > lo)
				lo = downs[i];
		}
		if (debug)
			fprintf(stderr, "hi=%d  lo=%d\n", hi, lo);
		if (hi != 0 || lo != 0) {
			snew(histo, hi + lo + 1);
			for (i=0; i < size_ions; i++) {
				if (ups[i] > 0)
					histo[lo + ups[i]] += 1;
				if (downs[i] > 0)
					histo[lo - downs[i]] += 1;
				if (ups[i] == 0 && downs[i] == 0)
					histo[lo] += 1;
			}
			/* write header */
			fp = xvgropen(opt2fn("-oe", NFILE, fnm),
				"Flux event distribution", "# of times fluxed", "# of particles");
			xvgr_legend(fp, 1, keys_events);
			for (i=0; i < hi + lo + 1; i++) 
				fprintf(fp, "%d  %d\n", i - lo, histo[i]);
			fclose(fp);
			free(histo);
		} else {
			fprintf(stderr, "No flux events. Skipping histogram.\n");
		}
	}
	/* write flux counts? */
	if (bCount) {
		fp = xvgropen(opt2fn("-oc", NFILE, fnm),
			"Count of flux events", "particle", "# of flux events");
		xvgr_legend(fp, 2, keys_count);
		for (i=0; i < size_ions; i++) {
			if (ups[i] > 0 || downs[i] > 0)
				fprintf(fp, "%d  %d  %d\n", index_ions[i], ups[i], downs[i]);
		}
		fclose(fp);
	}

	fprintf(stderr, "\n");
	fprintf(stdout, "Total flux: %d in %g ps\n", flux, (t - t0));
	fprintf(stdout, "Flux rate: %g / ns\n", flux / (t - t0) * 1000.0);

	/* free memory */
	free(side);
	free(position);
	free(location);
	free(ups);
	free(downs);
	/* close down */
	thanx(stderr);
	return 0;
}

