/*
 * NAME
 *   g_vesicle_pressure
 *     Calculate the internal pressure of a vesicle.
 * SYNOPSIS
 *   g_vesicle_pressure -s topol.tpr -f traj.xtc
 * AUTHOR
 *   Martti Louhivuori (m.j.louhivuori@rug.nl)
 * 
 * ---
 * Copyright (C) 2009  Martti Louhivuori
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * (version 2) as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
 * 02110-1301, USA.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "smalloc.h"
#include "typedefs.h"
#include "copyrite.h"
#include "statutil.h"
#include "tpxio.h"
#include "string2.h"
#include "vec.h"
#include "index.h"
#include "confio.h"

static const char *authors[] = {
	"Written by Martti Louhivuori (m.j.louhivuori@rug.nl)",
	"Copyright (c) 2009  Martti Louhivuori"
};
static const char *gpl[] = {
	"This program is free software; you can redistribute it and/or",
	"modify it under the terms of the GNU General Public License",
	"(version 2) as published by the Free Software Foundation."
};
void sp_print(FILE *out, const char *txt)
{
	int i, s;
	s = (int) (80 - strlen(txt)) / 2.0;
	for (i=0; i < s; i++) 
		fprintf(out, " ");
	fprintf(out, "%s\n", txt);
}


void calculate_com (rvec *com, int n, atom_id index[], rvec x[])
{
	int i;
	
	/* calculate the center of mass */
	/*   TODO: actually use masses! */
	(*com)[XX] = 0.0;
	(*com)[YY] = 0.0;
	(*com)[ZZ] = 0.0;
	for (i=0; i < n; i++) {
		rvec_add(*com, x[index[i]], *com);
	}
	svmul(1.0 / n, *com, *com);
}

real calculate_avg_radius (int n, atom_id index[], rvec x[], rvec com)
{
	rvec v;
	real r=0.0;
	int i;

	/* calculate the average distance to CoM */
	for (i=0; i < n; i++) {
		rvec_sub(x[index[i]], com, v);
		r = r + norm(v);
	}
	r = r / n;
	return r;
}

int read_pgrid (char *filename, matrix **output, ivec *dimensions)
{
	FILE *fp;
	bool bDouble;
    double box[DIM][DIM];
	float  buffer[DIM][DIM];
	int    n, i, j, k, x, y, z;
	matrix *grid;
	ivec   size;

	// no file to open?
	if (filename == NULL)
		return -1;
	// try to open file
	fp = fopen(filename, "r");
	if (fp == NULL) {
		fprintf(stderr, "ERROR: unable to read XYZ file '%s'\n", filename);
		exit(1);
	}
	// read meta-data
	fread(&bDouble, sizeof(int), 1, fp);
	if (bDouble)
		fread(box, sizeof(double), DIM*DIM, fp);
	else {
		fread(buffer, sizeof(float), DIM*DIM, fp);
		for (i=0; i < DIM; i++)
			for (j=0; j < DIM; j++)
				box[i][j] = buffer[i][j];
	}
	fread(size, sizeof(int), DIM, fp);
	// number of grid elements
	n = size[XX] * size[YY] * size[ZZ];
	// allocate memory
	grid = malloc(sizeof(matrix) * n);
	// read pressure field
	if (bDouble)
		fread(grid, sizeof(matrix), n, fp);
	else {
		for (i=0; i < n; i++) {
			fread(buffer, sizeof(float), DIM*DIM, fp);
			for (j=0; j < DIM; j++)
				for(k=0; k < DIM; k++)
					grid[i][j][k] = buffer[j][k];
		}
	}
	fclose(fp);
	// assign grid pointer
	*output = grid;
	// hack to copy grid size
	clear_ivec(*dimensions);
	ivec_add(*dimensions, size, *dimensions);
	return n;
}

int map (rvec x, real resolution, ivec size)
{
	int i, j, k;

	i = (int) (x[XX] / resolution);
	j = (int) (x[YY] / resolution);
	k = (int) (x[ZZ] / resolution);
	return (i*size[YY]*size[ZZ] + j*size[ZZ] + k);
}
int flatten (int i, int j, int k, ivec size)
{
	return (i*size[YY]*size[ZZ] + j*size[ZZ] + k);
}

void write_mask(char *filename, int *mask, int count, ivec dimensions, 
		real resolution)
{
	int node, i, j, k, n=0;
	float xx, yy, zz;
	FILE *fp;

	fp = fopen(filename, "w");
	fprintf(fp, "3D GRID MASK\n");
	fprintf(fp, "  %d\n", count);
	for (i=0; i < dimensions[XX]; i++) {
		xx = i * resolution;
		for (j=0; j < dimensions[YY]; j++) {
			yy = j * resolution;
			for (k=0; k < dimensions[ZZ]; k++) {
				zz = k * resolution;
				node = flatten(i, j, k, dimensions);
				if (!mask[node]) {
					fprintf(fp, "    1X     MASK%5d %3.3f %3.3f %3.3f\n", 
							i+1, xx, yy, zz);
					n++;
				}
			}
		}
	}
	fprintf(fp, "%.3f %.3f %.3f\n", dimensions[XX] * resolution, 
			dimensions[YY] * resolution, dimensions[ZZ] * resolution);
	fclose(fp);
}

int gmx_vesicle_pressure (int argc, char *argv[])
{
	static char *desc[] = {
		"Calculate the average radius of a lipid vesicle.[PAR]",
		"For each frame calculates the average distance from the ", 
		"selected atoms to their center, i.e. the average radius of the ",
		"vesicle.",
	};
	bool        verbose=FALSE, bMask=FALSE;
	int         fudge=0;
	real        cavity=0.0;
	t_topology  top;
	t_atoms    *atoms, dummies;
	t_atom      dummy;
	char        *name, **names;
	int        *types;
	matrix      box, *grid, avg;
	int         *mask;
	ivec        dimensions;
	int         n, i, j, k, l, shift, node, count=0;
	char        buf[255];
	rvec       *x, com, s;
	real        xx, yy, zz, a, e;
	int         ePBC, status, natom, size;
	atom_id    *index;
	char       *grpname;
	real        t, r, resolution=0.4;
	FILE       *fp;
	
	char *keys[] = {"Radius", "Std.dev."};
	char *keys3[] = {"Radius", "Std.dev.", "Std.dev. of radii"};
	
	t_pargs pa[] = {
		{ "-resolution", TRUE, etREAL, {&resolution}, 
			"spatial resolution of the pressure field (in nm)" }, 
		{ "-fudge", TRUE, etINT, {&fudge}, 
			"blur the water-lipid interface by N elements"}, 
		{ "-cavity", TRUE, etREAL, {&cavity}, 
			"ignore internal cavity of this radius (in nm)"},
		{ "-v", TRUE, etBOOL, {&verbose}, 
			"be loud and noisy" }
	};
	t_filenm fnm[] = {
		{ efTRX, "-f", NULL, ffREAD },
		{ efTPS, "-s", NULL, ffREAD },
		{ efNDX, NULL, NULL, ffOPTRD },
		{ efDAT, "-d", "local_pressure", ffREAD },
		{ efGRO, "-o", "mask", ffOPTWR },
	};
#define NFILE asize(fnm)
	CopyRight(stderr, argv[0]);
	for (i=0; i < (int)asize(authors); i++)
		sp_print(stderr, authors[i]);
	fprintf(stderr, "\n");
	for (i=0; i < (int)asize(gpl); i++)
		sp_print(stderr, gpl[i]);
	fprintf(stderr, "\n");
	parse_common_args(&argc, argv, PCA_CAN_VIEW | PCA_CAN_TIME | PCA_BE_NICE, 
			NFILE, fnm, asize(pa), pa, asize(desc), desc, 0, NULL);
	
	bMask = opt2bSet("-o", NFILE, fnm);
	/* read topology */
	read_tps_conf(ftp2fn(efTPS, NFILE, fnm), buf, &top, &ePBC, &x, NULL, 
			box, FALSE);
	atoms = &(top.atoms);
	/* read vesicle group */
	printf("Select group for the vesicle\n");
	get_index(atoms, ftp2fn_null(efNDX, NFILE, fnm), 1, &size, 
			&index, &grpname);
	/* read pressure data */
	n = read_pgrid(ftp2fn(efDAT, NFILE, fnm), &grid, &dimensions);
	if (debug) 
		fprintf(stderr, "grid n=%d dimensions=(%d,%d,%d)\n", n, dimensions[XX], dimensions[YY], dimensions[ZZ]);
	/* allocate memory */
	snew(mask, n);
	/* read first frame */
	natom = read_first_x(&status, ftp2fn(efTRX, NFILE, fnm), &t, &x, box);
	calculate_com(&com, size, index, x);
	r = calculate_avg_radius(size, index, x, com);
	if (debug) 
		fprintf(stderr, "\nr=%g natom=%d com=(%g,%g,%g)\n", r, natom, com[XX], com[YY], com[ZZ]);
	/* mask outer solvent and internal cavity */
	for (i=0; i < dimensions[XX]; i++) {
		xx = i * resolution;
		for (j=0; j < dimensions[YY]; j++) {
			yy = j * resolution;
			for (k=0; k < dimensions[ZZ]; k++) {
				zz = k * resolution;
				s[XX] = xx;
				s[YY] = yy;
				s[ZZ] = zz;
				rvec_dec(s, com);
				if (norm(s) > r || norm(s) < cavity) {
					node = flatten(i, j, k, dimensions);
					mask[node] = 1;
				}
			}
		}
	}
	/* main loop */
	do {
		for (i=0; i < size; i++) {
			node = map(x[index[i]], resolution, dimensions);
			mask[node] = 1;
		}
	} while (read_next_x(status, &t, natom, x, box));
	/* fudge edges? */
	if (fudge > 0) {
		for (l=0; l < n; l++) {
			if (!mask[l]) {
				for (i=-fudge; i <= fudge; i++) {
					shift = i * dimensions[YY] * dimensions[ZZ];
					for (j=-fudge; j <= fudge; j++) {
						shift += j * dimensions[ZZ];
						for (k=-fudge; k <= fudge; k++) {
							shift += l + k;
							if (shift >= 0 && shift < n && mask[shift])
								mask[l] = 1;
						}
					}
				}
			}
		}
	}

	/* report masked elements? */
	if (debug) {
		for (i=0; i < n; i++) {
			if (mask[i]) 
				count++;
		}
		fprintf(stdout, "%d grid elements masked out.\n", count);
		count = 0;
	}
	/* calculate avg. pressure */
	clear_mat(avg);
	for (i=0; i < dimensions[XX]; i++) {
		for (j=0; j < dimensions[YY]; j++) {
			for (k=0; k < dimensions[ZZ]; k++) {
				node = flatten(i, j, k, dimensions);
				if (mask[node] == 0) {
					m_add(avg, grid[node], avg);
					count++;
				}
			}
		}
	}
	/* report # of grid elements used? */
	if (verbose || debug) 
		fprintf(stdout, "\nUsing %d grid elements.\n", count);
	/* report pressure tension before normalisation? */
	if (debug) {
		fprintf(stdout, "Total, summed pressure tensor:\n");
		for (i=0; i < DIM; i++) 
			for (j=0; j < DIM; j++)
				fprintf(stderr, "  avg[%d][%d]=%g\n", i, j, avg[i][j]);
	}
	/* normalise pressure tensor */
	msmul(avg, (real) 1.0 / count, avg);
	/* report pressure tensor? */
	if (verbose || debug) {
		fprintf(stdout, "Normalised pressure tensor:\n");
		for (i=0; i < DIM; i++) 
			for (j=0; j < DIM; j++)
				fprintf(stderr, "  avg[%d][%d]=%g\n", i, j, avg[i][j]);
	}
	/* calculate average and std. deviation */
	a = (avg[XX][XX] + avg[YY][YY] + avg[ZZ][ZZ]) / 3.0;
	e = sqrt( (pow(avg[XX][XX] - a, 2) + pow(avg[YY][YY] - a, 2) 
				+ pow(avg[ZZ][ZZ] - a, 2)) / 3.0 );
	/* report average pressure and XX, YY, and ZZ components */
	fprintf(stdout, "\n");
	fprintf(stdout, "Components: (%g, %g, %g)\n", avg[XX][XX], avg[YY][YY], 
			avg[ZZ][ZZ]);
	fprintf(stdout, "Average pressure: (%g +/- %g) bar\n", a, e);
	/* write mask? */
	if (bMask) 
		write_mask(opt2fn("-o",NFILE,fnm), mask, count, dimensions, resolution);
	/* close down */
	close_trj(status);
	thanx(stderr);
	return 0;
}

